Package: sets
Version: 1.0-25
Title: Sets, Generalized Sets, Customizable Sets and Intervals
Description: Data structures and basic operations for ordinary sets,
             generalizations such as fuzzy sets, multisets, and
             fuzzy multisets, customizable sets, and intervals.
Authors@R: c(person(given = "David", family = "Meyer", role = c("aut", "cre"),
                    email = "David.Meyer@R-project.org",
		    comment = c(ORCID = "0000-0002-5196-3048")),	    
             person(given = "Kurt", family = "Hornik", role = "aut",
	            email = "Kurt.Hornik@R-project.org",
                    comment = c(ORCID = "0000-0003-4198-9911")),
             person(given = "Christian", family = "Buchta", role = "ctb"))
Depends: R (>= 2.7.0)
Suggests: proxy
Imports: graphics,grDevices,stats,utils
License: GPL-2
NeedsCompilation: yes
Packaged: 2023-12-05 17:39:41 UTC; meyer
Author: David Meyer [aut, cre] (<https://orcid.org/0000-0002-5196-3048>),
  Kurt Hornik [aut] (<https://orcid.org/0000-0003-4198-9911>),
  Christian Buchta [ctb]
Maintainer: David Meyer <David.Meyer@R-project.org>
Repository: CRAN
Date/Publication: 2023-12-06 09:22:33 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4.0.2; 2026-02-18 00:58:45 UTC; unix

Package: sets
Version: 1.0-22
Title: Sets, Generalized Sets, Customizable Sets and Intervals
Description: Data structures and basic operations for ordinary sets,
             generalizations such as fuzzy sets, multisets, and
             fuzzy multisets, customizable sets, and intervals.
Authors@R: c(person(given = "David", family = "Meyer", role = c("aut", "cre"),
                    email = "David.Meyer@R-project.org",
		    comment = c(ORCID = "0000-0002-5196-3048")),	    
             person(given = "Kurt", family = "Hornik", role = "aut",
	            email = "Kurt.Hornik@R-project.org",
                    comment = c(ORCID = "0000-0003-4198-9911")),
             person(given = "Christian", family = "Buchta", role = "ctb"))
Depends: R (>= 2.7.0)
Suggests: proxy
Imports: graphics,grDevices,stats,utils
License: GPL-2
NeedsCompilation: yes
Packaged: 2023-02-01 11:02:23 UTC; hornik
Author: David Meyer [aut, cre] (<https://orcid.org/0000-0002-5196-3048>),
  Kurt Hornik [aut] (<https://orcid.org/0000-0003-4198-9911>),
  Christian Buchta [ctb]
Maintainer: David Meyer <David.Meyer@R-project.org>
Repository: CRAN
Date/Publication: 2023-02-01 13:22:05 UTC

useDynLib("sets", .registration = TRUE)

importFrom("graphics", "axis", "barplot", "lines", "par", "plot", "text")
importFrom("grDevices", "grey.colors")
importFrom("stats", "approxfun", "dnorm", "median", "na.omit", "weighted.mean")
importFrom("utils", "combn", "str")

export("%e%",
       "%D%",
       "as.set",
       "canonicalize_set_and_mapping",
       "is.set",
       "make_set_with_order",
       "set",
       "set_cardinality",
       "set_cartesian",
       "set_combn",
       "set_complement",
       "set_contains_element",
       "set_intersection",
       "set_is_empty",
       "set_is_equal",
       "set_is_proper_subset",
       "set_is_subset",
       "set_outer",
       "set_power",
       "set_similarity",
       "set_dissimilarity",
       "set_symdiff",
       "set_union",
       "sets_options"
       )

S3method("[", "set")
S3method("[[", "set")
S3method("[<-", "set")
S3method("[[<-", "set")
S3method("Math","set")
S3method("Summary","set")
S3method("all.equal","set")
S3method("as.list", "set")
S3method("as.vector", "set")
S3method("make_set_with_order", "character")
S3method("make_set_with_order", "data.frame")
S3method("make_set_with_order", "matrix")
S3method("make_set_with_order", "default")
S3method("make_set_with_order", "factor")
S3method("make_set_with_order", "integer")
S3method("make_set_with_order", "list")
S3method("make_set_with_order", "logical")
S3method("make_set_with_order", "NULL")
S3method("make_set_with_order", "numeric")
S3method("make_set_with_order", "ordered")
S3method("make_set_with_order", "set")
S3method("make_set_with_order", "gset")
S3method("make_set_with_order", "cset")
S3method("make_set_with_order", "tuple")
S3method("c","set")
S3method("format", "set")
S3method("length", "set")
S3method("lengths", "set")
S3method("mean", "set")
S3method("median", "set")
S3method("plot","set")
S3method("lines","set")
S3method("print","set")
S3method("print","summary.set")
S3method("rep","set")
S3method("str","set")
S3method("summary","set")

## Generalized set stuff.
export("e")
export("is_element")
S3method("format","gset_element")
S3method("LABEL","gset_element")
S3method("print","gset_element")
S3method("as.list","gset_element")
S3method("as.vector","gset_element")

export("%e%",
       "%D%",
       "as.gset",
       "gset",
       "gset_bound",
       "gset_cardinality",
       "gset_cartesian",
       "gset_charfun",
       "gset_combn",
       "gset_complement",
       "gset_core",
       "gset_concentrate",
       "gset_contains_element",
       "gset_defuzzify",
       "gset_difference",
       "gset_dilate",
       "gset_has_missings",
       "gset_height",
       "gset_intersection",
       "gset_is_crisp",
       "gset_is_empty",
       "gset_is_equal",
       "gset_is_fuzzy_multiset",
       "gset_is_fuzzy_set",
       "gset_is_multiset",
       "gset_is_proper_subset",
       "gset_is_set",
       "gset_is_set_or_fuzzy_set",
       "gset_is_set_or_multiset",
       "gset_is_subset",
       "gset_memberships",
       "gset_transform_memberships",
       "gset_normalize",
       "gset_outer",
       "gset_peak",
       "gset_power",
       "gset_product",
       "gset_mean",
       "gset_similarity",
       "gset_dissimilarity",
       "gset_sum",
       "gset_support",
       "gset_symdiff",
       "gset_union",
       "gset_universe",
       "is.gset"
       )

#S3method("Ops","gset")
S3method("Summary","gset")
S3method("[", "gset")
S3method("[[", "gset")
S3method("[<-", "gset")
S3method("[[<-", "gset")
S3method("all.equal","gset")
S3method("as.gset", "character")
S3method("as.gset", "data.frame")
S3method("as.gset", "matrix")
S3method("as.gset", "default")
S3method("as.gset", "factor")
S3method("as.gset", "gset")
S3method("as.gset", "cset")
S3method("as.gset", "integer")
S3method("as.gset", "list")
S3method("as.gset", "logical")
S3method("as.gset", "numeric")
S3method("as.gset", "ordered")
S3method("as.gset", "tuple")
S3method("as.list", "gset")
S3method("as.vector", "gset")
S3method("c","gset")
S3method("cut","gset")
S3method("format", "gset")
S3method("length", "gset")
S3method("lengths", "gset")
S3method("mean", "gset")
S3method("median", "gset")
S3method("plot", "gset")
S3method("lines", "gset")
S3method("plot", "charfun_generator")
S3method("lines", "charfun_generator")
S3method("print","gset")
S3method("print","gset_charfun")
S3method("print","summary.gset")
S3method("rep","gset")
S3method("str","gset")
S3method("summary","gset")

## Customizable set stuff
export("as.cset",
       "cset",
       "cset_bound",
       "cset_cardinality",
       "cset_cartesian",
       "cset_charfun",
       "cset_combn",
       "cset_complement",
       "cset_concentrate",
       "cset_contains_element",
       "cset_core",
       "cset_defuzzify",
       "cset_difference",
       "cset_dilate",
       "cset_has_missings",
       "cset_height",
       "cset_intersection",
       "cset_is_crisp",
       "cset_is_empty",
       "cset_is_equal",
       "cset_is_fuzzy_multiset",
       "cset_is_fuzzy_set",
       "cset_is_multiset",
       "cset_is_proper_subset",
       "cset_is_set",
       "cset_is_set_or_fuzzy_set",
       "cset_is_set_or_multiset",
       "cset_is_subset",
       "cset_matchfun",
       "cset_matchfun<-",
       "cset_memberships",
       "cset_transform_memberships",
       "cset_normalize",
       "cset_orderfun",
       "cset_orderfun<-",
       "cset_outer",
       "cset_power",
       "cset_product",
       "cset_mean",
       "cset_peak",
       "cset_similarity",
       "cset_dissimilarity",
       "cset_sum",
       "cset_support",
       "cset_symdiff",
       "cset_union",
       "cset_universe",
       "is.cset",
       "matchfun"
       )

S3method("Ops","cset")
S3method("Summary","cset")
S3method("[", "cset")
S3method("[[", "cset")
S3method("[<-", "cset")
S3method("[[<-", "cset")
S3method("all.equal","cset")
S3method("as.cset", "character")
S3method("as.cset", "data.frame")
S3method("as.cset", "matrix")
S3method("as.cset", "default")
S3method("as.cset", "factor")
S3method("as.cset", "cset")
S3method("as.cset", "integer")
S3method("as.cset", "list")
S3method("as.cset", "logical")
S3method("as.cset", "numeric")
S3method("as.cset", "ordered")
S3method("as.cset", "tuple")
S3method("as.list", "cset")
S3method("as.vector", "cset")
S3method("as.character", "cset")
S3method("c","cset")
S3method("cut","cset")
S3method("format", "cset")
S3method("length", "cset")
S3method("lengths", "cset")
S3method("mean", "cset")
S3method("median", "cset")
S3method("na.omit", "cset")
S3method("plot", "cset")
S3method("lines", "cset")
S3method("print","cset")
S3method("print","cset_charfun")
S3method("print","summary.cset")
S3method("rep","cset")
S3method("str","cset")
S3method("summary","cset")

## Closure and reduction
export("binary_closure")
export("closure")
S3method("closure","set")
S3method("closure","default")

export("binary_reduction")
export("reduction")
S3method("reduction","set")
S3method("reduction","default")

## Tuple stuff.
export("%e%",
       "as.tuple",
       "is.tuple",
       "pair",
       "singleton",
       "triple",
       "tuple",
       "tuple_is_ntuple",
       "tuple_is_pair",
       "tuple_is_singleton",
       "tuple_is_triple",
       "tuple_outer"
       )

S3method("[", "tuple")
S3method("[<-", "tuple")
S3method("Math","tuple")
S3method("Ops","tuple")
S3method("Summary","tuple")
S3method("print","tuple")
S3method("format", "tuple")
S3method("as.tuple", "default")
S3method("as.tuple", "factor")
S3method("as.tuple", "integer")
S3method("as.tuple", "list")
S3method("as.tuple", "logical")
S3method("as.tuple", "character")
S3method("as.tuple", "numeric")
S3method("as.tuple", "ordered")
S3method("as.tuple", "tuple")
S3method("as.tuple", "data.frame")
S3method("as.tuple", "set")
S3method("as.tuple", "cset")
S3method("as.tuple", "gset")
S3method("as.list", "tuple")
S3method("as.vector", "tuple")
S3method("plot","tuple")
S3method("lines","tuple")
S3method("c", "tuple")
S3method("mean", "tuple")
S3method("median", "tuple")
S3method("rep", "tuple")

## LABEL(S)
export("LABEL",
       "LABELS")

S3method("LABEL", "character")
S3method("LABEL", "default")
S3method("LABEL", "factor")
S3method("LABEL", "integer")
S3method("LABEL", "logical")
S3method("LABEL", "matrix")
S3method("LABEL", "numeric")
S3method("LABEL", "set")
S3method("LABEL", "interval")
S3method("LABEL", "gset")
S3method("LABEL", "cset")
S3method("LABEL", "tuple")
S3method("LABEL", "list")


### fuzzy logic stuff
export(".N.",
       ".T.",
       ".S.",
       ".I.",
       "%is%",
       "charfun_generator",
       "fuzzy_bell",
       "fuzzy_bell_gset",
       "fuzzy_cone",
       "fuzzy_cone_gset",
       "fuzzy_inference",
       "fuzzy_logic",
       "fuzzy_normal",
       "fuzzy_normal_gset",
       "fuzzy_partition",
       "fuzzy_pi3",
       "fuzzy_pi4",
       "fuzzy_pi3_gset",
       "fuzzy_pi4_gset",
       "fuzzy_rule",
       "fuzzy_sigmoid",
       "fuzzy_sigmoid_gset",
       "fuzzy_system",
       "fuzzy_trapezoid",
       "fuzzy_trapezoid_gset",
       "fuzzy_triangular",
       "fuzzy_triangular_gset",
       "fuzzy_tuple",
       "fuzzy_two_normals",
       "fuzzy_two_normals_gset",
       "fuzzy_variable",
       "is.charfun_generator"
       )

S3method("print","fuzzy_rule")
S3method("print","fuzzy_system")
S3method("print","fuzzy_variable")
S3method("plot","fuzzy_system")
S3method("plot","fuzzy_variable")

## interval stuff
export("%..%",
       "%<%",
       "%>%",
       "%<=%",
       "%>=%",
       "as.interval",
       "integers",
       "integers2reals",
       "interval",
       "interval_complement",
       "interval_contains_element",
       "interval_difference",
       "interval_division",
       "interval_domain",
       "interval_intersection",
       "interval_is_bounded",
       "interval_is_closed",
       "interval_is_countable",
       "interval_is_degenerate",
       "interval_is_empty",
       "interval_is_equal",
       "interval_is_finite",
       "interval_is_greater_than",
       "interval_is_greater_than_or_equal",
       "interval_is_half_bounded",
       "interval_is_left_bounded",
       "interval_is_left_closed",
       "interval_is_left_open",
       "interval_is_less_than",
       "interval_is_less_than_or_equal",
       "interval_is_left_unbounded",
       "interval_is_proper",
       "interval_is_proper_subinterval",
       "interval_is_right_bounded",
       "interval_is_right_closed",
       "interval_is_right_open",
       "interval_is_right_unbounded",
       "interval_is_subinterval",
       "interval_is_unbounded",
       "interval_is_uncountable",
       "interval_measure",
       "interval_power",
       "interval_product",
       "interval_sum",
       "interval_symdiff",
       "interval_union",
       "is.interval",
       "naturals",
       "naturals0",
       "reals",
       "reals2integers"
       )

S3method("Math", "interval")
S3method("abs", "interval")
S3method("as.character", "interval")
S3method("as.double", "interval")
S3method("as.interval", "default")
S3method("as.interval", "interval")
S3method("as.interval", "logical")
S3method("as.interval", "list")
S3method("as.interval", "NULL")
S3method("as.interval", "numeric")
S3method("as.interval", "set")
S3method("as.interval", "tuple")
S3method("as.list", "interval")
S3method("as.vector", "interval")
S3method("c", "interval")
S3method("format", "interval")
S3method("[", "interval")
S3method("[[", "interval")
S3method("length", "interval")
S3method("make_set_with_order", "interval")
S3method("max", "interval")
S3method("mean", "interval")
S3method("min", "interval")
S3method("Ops", "interval")
S3method("print", "interval")
S3method("print", "summary.interval")
S3method("prod", "interval")
S3method("range", "interval")
S3method("str", "interval")
S3method("sum", "interval")
S3method("summary", "interval")

### internal dispatch
S3method(".as.list", "default")
S3method(".as.list", "set")
S3method(".as.list", "gset")

### intended for internal usage only
export(".cartesian_product",
       ".exact_match",
       ".get_memberships",
       ".make_gset_from_support_and_memberships",
       ".make_set_of_tuples_from_list_of_lists",
       ".ntuple",
       ".set_subset",
       "fuzzy_logic_predicates")

# Sets

This is a Haskell package which defines sets in a more general way than Data.Set.

Data.Set only allows to create sets of types which implements the Ord typeclass, this package introduces HomogeneousSet which does not require the typeclass Ord.

The PureSet type allows to create nested sets easily. It may be used to study set theory for example.

## General info

The homogenous sets implemented in this package require the type contained in the set to implement the typeclass Eq. It ensures there are no duplicate element in the set and that the order of elements does not matter when testing equality. It is slower than Data.Set because we do not require the Ord typeclass, if you only use types which are orderable, use Data.Set instead.

The pure set type implemented in this package is a tree like structure where the order of branches does not matter. It allows arbitrary nesting of sets which is useful to do set theoretic constructions.

## Installation

`cabal install Sets`

## Usage

Example usage of homogenous sets :

```haskell
ghci> import HomogeneousSet
ghci> data Foo = Foo Int Char deriving (Eq) -- an arbitrary type which is not required to implement >Ord typeclass
ghci> s1 = set [Foo 3 'a', Foo 2 'c', Foo 3 'a']
ghci> s2 = set [Foo 2 'c', Foo 3 'a']
ghci> s1 == s2
True
```

Example usage of pure sets :

```haskell
ghci> import PureSet
ghci> numberToSet 3
(pureSet [(pureSet []),(pureSet [(pureSet [])]),(pureSet [(pureSet []),(pureSet [(pureSet [])])])])
ghci> putStrLn.prettify $ numberToSet 3
{{}, {{}}, {{}, {{}}}}
```

## Contribution

Any input is appreciated ! Send an email for any remark or question.
The git repository : https://gitlab.utc.fr/gsabbagh/sets

closure <-
function(x, operation, ...)
    UseMethod("closure")

closure.default <-
function(x, operation, ...)
    stop("Not implemented.")

closure.set <-
function(x, operation = c("union", "intersection"), ...)
{
    if (length(x) < 2L) return(x)
    if (!all(sapply(x, is.cset)))
        stop("closure only defined on set of (c,g)sets.")

    if (all(sapply(x, is.set))) {

        dom <- .as.list(do.call(set_union, x))
        x <- lapply(x, .make_list_elements)

        members <-
            binary_closure(do.call(rbind, lapply(x, function(i) dom %in% i)),
                           operation)

        .make_set_from_list(.list_sort(apply(members, 1L,
                                             function(i) .make_set_from_list(dom[i])
                                             )
                                       )
                            )
    } else if (all(sapply(x, is.gset))) {
        operation <- paste0("gset_", match.arg(operation))
        len <- 0
        while ((newlen <- length(gset_support(x))) != len) {
            len <- newlen
            x <- c(x,
                   as.set(lapply(gset_combn(x, 2L),
                                 function(i) do.call(gset_union, i))
                          )
                   )
        }
        x
    } else {
        operation <- paste0("cset_", match.arg(operation))
        len <- 0
        while ((newlen <- length(cset_support(x))) != len) {
            len <- newlen
            x <- c(x,
                   as.set(lapply(cset_combn(x, 2L),
                                 function(i) do.call(cset_union, i))
                          )
                   )
        }
        x
    }
}

binary_closure <-
function(x, operation = c("union", "intersection"))
    .Call(sets_closure, x,
          pmatch(match.arg(operation), c("union", "intersection")))

##########################
######### Tuples #########
##########################

### constructors

tuple <-
function(...)
{
    l <- list(...)
    nam <- names(l)
    if (!is.null(nam)) {
        nam <- rep_len(nam, length.out = length(l))
        names(l) <- nam
    }
    .make_tuple_from_list(l)
}

.make_tuple_from_list <-
function(x)
    .structure(x, class = "tuple")

singleton <-
function(...)
{
    if(length(list(...)) != 1L)
        stop("Exactly one argument expected.")
    tuple(...)
}

pair <-
function(...)
{
    if(length(list(...)) != 2L)
        stop("Exactly two arguments expected.")
    tuple(...)
}

triple <-
function(...)
{
    if(length(list(...)) != 3L)
        stop("Exactly three arguments expected.")
    tuple(...)
}

## quadruple <- function(...) {
##   if(length(list(...)) != 4L)
##     stop("Need four components.")
##   tuple(...)
## }

## quintuple <- function(...) {
##   if(length(list(...)) != 5L)
##     stop("Need five components.")
##   tuple(...)
## }

## sextuple <- function(...) {
##   if(length(list(...)) != 6L)
##     stop("Need six components.")
##   tuple(...)
## }

## septuple <- function(...) {
##   if(length(list(...)) != 7L)
##     stop("Need seven components.")
##   tuple(...)
## }

## octuple <- function(...) {
##   if(length(list(...)) != 8L)
##     stop("Need eight components.")
##   tuple(...)
## }

## nonuple <- function(...) {
##   if(length(list(...)) != 9L)
##     stop("Need nine components.")
##   tuple(...)
## }

## decuple <- function(...) {
##   if(length(list(...)) != 10L)
##     stop("Need ten components.")
##   tuple(...)
## }

### predicates

is.tuple <-
function(x)
    inherits(x, "tuple")

tuple_is_ntuple <-
function(x, n)
    is.tuple(x) && (length(x) == n)

tuple_is_singleton <-
function(x)
    tuple_is_ntuple(x, 1L)

tuple_is_pair <-
function(x)
    tuple_is_ntuple(x, 2L)

tuple_is_triple <-
function(x)
    tuple_is_ntuple(x, 3L)

## tuple_is_quadruple <- function(x)
##   tuple_is_ntuple(x, 4L)

## tuple_is_quintuple <- function(x)
##   tuple_is_ntuple(x, 5L)

## tuple_is_sextuple <- function(x)
##   tuple_is_ntuple(x, 6L)

## tuple_is_septuple  <- function(x)
##   tuple_is_ntuple(x, 7L)

## tuple_is_octuple <- function(x)
##   tuple_is_ntuple(x, 8L)

## tuple_is_nonuple <- function(x)
##   tuple_is_ntuple(x, 9L)

## tuple_is_decuple <- function(x)
##   tuple_is_ntuple(x, 10L)

### operators, methods

c.tuple <-
function(...)
    do.call(tuple,
            unlist(lapply(list(...),
                          function(i) if (is.tuple(i)) i else list(i)),
                   recursive = FALSE)
            )

"[.tuple" <-
function(x, i)
    .make_tuple_from_list(NextMethod("["))

"[<-.tuple" <-
function(x, i, value)
    .make_tuple_from_list(NextMethod("[<-"))

rep.tuple <-
function(x, ...)
    .make_tuple_from_list(rep(.as.list(x), ...))

Ops.tuple <-
function(e1, e2)
{
    if (!missing(e2)) {
        len <- max(length(e1), length(e2))
        e1 <- rep_len(e1, length.out = len)
        e2 <- rep_len(e2, length.out = len)
        ret <- lapply(seq_len(len),
                      function(i) do.call(.Generic, list(e1[[i]], e2[[i]])))
    } else {
        ret <- lapply(seq_along(e1),
                      function(i) do.call(.Generic, list(e1[[i]])))
    }

    .make_tuple_from_list(ret)

}

Math.tuple <-
function(x, ...)
    as.tuple(get(.Generic)(unlist(x), ...))

Summary.tuple <-
function(..., na.rm = FALSE)
    do.call(.Generic, c(as.list(...), na.rm = na.rm))

mean.tuple <-
function(x, ...)
{
    x <- as.numeric(x)
    NextMethod()
}

## <FIXME>
## Simplify when depending on R >= 3.4.0
if(is.na(match("...", names(formals(median))))) {
    median.tuple <- function(x, na.rm = FALSE) {
        x <- as.numeric(x)
        NextMethod()
    }
} else {
    median.tuple <- function(x, na.rm = FALSE, ...) {
        x <- as.numeric(x)
        NextMethod()
    }
}
## </FIXME>

tuple_outer <- set_outer

### output

print.tuple <-
function(x, ...)
{
    writeLines(strwrap(format(x, ...), exdent = 1L))
    invisible(x)
}

format.tuple <-
function(x, ...)
    .format_set_or_tuple(x, "(", ")", ...)

.ntuple <-
function(i, a = FALSE, A = FALSE, plural = FALSE)
{
    if (plural && (a || A))
      stop("Wrong arguments.")
    if (is.tuple(i))
        i <- length(i)
    if (i == 0L)
        paste(if (A) "The" else "the", "empty tuple")
    else if (i > 10L)
        paste(if (A) "A " else if (a) "a " else "",
              i,
              if (plural) "-tuples" else "-tuple",
              sep = "")
    else
        paste(if (a) .TUPLESa[i] else if (A) .TUPLESA[i] else "",
              if (plural) .P_TUPLES[i] else  .S_TUPLES[i], sep = "")
}

### constants

.S_TUPLES <- c("singleton", "pair", "triple", "quadruple", "quintuple",
               "sextuple", "septuple", "octuple", "nonuple", "decuple")
.P_TUPLES <- c("singletons", "pairs", "triples", "quadruples", "quintuples",
               "sextuples", "septuples", "octuples", "nonuples", "decuples")

.TUPLESa <- c("a ", "a ", "a ", "a ", "a ", "a ", "a ", "an ", "a ", "a ")
.TUPLESA <- c("A ", "A ", "A ", "A ", "A ", "A ", "A ", "An ", "A ", "A ")

## * Element function, used for the creation of gsets by elements

e <-
function(x, memberships = 1L)
{
    if (is_element(x)) return(x)
    if (length.set(memberships) > 1L)
        memberships <- as.gset(memberships)
    else if (isTRUE(memberships >= 1))
        memberships <- as.integer(memberships)
    .stop_if_memberships_are_invalid(memberships)
    .make_element_from_support_and_memberships(x, memberships)
}

.make_element_from_support_and_memberships <-
function(x, memberships)
    .structure(list(x),
               memberships = memberships,
               class = "gset_element")


print.gset_element <-
function(x, ...)
{
    writeLines(format(x, ...))
    invisible(x[[1]])
}

format.gset_element <-
function(x, ...)
{
    paste(paste(LABEL(x[[1]], ...), collapse = " "),
          " [",
          paste(format(.get_memberships(x)), collapse = ", "),
          "]",
          sep = "")
}

LABEL.gset_element <-
function(x, limit, ...)
    format(x, ...)


as.vector.gset_element <-
    function(x, ...)
        list(x)

as.list.gset_element <-
    function(x, ...)
        unclass(x)

##### interval class
### FIXME: actually, multi-intervals -> change class name?

## FIXME: which functions should convert their arguments via
## as.interval?

## constructors

## FIXME: allow for "boxes" (dimensionality > 1)?

## FIXME: what should interval() / integers() etc. return: the domain,
## or an empty set? Currently, intergers/reals/naturals return the domain,
## interval the empty interval.

"%..%" <-
integers <-
function(l = NULL, r = NULL)
{
    if (is.null(l)) l <- -Inf
    if (is.null(r)) r <- Inf
    if (!is.numeric(l) || !is.numeric(r))
        stop("Need numeric bounds.")
    l <- round(l)
    r <- round(r)
    interval(l = l, r = r, domain = "Z")
}

naturals <-
function(l = NULL, r = NULL)
{
    if (is.null(l)) l <- 1
    if (is.null(r)) r <- Inf
    if (!is.numeric(l) || !is.numeric(r))
        stop("Need numeric bounds.")
    l <- round(l)
    r <- round(r)
    if (l < 1L) stop("Positive integers start with 1.")
    interval(l = l, r = r, domain = "N")
}

naturals0 <-
function(l = NULL, r = NULL)
{
    if (is.null(l)) l <- 0
    if (is.null(r)) r <- Inf
    if (!is.numeric(l) || !is.numeric(r))
        stop("Need numeric bounds.")
    l <- round(l)
    r <- round(r)
    if (l < 0) stop("Non-negative integers start with 0.")
    interval(l = l, r = r, domain = "N0")
}

reals <-
function(l = NULL, r = NULL,
         bounds = c("[]", "[)", "(]", "()",
                    "[[", "]]", "][",
                    "open", "closed", "left-open", "right-open",
                    "left-closed", "right-closed")
         )
{
    if (is.null(l)) l <- -Inf
    if (is.null(r)) r <- Inf
    if (!is.numeric(l) || !is.numeric(r))
        stop("Need numeric bounds.")
    interval(l = l, r = r, bounds = bounds, domain = "R")
}

interval <-
function(l = NULL, r = l,
         bounds = c("[]", "[)", "(]", "()",
                    "[[", "]]", "][",
                    "open", "closed", "left-open", "right-open",
                    "left-closed", "right-closed"),
         domain = NULL)
{
    ## bounds
    bounds <- match.arg(bounds)
    bounds <- switch(bounds,
                  "open"=, "]["= "()", "closed"= "[]",
                   "left-open"=, "right-closed"=, "]]"= "(]",
                     "right-open"=, "left-closed"=, "[["= "[)", bounds)

    ## check for empty interval
    if (is.null(l) || is.null(r) || r < l ||
        isTRUE(all.equal(l, r)) && bounds != "[]")
        return(.empty_interval())

    ## domain
    if (is.null(domain)) {
        ## guess domain from mode
        domain <- if ((is.double(l) || is.double(r)) &&
                      !(is.integer(l) && is.infinite(r)) &&
                      !(is.integer(r) && is.infinite(l)))
            "R"
        else if (l > 0)
            "N"
        else if (l > -1L)
            "N0"
        else
            "Z"
    } else {
        ## enforce domain
        domain <- match.arg(domain, c("R", "Z", "N", "N0"))
        if (domain == "R") {
            l <- as.double(l)
            r <- as.double(r)
        } else {
            l <- round(l)
            r <- round(r)
            if (domain == "N" && l < 1L)
                stop("Positive integers start with 1.")
            if (domain == "N0" && l < 0)
                stop("Non-negative integers start with 0.")
        }
    }

    if (!is.numeric(l) || !is.numeric(r))
        stop("Need numeric bounds.")

    ## create interval
    .make_interval_from_list_and_domain(list(list(l = l, r = r,
                                                  lc = substr(bounds, 1L, 1L) == "[",
                                                  rc = substr(bounds, 2L, 2L) == "]"
                                                  )
                                             ),
                                        domain = domain
                                        )
}

## subsetting

`[.interval` <-
function(x, i)
{
    i <- i[i <= length(x)]
    .make_interval_from_list_and_domain(NextMethod(), interval_domain(x))
}

## subscripting
## FIXME: should that be the same than subsetting?

`[[.interval` <-
function(x, i)
    .make_interval_from_list_and_domain(list(NextMethod()), interval_domain(x))

## union = combine

interval_union <-
function(...) c.interval(...)

c.interval <-
function(...)
{
    ## coerce to intervals
    l <- lapply(list(...), as.interval)

    ## remove empty intervals
    l <- l[!sapply(l, interval_is_empty)]
    if (length(l) < 1L) return(.empty_interval())

    ## check domains
    dom <- .maximal_domain(l)

    ## split countable intervals, if any uncountable exists
    countables <- sapply(l, interval_is_countable)
    if (any(countables) && !all(countables))
        l[countables] <- lapply(l[countables], integers2reals)

    l <- unlist(l, recursive = FALSE)

    ## canonicalize
    .make_interval_from_list_and_domain(.canonicalize_interval(l, dom == "R"),
                                        domain = dom)
}

## complement

interval_complement <-
function(x, y = NULL)
{
    x <- as.interval(x)
    x_is_c <- interval_is_countable(x)

    if (!is.null(y)) {
        y <- as.interval(y)
        y_is_c <- interval_is_countable(y)

        ## trivial cases
        if (interval_is_empty(y) || interval_is_empty(x))
            return(y)

        ## promote/reduce x, depending on y
        if (x_is_c && !y_is_c)
            x <- integers2reals(x, min(y), max(y))
        if (!x_is_c && y_is_c)
            x <- reals2integers(x)
        attr(x, "domain") <- attr(y, "domain")

        return(interval_complement(interval_union(x, interval_complement(y))))
    }

    if (interval_is_empty(x))
        return(reals())

    dom <- interval_domain(x)

    i_is_c <- interval_is_countable(x)

    ## bounded parts
    len <- length.interval(x)
    l <- if (len > 1L)
        lapply(seq_len(len - 1L),
               function(i) list(l = max(x[[i]]) + x_is_c,
                                r = min(x[[i + 1L]]) - x_is_c,
                                lc = x_is_c || interval_is_right_open(x[[i]]),
                                rc = x_is_c || interval_is_left_open(x[[i + 1L]])
                                )
               )
    else
        list()

    ## unbounded parts
    if (dom %in% c("R","Z") && interval_is_left_bounded(x) ||
        dom == "N" && min(x) > 1L ||
        dom == "N0" && min(x) > 0L)
        l <- c(list(list(l = if (dom == "N0") 0 else if (dom == "N") 1 else -Inf,
                         r = min(x) - x_is_c,
                         lc = TRUE,
                         rc = x_is_c || interval_is_left_open(x)
                         )
                    ),
               l
               )

    if (interval_is_right_bounded(x))
        l <- c(l,
               list(list(l = max(x) + x_is_c,
                         r = +Inf,
                         rc = TRUE,
                         lc = x_is_c || interval_is_right_open(x)
                         )
                    )
               )

    .make_interval_from_list_and_domain(l, domain = dom)
}

## intersection
interval_intersection <-
function(...)
{
    ## coerce
    len <- length(l <- lapply(list(...), as.interval))

    ## trivial case
    if (any(sapply(l, interval_is_empty)))
        return(.empty_interval())

    ## find minimal domain
    dom <- .minimal_domain(l)
    Dom <- switch(dom, N = naturals(), N0 = naturals0(), Z = integers(), reals())

    ## check for mixture of countables and uncountables
    uc <- sapply(l, interval_is_uncountable)
    if (any(uc) && !all(uc))
        l[uc] <- lapply(l[uc], reals2integers)

    ## workhorse
    IS <- function(x, y)
        interval_complement(interval_union(interval_complement(x),
                                           interval_complement(y),
                                           if (dom == "N") integers(-Inf, 0),
                                           if (dom == "N0") integers(-Inf, -1)
                                           ),
                            Dom)

    ## reduce
    if (len < 1L)
        .empty_interval()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(IS, l)
}

## symdiff
interval_symdiff <-
function(...)
{
    len <- length(l <- lapply(list(...), as.interval))
    SD <- function(x, y) interval_union(interval_complement(x, y),
                                        interval_complement(y, x))
    if (len < 1L)
        .empty_interval()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(SD, l)
}

## length
length.interval <-
function(x)
    length(unclass(x))

interval_measure <-
function(x)
    sum(sapply(unclass(x), `[[`, "r") - sapply(unclass(x), `[[`, "l"))

## Summary methods

## FIXME: min/max not defined for open intervals??
## -> we actually compute the infimum  / supremum here.

min.interval <-
function(...)
{
    l <- list(...)
    l <- unclass(do.call(c, l[-length(l)]))
    l[[1L]]$l
}

max.interval <-
function(...)
{
    l <- list(...)
    l <- unclass(do.call(c, l[-length(l)]))
    l[[length(l)]]$r
}

range.interval <- function(...) c(min(...), max(...))

sum.interval <-
function(...)
{
    l <- list(...)
    do.call(interval_sum, l[-length(l)])
}

prod.interval <-
function(...)
{
    l <- list(...)
    do.call(interval_product, l[-length(l)])
}

mean.interval <-
function(...)
{
    l <- list(...)
    do.call(interval_sum, l) / length(l)
}

## operators

Ops.interval <-
function(e1, e2)
{
    if (nargs() == 1L) {
        if(!((ch <- as.character(.Generic)) %in% c("-", "!")))
            stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                          .Generic, .Class),
                 domain = NA)
        if (ch == "!")
            return(interval_complement(e1))
        else
            return(0L-e1)
    }

    if(!(as.character(.Generic)
         %in% c("<", "<=", ">", ">=", "==", "!=",
                "&", "|", "*", "+", "-", "/", "^")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    switch(.Generic,
           "|"  = interval_union(e1, e2),
           "&"  = interval_intersection(e1, e2),

           "+"  = interval_sum(e1, e2),
           "-"  = interval_difference(e1, e2),
           "*"  = interval_product(e1, e2),
           "/"  = interval_division(e1, e2),

           "<"  = interval_is_proper_subinterval(e1, e2),
           "<=" = interval_is_subinterval(e1, e2),
           ">"  = interval_is_proper_subinterval(e2, e1),
           ">=" = interval_is_subinterval(e2, e1),
           "==" = interval_is_equal(e1, e2),
           "!=" = !interval_is_equal(e1, e2),

           "^" = if (is.interval(e1) && is.numeric(e2))
               interval_power(e1, e2)
           else if (min(e1) > 0)
               exp(log(e1) * e2)
           else stop("Exponentiation undefined for negative base.")
           )
}

abs.interval <-
function(x)
{
    dom <- interval_domain(x)
    FUN <- function(i) if (i$l >= 0)
        .make_interval_from_list_and_domain(list(i), dom)
    else if (i$r <= 0)
        .make_interval_from_list_and_domain(list(list(l = abs(i$r),
                                                      r = abs(i$l),
                                                      lc = i$rc,
                                                      rc = i$lc
                                                      )
                                                 ),
                                            domain = dom
                                            )
    else
        .make_interval_from_list_and_domain(list(list(l = 0,
                                                      r = max(abs(i$l), i$r),
                                                      lc = TRUE,
                                                      rc = if (abs(i$l) > i$r) i$lc else i$rc
                                                      )
                                                 ),
                                            domain = dom
                                            )

    do.call(c.interval, lapply(unclass(x), FUN))
}

Math.interval <-
function(x, ...)
{
    if (.Generic %in% c("log", "sqrt") && min(x) < 0)
        stop(.Generic, " only defined for positive reals.")

    if (!.Generic %in% c("sqrt", "exp", "log", "log2", "log10",
                         "trunc", "round", "ceiling", "floor",
                         "sin", "cos", "tan", "asin", "acos", "atan"))
        stop(.Generic, " not implemented for intervals.")

    dom <- interval_domain(x)
    FUN <- if (dom == "R")
        function(i) {
            if (.Generic %in% c("sin", "cos", "tan")) {
                if (.Generic %in% c("sin", "cos") &&
                    (i$r - i$l > 2 * pi ||
                     isTRUE(all.equal(i$r - i$l, 2 * pi)))) {
                    l <- -1
                    r <- 1
                } else if (.Generic == "tan" &&
                           (i$r - i$l > pi ||
                            isTRUE(all.equal(i$r - i$l, pi)))) {
                    l <- -Inf
                    r <- Inf
                } else {
                    shift <- ceiling(i$l / (2 * pi)) * 2 * pi
                    l <- i$l - shift
                    r <- i$r - shift

                    p <- pi * -3:3 / 2
                    fp <- p[p > l & p < r |
                            sapply(p, function(i) isTRUE(all.equal(i, l))) |
                            sapply(p, function(i) isTRUE(all.equal(i, r)))]

                    x <- get(.Generic)(c(l, fp, r))
                    l <- min(x)
                    r <- max(x)
                }
            } else if (.Generic == "acos") { ## cos is monotonic decreasing
                l <- get(.Generic)(i$r, ...)
                r <- get(.Generic)(i$l, ...)
            } else { ## all monotonic increasing funs
                l <- get(.Generic)(i$l, ...)
                r <- get(.Generic)(i$r, ...)
            }

            ret <- .make_interval_from_list_and_domain(list(list(l = l,
                                                                 r = r,
                                                                 lc = i$lc,
                                                                 rc = i$rc
                                                                 )
                                                            ),
                                                       domain = dom
                                                       )

            ## dicretize for rounding functions
            if (.Generic %in% c("trunc", "round", "floor")) {
                digits <- list(...)
                if (.Generic != "round" || !length(digits))
                    digits <- 0
                ret <- do.call(c.interval,
                               as.list(seq.int(unclass(ret)[[1]]$l,
                                               unclass(ret)[[1]]$r,
                                               by = 10 ^ -digits[[1]])))
            }
            ret
    } else
        function(i) as.interval(lapply(seq.int(i$l, i$r), .Generic, ...))

    do.call(c.interval, lapply(unclass(x), FUN))
}

## interval arithmetic

.interval_ops <-
function(l, OP)
{
    len <- length(l <- lapply(unclass(l),
                              function(i) integers2reals(as.interval(i))))
    FUN <- function(x, y)
        do.call(c.interval, set_outer(unclass(x), unclass(y), OP))

    if (len < 1L)
        .empty_interval()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(FUN, unclass(l))
}

.bounds <- function(x, y)
    c("()", "[)", "(]", "[]")[1 + x + 2 * y]

interval_sum <-
function(...)
{
   FUN <- function(x, y)
       interval(x$l + y$l, x$r + y$r,
                .bounds(x$lc && y$lc, x$rc && y$rc))

   .interval_ops(list(...), FUN)
}

interval_difference <-
function(...)
{
    FUN <- function(x, y)
        interval(x$l - y$r, x$r - y$l,
                 .bounds(x$lc && y$rc, x$rc && y$lc))

   .interval_ops(list(...), FUN)

}

interval_product <-
function(...)
{
    FUN <- function(x, y) {
        p <- c(x$l * y$l, x$l * y$r, x$r * y$l, x$r * y$r)
        b <- c(x$lc && y$lc, x$lc && y$rc, x$rc && y$lc, x$rc && y$rc)
        mi <- which.min(p)
        ma <- which.max(p)
        interval(p[mi], p[ma], .bounds(b[mi], b[ma]))
    }

    .interval_ops(list(...), FUN)

}

interval_division <-
function(...)
{
    FUN <- function(x, y) {
        if (0 < y$l || 0 > y$r)
            interval_product(interval(x$l, x$r, .bounds(x$lc, x$rc)),
                             interval(1 / y$r, 1 / y$l, .bounds(y$rc, y$lc)))
        else if (x$l < 0 && 0 < x$r && y$l <= 0 && 0 <= y$r ||
                 y$l == 0 && y$r == 0)
            reals()
        else if (x$r <= 0 && y$l < 0 && y$r == 0)
            interval(x$r / y$l, Inf, .bounds(x$rc && y$lc, TRUE))
        else if (x$r <= 0 && y$l < 0 && 0 < y$r)
            c.interval(interval(-Inf, x$r / y$r, .bounds(TRUE, x$rc && y$rc)),
                       interval(x$r / y$l, Inf, .bounds(x$rc && y$lc, TRUE)))
        else if (x$r <= 0 && y$r > 0 && y$l == 0)
            interval(-Inf, x$r / y$r, .bounds(TRUE, x$rc && y$rc))

        else if (x$l >= 0 && y$l < 0 && y$r == 0)
            interval(-Inf, x$l / y$l, .bounds(TRUE, x$lc && y$lc))
        else if (x$l >= 0 && y$l < 0 && 0 < y$r)
            c.interval(interval(-Inf, x$l / y$l, .bounds(TRUE, x$lc && y$lc)),
                       interval(x$l / y$r, Inf, .bounds(x$lc && y$rc, TRUE)))
        else if (x$l >= 0 && y$r > 0 && y$l == 0)
            interval(x$l / y$r, Inf, .bounds(x$lc && y$rc, TRUE))
        else
            stop("This should not happen.")
    }

    .interval_ops(list(...), FUN)
}

interval_power <-
function(x, n)
{
    x <- as.interval(x)

    if (!isTRUE(all.equal(as.integer(n), n)))
        stop("Power of intervals only defined for integer exponents.")

    if (n == 0)
        return(reals(1,1))

    if (n < 0)
        return(1 / interval_power(x, abs(n)))

    FUN <- function(X)
    {
        if (n %% 2 == 0 && (X$l < 0 || X$r < 0))
            interval_power(abs(x), n)
        else
            .make_interval_from_list_and_domain(list(list(l = X$l ^ n,
                                                          r = X$r ^ n,
                                                          lc = X$lc,
                                                          rc = X$rc
                                                          )
                                                     ),
                                                interval_domain(x)
                                                )
    }

    do.call(c.interval, lapply(unclass(x), FUN))

}

## coercing functions

integers2reals <-
function(x, min = -Inf, max = Inf)
{
    if(interval_is_uncountable(x)) return(x)

    ## split
    x <- unlist(lapply(unclass(x),
                       function(i) unlist(lapply(seq.int(from = max(i$l, min),
                                                         to = min(i$r, max)),
                                                 interval),
                                          recursive = FALSE)
                       ),
                recursive = FALSE)

    ## canonicalize
    .make_interval_from_list_and_domain(.canonicalize_interval(x, reals = TRUE),
                                        "R")
}

reals2integers <-
function(x)
{
    if(interval_is_countable(x)) return(x)
    x <- unclass(x)
    for(i in seq_along(x)) {
        X <- x[[i]]
        if (!X$lc || X$l != round(X$l)) X$l <- floor(X$l) + 1
        if (!X$rc || X$r != round(X$r)) X$r <- trunc(X$r)
        if (X$r < X$l)
            X <- NULL
        else
            X$lc <- X$rc <- TRUE
        x[[i]] <- X
    }

    ## canonicalize
    .make_interval_from_list_and_domain(.canonicalize_interval(x, reals = FALSE),
                                        domain = "Z")
}

as.character.interval <-
function(x, ...)
{
    if (length.interval(x) < 1L)
        return("{}")

    if (interval_is_uncountable(x)) {
        bounds <- if (sets_options("openbounds") == "][")
            c("]", "[", "[", "]")
        else
            c("(", "[", ")", "]")

        ## merge adjacent degenerate intervals into sets
        .merge = function(x, y) {
            i = length(x)
            X <- x[[i]]
            Y <- y[[1]]
            if ((is.set(X) || isTRUE(all.equal(X$l, X$r))) &&
                isTRUE(all.equal(Y$l, Y$r))) {
                if (!is.set(X)) X <- set(X$l)
                c(x[-i], list(c(X, Y$l)))
            } else
            c(x, y)
        }
        l <- Reduce(.merge, lapply(unclass(x), list))

        ## format individual interval
        .format <- function(x) {
            if (is.set(x))
                format(x)
            else if (isTRUE(all.equal(x$l, x$r)))
                paste("{", as.character(x$l), "}", sep = "")
            else
                paste(bounds[1 + x$lc], x$l, ", ", x$r,
                      bounds[3 + x$rc], sep = "")
        }

        paste(sapply(unclass(l), .format), collapse = " U ")

    } else {
        .format <- function(x)
            if (x$l == x$r) x$l else paste(x$l, "..", x$r, sep = "")
        paste(sapply(unclass(x), .format), collapse = ", ")
    }
}

as.interval <-
function(x)
    UseMethod("as.interval")

as.interval.default <- function(x) stop("Not implemented.")
as.interval.numeric <-
function(x)
{
    if (all(is.finite(x)) && isTRUE(all.equal(as.integer(x), x)))
        return(do.call(c.interval, lapply(unclass(x),
                                          function(i) integers(i, i))))
    else
        reals(min(x, na.rm = TRUE), max(x, na.rm = TRUE))
}

as.interval.interval <- function(x) x
as.interval.set <- function(x) as.interval(unclass(x))
as.interval.list <- function(x) do.call(c.interval, x)
as.interval.tuple <- function(x) interval(x[[1]], x[[2]])
as.interval.logical <-
as.interval.NULL <- function(x) .empty_interval()

as.vector.interval <-
    function(x, ...)
        as.list.interval(x, ...)

as.list.interval <-
    function(x, ...)
        lapply(seq_along(x), function(i) x[[i]])

make_set_with_order.interval <-
function(x) {
    if (interval_domain(x) == "R") {
        FUN <- function(i)
            if (isTRUE(all.equal(i$l, i$r))) i$l else pair(i$l, i$r)
        .make_set_with_order(.make_set_from_list(lapply(unclass(x), FUN)))
    } else {
        .make_set_with_order(.make_set_from_list(unlist(lapply(unclass(x),
                                                               function(i) seq.int(i$l, i$r)))))
    }
}

as.double.interval <-
function(x, ...) {
    if (length.interval(x) < 1L)
        return(NULL)
    r <- range(x)
    x <- as.list(reals2integers(x))
    if (r[1] == -Inf) x <- x[-1]
    l <- length(x)
    if (l > 0L && r[2] == Inf) x <- x[-l]
    unlist(lapply(unclass(x), function(i) seq.int(i$l, i$r)))
}

### print and summary stuff

print.interval <-
function(x, ...)
{
    writeLines(as.character(x))
    invisible(x)
}

summary.interval <-
function(object, ...)
{
    len <- length.interval(object)
    uc <- interval_is_uncountable(object)
    out <- if (len < 1L)
        gettext("The empty interval.")
    else if (len < 2L) {
        if (uc)
            gettext("An interval.")
        else
            gettext("A sequence.")
    } else {
        if (uc)
            gettextf("A union of %d disjoint intervals.", len)
        else
            gettextf("A union of %d sequences.", len)
    }
    .structure(out, class = "summary.interval")
}

print.summary.interval <-
function(x, ...)
{
    writeLines(x)
    invisible(x)
}

format.interval <-
function(x, ...)
    as.character(x)

## extractors
interval_domain <-
function(x)
    attr(as.interval(x), "domain")

### predicates
interval_is_uncountable <-
function(x)
    interval_domain(x) == "R"

interval_is_countable <- Negate(interval_is_uncountable)

interval_is_bounded <-
interval_is_finite <-
function(x)
    interval_is_left_bounded(x) && interval_is_right_bounded(x)

interval_is_left_bounded <-
function(x)
    is.finite(min(x)) || interval_is_left_open(x)

interval_is_right_bounded <-
function(x)
    is.finite(max(x)) || interval_is_right_open(x)

interval_is_left_unbounded <-
function(x)
    interval_is_left_closed(x) && is.infinite(min(x))

interval_is_right_unbounded <-
function(x)
    interval_is_right_closed(x) && is.infinite(max(x))

interval_is_unbounded <-
function(x)
    interval_is_left_unbounded(x) && interval_is_right_unbounded(x)

interval_is_half_bounded <-
function(x)
    xor(interval_is_left_bounded(x), interval_is_right_bounded(x))

interval_is_degenerate <-
function(x)
    interval_is_empty(x) || isTRUE(all.equal(min(x), max(x)))

interval_is_proper <- Negate(interval_is_degenerate)

interval_is_empty <-
function(x)
    length.interval(x) < 1L

interval_is_closed <-
function(x)
    interval_is_left_closed(x) && interval_is_right_closed(x)

interval_is_left_closed <-
function(x)
    unclass(as.interval(x))[[1]]$lc

interval_is_right_closed <-
function(x) {
    x <- as.interval(x)
    i <- length.interval(x)
    unclass(x)[[i]]$rc
}

interval_is_left_open <- Negate(interval_is_left_closed)

interval_is_right_open <- Negate(interval_is_right_closed)

interval_is_equal <-
function(x, y)
{
    x <- as.interval(x)
    y <- as.interval(y)
    ctbls <- interval_is_countable(x) && interval_is_countable(y)
    uc <- interval_is_uncountable(x) && interval_is_uncountable(y)
    x <- unclass(x)
    y <- unclass(y)
    length(x) == length(y) && (ctbls || uc) &&
        isTRUE(all.equal(sapply(x, `[[`, "l"), sapply(y, `[[`, "l"))) &&
        isTRUE(all.equal(sapply(x, `[[`, "r"), sapply(y, `[[`, "r"))) &&
        (ctbls || all(sapply(x, `[[`, "lc") == sapply(y, `[[`, "lc")) &&
                  all(sapply(x, `[[`, "rc") == sapply(y, `[[`, "rc")))
}

"%<%" <-
interval_is_less_than <-
function(x, y)
{
    mx <- max(x)
    my <- min(y)
    mx < my ||
      mx == my && (interval_is_right_open(x) || interval_is_left_open(y))
}

"%>%" <-
interval_is_greater_than <-
function(x, y)
{
    mx <- min(x)
    my <- max(y)
    mx > my ||
      mx == my && (interval_is_right_open(x) || interval_is_left_open(y))
}

"%<=%" <-
interval_is_less_than_or_equal <-
function(x, y)
    max(x) <= min(y)

"%>=%" <-
interval_is_greater_than_or_equal <-
function(x, y)
    min(x) >= max(y)

interval_is_proper_subinterval <-
function(x, y)
    interval_is_subinterval(x, y) && !interval_is_equal(x, y)

interval_is_subinterval <-
function(x, y)
{
    x <- as.interval(x)
    y <- as.interval(y)

    ## trivial cases
    if (interval_is_empty(x))
        return(TRUE)

    if (interval_is_empty(y))
        return(FALSE)

    if (interval_is_uncountable(x) &&
        interval_is_proper(x) &&
        interval_is_countable(y))
        return(FALSE)

    ## FIXME: is Real(Inf) == Integer(Inf)?

    ## check if bounds are nested
    if (min(x) < min(y) || max(x) > max(y))
        return(FALSE)

    ## check each element in x against y
    uc <- interval_is_uncountable(x) || interval_is_uncountable(y)
    l <- length(y <- unclass(y))
    j <- 1
    .not_contained <-
        function(a, b) {
            a$l < b$l || a$r > b$r ||
            uc && (isTRUE(all.equal(a$l, b$l)) && a$lc && !b$lc ||
                   isTRUE(all.equal(a$r, b$r)) && a$rc && !b$rc)
        }
    for (i in x) {
        while (.not_contained(i, y[[j]])) {
            j <- j + 1
            if (j > l) return(FALSE)
        }
    }
    TRUE
}

interval_contains_element <-
function(x, y)
    interval_is_subinterval(y, x)

is.interval <- function(x) inherits(x, "interval")

## str method
str.interval <- function(object, ...) str(unclass(object), ...)

### internal functions

.empty_interval <-
function()
    .make_interval_from_list_and_domain(list(), NA)

.make_interval_from_list_and_domain <- function(x, domain = "R")
    .structure(x, class = "interval", domain = domain)

.sort_interval <-
function(x)
    x[order(sapply(unclass(x), `[[`, "l"), sapply(unclass(x), `[[`, "r"))]

.merge_uncountables <-
function(x, y)
{
    len <- length(x)
    if (len < 1L) return(y)
    if (length(y) < 1L) return(x)
    X <- x[[len]]
    Y <- y[[1L]]
    if (Y$l < X$r || isTRUE(all.equal(Y$l, X$r)) && (X$rc || Y$lc)) {
        ## X and Y overlap
        if (isTRUE(all.equal(Y$l, X$l)) && (X$lc != Y$lc))
            ## X left-open, but Y left-closed -> left-close X
            X$lc <- TRUE
        if (Y$r < X$r || isTRUE(all.equal(Y$r, X$r)) && (!Y$rc || X$rc)) {
            ## Y included in X -> discard Y
            c(x[-len], list(X))
        } else {
            ## true overlap -> merge X and Y
            Y$l <- X$l
            Y$lc <- X$lc
            c(x[-len], list(Y))
        }
    } else c(x, y) ## X and Y are disjoint -> return both
 }

.merge_countables <-
function(x, y)
{
    len <- length(x)
    if (len < 1L) return(y)
    if (length(y) < 1L) return(x)
    X <- x[[len]]
    Y <- y[[1L]]
    if (Y$l - X$r < 2L) {
        ## X and Y overlap
        if (Y$r <= X$r)
            ## Y included in X -> discard Y
            c(x[-len], list(X))
        else {
            ## true overlap -> merge X and Y
            Y$l <- X$l
            c(x[-len], list(Y))
        }
    } else c(x, y) ## X and Y are disjoint -> return both
}

.simplify_interval <-
function(x, reals = TRUE)
{
    Reduce(if (reals) .merge_uncountables else .merge_countables,
           lapply(unclass(x), list))
}

.canonicalize_interval <-
function(x, reals = TRUE)
    .simplify_interval(.sort_interval(x), reals = reals)

.minimal_domain <-
function(l)
{
    dom <- rep.int(sapply(l, interval_domain), sapply(l, length.interval))
    if (any(dom == "N"))
        "N"
    else if (any(dom == "N0"))
        "N0"
    else if (any(dom == "Z"))
        "Z"
    else
        "R"
}

.maximal_domain <-
function(l)
{
    dom <- rep.int(sapply(l, interval_domain), sapply(l, length.interval))
    if (any(dom == "R"))
        "R"
    else if (any(dom == "Z"))
        "Z"
    else if (any(dom == "N0"))
        "N0"
    else
        "N"
}

## extractors

gset_memberships <-
function(x, filter = NULL)
{
    ex <- if(is.null(filter))
        .get_memberships
    else
        function(i) .get_memberships(i[filter])
    if (is.tuple(x)) lapply(x, ex) else ex(x)
}

gset_support <-
function(x)
    as.set(.get_support(x))

gset_core <-
function(x, na.rm = FALSE)
{
    m <- as.double(sapply(.get_memberships(x), max) >= 1)
    if (na.rm) m[is.na(m)] <- FALSE
    .make_gset_from_support_and_memberships(.get_support(x), m)
}

gset_height <-
function(x, na.rm = FALSE)
    max(unlist(.get_memberships(x)), na.rm = na.rm)

gset_peak <-
function(x, na.rm = FALSE)
{
    m <- as.double(sapply(.get_memberships(x), max) ==
                   gset_height(x, na.rm = TRUE))
    if (na.rm) m[is.na(m)] <- FALSE
    .make_gset_from_support_and_memberships(.get_support(x), m)
}

gset_charfun <-
function(x)
{
    if (!is.gset(x))
        stop("Argument 'x' must be a generalized set.")
    ret <- function(e) {
        if (missing(e)) return(.get_memberships(x))
        if (is_element(e)) e <- list(e)
        ret <- .memberships_for_support(x, e)
        if (is.list(ret) && (length(ret) < 2L))
            ret[[1L]]
        else
            ret
    }
    .structure(ret, class = "gset_charfun")
}

gset_universe <-
function(x)
    .get_universe(x)

gset_bound <-
function(x)
    .get_bound(x)

print.gset_charfun <-
function(x, ...)
{
    writeLines("The characteristic function of a generalized set.")
    invisible(x)
}

### * cset extractors (mostly copies of the gset extractors)

cset_memberships <- gset_memberships

cset_support <- gset_support

cset_core <- gset_core

cset_height <- gset_height

cset_peak <- gset_peak

cset_charfun <-
function(x)
{
    if (!is.cset(x))
        stop("Argument 'x' must be a customizable set.")
    matchfun <- cset_matchfun(x)
    ret <- function(e) {
        if (missing(e)) return(.get_memberships(x))
        if (is_element(e)) e <- list(e)
        ret <- .memberships_for_support(x, e, matchfun)
        if (is.list(ret) && (length(ret) < 2L))
            ret[[1L]]
        else
            ret
    }
    .structure(ret, class = "cset_charfun")
}

cset_universe <- gset_universe

cset_bound <- gset_bound

print.cset_charfun <-
function(x, ...)
{
    writeLines("The characteristic function of a customizable set.")
    invisible(x)
}

cset_orderfun <-
function(x)
{
    FUN <- .orderfun(x)
    if (is.null(FUN))
        .list_order
    else
        FUN
}

"cset_orderfun<-" <-
function(x, value) {
    attr(x, "orderfun") <- value
    x
}

cset_matchfun <-
function(x)
{
    FUN <- .matchfun(x)
    if (is.null(FUN))
        .exact_match
    else
        FUN
}

"cset_matchfun<-" <-
function(x, value)
    cset(x, matchfun = value, orderfun = .orderfun(x))

## internal stuff

.get_memberships <-
function(x)
{
    m <- attr(x, "memberships")
    if (is.null(m)) rep.int(1L, length.set(x)) else m
}

.set_memberships <-
function(x, value)
{
    attr(x, "memberships") <- value
    x
}

.get_fuzzy_multi_memberships <-
function(x)
{
    if (!gset_is_fuzzy_multiset(x))
        lapply(.get_memberships(x), function(i) gset(1L, i))
    else
        .get_memberships(x)
}

.get_universe <-
function(x)
{
    u <- attr(x, "universe")
    if (is.null(u))
        u <- sets_options("universe")
    if (!is.null(u))
        as.set(if (is.interval(u)) as.numeric(u) else eval(u))
    else
        as.set(.get_support(x))
}

.set_universe <-
function(x, value)
{
    attr(x, "universe") <- value
    x
}

.get_bound <-
function(x)
{
    b <- attr(x, "bound")
    if (is.null(b))
        b <- sets_options("bound")
    if (!is.null(b))
        b
    else if (set_is_empty(x))
        return(0)
    else
        max(.multiplicities(.get_memberships(x)), na.rm = TRUE)
}

.set_bound <-
function(x, value)
{
    attr(x, "bound") <- value
    x
}

.get_support <-
function(x)
{
    ## simplify, if all elements are of same class and of length 1:
    cl <- unlist(lapply(x, class))
    len <- lapply(x, length)

    ret <- if (all(cl[1] == cl) && all(len == 1L))
        unlist(x, recursive = FALSE)
    else
        x

    if (is.null(ret))
        list()
    else if (is.recursive(ret))
        .as.list(x)
    else
        ret
}

.multiplicities <-
function(x)
{
    if (is.list(x))
        unlist(lapply(x, function(i) sum(.get_memberships(i))))
    else if(any(x > 1, na.rm = TRUE))
        x
    else
        rep.int(1L, length(x))
}

.orderfun <-
function(x)
    attr(x, "orderfun")

.matchfun <-
function(x)
    attr(x, "matchfun")

## set Summary methods
Summary.set <-
function(..., na.rm = FALSE)
{
    l <- list(...)
    if (.Generic == "sum")
        return(Reduce(function(i, j) sum(i, as.numeric(j), na.rm = na.rm),
                      l, 0))
    else if (.Generic == "prod")
        return(Reduce(function(i, j) prod(i, as.numeric(j), na.rm = na.rm),
                      l, 1))
    do.call(.Generic, c(na.rm = na.rm, do.call(set_union, l)))
}

mean.set <-
function(x, ...)
{
    mean(as.numeric(x), ...)
}

## <FIXME>
## Simplify when depending on R >= 3.4.0
if(is.na(match("...", names(formals(median))))) {
    median.set <- function(x, na.rm = FALSE) {
        median(as.numeric(x), na.rm = na.rm)
    }
} else {
    median.set <- function(x, na.rm = FALSE, ...) {
        median(as.numeric(x), na.rm = na.rm, ...)
    }
}
## </FIXME>

## gset Summary methods

Summary.gset <-
function(..., na.rm = FALSE)
{
    l <- list(...)
    if (any(sapply(l, gset_is_fuzzy_multiset)))
        stop("Operation not defined for fuzzy multisets.")
    l <- lapply(l, function(i) as.set(as.numeric(i) * .get_memberships(i)))
    do.call(.Generic, c(l, na.rm = na.rm))
}

mean.gset <-
function(x, ..., na.rm = FALSE)
{
    if (gset_is_fuzzy_multiset(x))
        stop("Operation not defined for fuzzy multisets.")

    v <- as.numeric(x)
    m <- .get_memberships(x)
    if (na.rm && any(nas <- is.na(m))) {
        v <- v[!nas]
        m <- m[!nas]
    }

    weighted.mean(v, m, na.rm = na.rm)
}

## <FIXME>
## Simplify when depending on R >= 3.4.0
if(is.na(match("...", names(formals(median))))) {
    median.gset <- function(x, na.rm = FALSE) {
        if (gset_is_fuzzy_multiset(x))
            stop("Operation not defined for fuzzy multisets.")
        x <- if (gset_is_fuzzy_set(x, na.rm = TRUE))
                 as.numeric(x) * .get_memberships(x)
             else {
                 n <- as.numeric(x)
                 m <- .get_memberships(x)
                 n[is.na(m)] <- NA
                 m[is.na(m)] <- 1
                 rep.int(n, times = m)
             }
        median(x, na.rm = na.rm)
    }
} else {
    median.gset <- function(x, na.rm = FALSE, ...) {
        if (gset_is_fuzzy_multiset(x))
            stop("Operation not defined for fuzzy multisets.")
        x <- if (gset_is_fuzzy_set(x, na.rm = TRUE))
                 as.numeric(x) * .get_memberships(x)
             else {
                 n <- as.numeric(x)
                 m <- .get_memberships(x)
                 n[is.na(m)] <- NA
                 m[is.na(m)] <- 1
                 rep.int(n, times = m)
             }
        median(x, na.rm = na.rm, ...)
    }
}
## </FIXME>

## cset Summary methods
## FIXME: can we call Summary.gset directly?

Summary.cset <-
function(..., na.rm = FALSE)
{
    l <- list(...)
    if (any(sapply(l, gset_is_fuzzy_multiset)))
        stop("Operation not defined for fuzzy multisets.")
    l <- lapply(l, function(i) as.set(as.numeric(i) * .get_memberships(i)))
    do.call(.Generic, c(l, na.rm = na.rm))
}

mean.cset <-
function(x, ..., na.rm = FALSE)
    mean.gset(x, ..., na.rm = na.rm)

## <FIXME>
## Simplify when depending on R >= 3.4.0
if(is.na(match("...", names(formals(median))))) {
    median.cset <- function(x, na.rm = FALSE) 
        median.gset(x, na.rm = na.rm)
} else {
    median.cset <- function(x, na.rm = FALSE, ...) 
        median.gset(x, na.rm = na.rm, ...)
}
## </FIXME>

sets_options <-
local({
    options <- list(quote = TRUE, hash = TRUE, openbounds = "()")
    function(option, value) {
        if (missing(option)) return(options)
        if (missing(value))
            options[[option]]
        else
            options[[option]] <<- value
    }
})

set_similarity <-
function(x, y, method = "Jaccard")
{
    method <- match.arg(method)

    if (method == "Jaccard") {
        if (set_is_empty(x) && set_is_empty(y))
            1
        else
            length(set_intersection(x, y)) / length(set_union(x, y))
    }
}

gset_similarity <-
function(x, y, method = "Jaccard")
{
    method <- match.arg(method)

    if (method == "Jaccard") {
        if (gset_is_empty(x) && gset_is_empty(y))
            1
        else
            gset_cardinality(gset_intersection(x, y)) / gset_cardinality(gset_union(x, y))
    }
}

cset_similarity <-
function(x, y, method = "Jaccard")
{
    method <- match.arg(method)

    if (method == "Jaccard") {
        if (cset_is_empty(x) && cset_is_empty(y))
            1
        else
            cset_cardinality(cset_intersection(x, y)) / cset_cardinality(cset_union(x, y))
    }
}

set_dissimilarity <-
function(x, y, method = c("Jaccard", "Manhattan", "Euclidean", "L1", "L2"))
{
    method <- match.arg(method)

    switch(method,
           Jaccard = 1 - set_similarity(x, y, "Jaccard"),
           Manhattan = , L1 = sqrt(length(set_symdiff(x, y))),
           Euclidean = , L2 = length(set_symdiff(x, y))
           )
}

gset_dissimilarity <-
function(x, y, method = c("Jaccard", "Manhattan", "Euclidean", "L1", "L2"))
{
    method <- match.arg(method)

    switch(method,
           Jaccard = 1 - gset_similarity(x, y, "Jaccard"),
           Manhattan =, L1 = .distfunL1(x, y),
           Euclidean =, L2 = .distfunL2(x, y)
           )
}

cset_dissimilarity <-
function(x, y, method = c("Jaccard", "Manhattan", "Euclidean", "L1", "L2"))
{
    method <- match.arg(method)

    switch(method,
           Jaccard = 1 - cset_similarity(x, y, "Jaccard"),
           Manhattan =, L1 = .distfunL1(x, y),
           Euclidean =, L2 = .distfunL2(x, y)
           )
}

### internal

.distfunL2 <- function(x, y) {
    if (cset_is_multiset(x, na.rm = TRUE))
        sqrt(.distfunL1(x, y))
    else
        sqrt(sum(unlist(.apply_connector_to_list_of_gsets_and_support(list(x, y),
             set_union(x, y), function(i,j) crossprod(i - j)[,,drop = T]))))
}

.distfunL1 <- function(x, y)
    sum(unlist(.apply_connector_to_list_of_gsets_and_support(list(x, y),
                    set_union(x, y), function(i,j) sum(abs(i - j)))))

### LABELS
###
## We need a function that produces "nice" labels from any object mainly
## for printing, but also for dimnames.  This function should do more
## than as.character(), and less than format ...
##
## Normally, one uses LABELS() and adds extensions by writing methods
## for LABEL().
##
## What we do in LABELS() is the following:
##
## 1. transform the given object to a list
## 2. check names attribute; if any, use these as default
## 3. for all components with empty name, use LABEL() to compute a
##    "simple" representation
## 4. optionally, truncate strings to specified length
## 5. optionally, apply make.unique() to the result
##
## Generally, LABEL() uses format() if the argument is of length 1, and
## creates a type specification otherwise.
## Exception: we also accept "small" sets and pairs since they can well
## be distinguished even if they are nested.  Currently, "small" means a
## length of 5 which is sort of ad-hoc.

LABELS <-
function(x, max_width = NULL, dots = "...", unique = FALSE, limit = NULL, ...)
{
    x <- as.list(x)
    l <- length(x)

    ## recycle max_width and dots as needed
    if (!is.null(max_width))
        max_width <- rep_len(max_width, length.out = l)
    dots <- rep_len(dots, length.out = l)

    ## check existing labels
    ret <- names(x)
    if (is.null(ret))
        ret <- rep.int("", l)

    ## create a label for components without given one
    empty <- is.na(ret) | (ret == "")
    if (any(empty))
        ret[empty] <- sapply(x[empty], LABEL, limit, ...)

    ## check maximum width (max_width == NULL => unbounded)
    if (!is.null(max_width)) {
        too_long <- nchar(ret, "width") > max_width
        if (any(too_long)) {
            ret[too_long] <- strtrim(ret[too_long], max_width[too_long])

            ## possibly add dots
            if (!is.null(dots))
                ret[too_long] <- paste0(ret[too_long], dots[too_long])
          }
    }

    if (unique)
      ret <- make.unique(ret)

    ret
}

LABEL <-
function(x, limit = NULL, ...)
    UseMethod("LABEL")

LABEL.default <-
function(x, limit = NULL, ...)
    paste0("<<", class(x)[1L], ">>")

LABEL.matrix <-
function(x, limit = NULL, ...)
    sprintf("<<%ix%i matrix>>", nrow(x), ncol(x))

LABEL.numeric <-
LABEL.factor <-
LABEL.integer <-
LABEL.logical <-
function(x, limit = NULL, ...) {
    if (is.null(limit))
        limit <- 2L
    .format_or_class(x, limit, ...)
}

LABEL.character <-
function(x, limit = NULL, quote = sets_options("quote"), ...) {
    if (is.null(limit))
        limit <- 2L
    if (quote)
        x <- ifelse(is.na(x), x, paste0("\"", x, "\""))
    .format_or_class(x, limit, ...)
}

LABEL.list <-
function(x, limit = NULL, ...) {
    if (is.null(limit))
        limit <- 1L
    .format_or_class(x, limit, ...)
}

LABEL.set <-
LABEL.gset <-
LABEL.cset <-
LABEL.tuple <-
LABEL.interval <-
function(x, limit = NULL, ...) {
    if (is.null(limit))
        limit <- 6L
    .format_or_class(x, limit, ...)
}

.format_or_class <-
function(x, limit, ...)
{
    l <- length.set(x)
    if (l < limit) {
        if (is.integer(x))
            format(ifelse(is.na(x), x, paste0(x, "L")), ...)
        else
            format(x, ...)
    } else
        paste0("<<", class(x)[1L], "(", l, ")>>")
}

c.set <-
set_union <-
function(...)
{
    ret <- do.call(.set,
                   unlist(lapply(list(...), .make_list_elements), recursive = FALSE))
    n <- names(ret)
    if (!is.null(n) && any(duplicated(n[n != ""])))
        names(ret) <- NULL
    ret
}

c.gset <-
gset_union <-
function(...)
{
    l <- list(...)

    ## compute target support
    support <- do.call(set_union, l)

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, gset_is_set))))
        return(support)

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, .S.)
}

c.cset <-
cset_union <-
function(...)
{
    l <- list(...)

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## compute target support using correct matchfuns
    support <- cset(do.call(set_union, l),
                    matchfun = matchfun,
                    orderfun = orderfun)

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, gset_is_set))))
        return(support)

    ## create gset by applying conorm, and then make cset
    .make_cset_from_gset_and_orderfun_and_matchfun(
         .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                 support,
                                                                 .S.,
                                                                 matchfun),
         orderfun,
         matchfun)
}

reduction <-
function(x, operation, ...)
    UseMethod("reduction")

reduction.default <-
function(x, operation, ...)
    stop("Not implemented.")

reduction.set <-
function(x, operation = c("union", "intersection"), ...)
{
    operation <- match.arg(operation)
    if (length(x) < 2L) return(x)
    if (!all(sapply(x, is.cset)))
        stop("reduction only defined on set of (c,g)sets.")

    if (all(sapply(x, is.set))) {

        dom <- .as.list(do.call(set_union, x))
        x <- lapply(x, .make_list_elements)

        members <-
            binary_reduction(do.call(rbind, lapply(x, function(i) dom %in% i)),
                             operation)

        .make_set_from_list(.list_sort(apply(members, 1L,
                                             function(i) .make_set_from_list(dom[i])
                                             )
                                       )
                            )
    } else if (all(sapply(x, is.gset))) {
        clo <- closure(x, operation)
        FUN <- function(e)
            !isTRUE(gset_is_equal(closure(gset_difference(x, set(e)),
                                          operation), clo))
        as.set(Filter(FUN, .as.list(x)))
    } else {
        clo <- closure(x, operation)
        FUN <- function(e)
            !isTRUE(cset_is_equal(closure(cset_difference(x, set(e)),
                                          operation), clo))
        as.set(Filter(FUN, .as.list(x)))
    }
}

binary_reduction <-
function(x, operation = c("union", "intersection"))
    .Call(sets_reduction, x,
          pmatch(match.arg(operation), c("union", "intersection")))

## Support for fuzzy sets needs the specification of negation,
## conjunction and disjunction via the functions N, T (t-norm) and S
## (t-conorm), respectively.
##
## We try to make our code "work" with arbitrary such triples by using
## functions .N., .T. and .S., respectively, and allow setting the fuzzy
## logic system via a dynamic variable.  If this turns out to be too
## cumbersome, we could also move back to hard-wiring the "usual" Zadeh
## connectives 1 minus, min and max, and provide ways to unlock the
## locked bindings as needed.

## .N. <- function(x) 1 - x
## .T. <- function(x, y) pmin(x, y)
## .S. <- function(x, y) pmax(x, y)
## .I. <- function(x, y) ifelse(x <= y, 1, y)

.N. <- function(x) fuzzy_logic()$N(x)

## ensure that .T.(NA, 0) == 0
.T. <- function(x, y)
    `[<-`(fuzzy_logic()$T(x, y),
          xor(is.na(y), is.na(x)) & pmin.int(x, y, na.rm = TRUE) == 0,
          0)

## ensure that .S.(NA, 1) == 1
.S. <- function(x, y)
    `[<-`(fuzzy_logic()$S(x, y),
          xor(is.na(y), is.na(x)) & pmax.int(x, y, na.rm = TRUE) == 1,
          1)

## ensure that .I.(0, NA) == 1
.I. <- function(x, y)
    `[<-`(fuzzy_logic()$I(x, y),
          is.na(y) & !is.na(x) & x == 0,
          1)

## Use dynamic variables for the fuzzy connectives.
## Note that there are also parametric fuzzy logic families, which we
## can set via
##   fuzzy_logic(NAME, PARAMS)

## One might think that using an environment for storing the fuzzy logic
## connectives, along the lines of
##   fuzzy_logic_db <- new.env()
## and adding
##   for(nm in names(family))
##     assign(nm, family[[nm]], envir = fuzzy_logic_db)
## and using
##   .N. <- function(...) fuzzy_logic_db$N(...)
## would be somewhat faster than
##   .N. <- function(...) fuzzy_logic()$N(...)
## but we could not find conclusive evidence for performance gains
## either way.

fuzzy_logic <-
local({
    family <- NULL
    function(new, ...) {
        if(!missing(new))
            family <<- fuzzy_logic_families[[new]](...)
        else
            family
    }
})

fuzzy_logic_family <-
function(name, T, S, N, I = NULL, params = NULL, meta = NULL)
{
    if(is.null(I))
        I <- function(x, y)
            stop("Implication not available.", call. = FALSE)
    .structure(list(name = name, T = T, S = S, N = N, I = I,
                    params = params, meta = meta),
               class = "fuzzy_logic_family")
}

fuzzy_logic_predicates <-
function()
    fuzzy_logic()$meta

fuzzy_logic_family_Zadeh <-
function()
    fuzzy_logic_family(name = "Zadeh",
                       N = function(x) 1 - x,
                       T = function(x, y) pmin(x, y),
                       S = function(x, y) pmax(x, y),
                       I = function(x, y) ifelse(x <= y, 1, y),
                       meta =
                       list(is_de_Morgan_triple = TRUE,
                            N_is_standard = TRUE,
                            T_is_continuous = TRUE,
                            T_is_Archimedean = FALSE)
                       )

fuzzy_logic_family_drastic <-
function()
    fuzzy_logic_family(name = "drastic",
                       N = function(x) 1 - x,
                       T = function(x, y)
                       ifelse(pmax(x, y) == 1, pmin(x, y), 0),
                       S = function(x, y)
                       ifelse(pmin(x, y) == 0, pmax(x, y), 1),
                       meta =
                       list(is_de_Morgan_triple = TRUE,
                            N_is_standard = TRUE,
                            T_is_continuous = FALSE,
                            T_is_Archimedean = TRUE,
                            T_generator = function(x) {
                                ifelse(x < 1, 2 - x, 0)
                            }
                            )
                       )

fuzzy_logic_family_product <-
function()
    fuzzy_logic_family(name = "product",
                       N = function(x) 1 - x,
                       T = function(x, y) x * y,
                       S = function(x, y) x + y - x * y,
                       I = function(x, y) pmin(y / x, 1),
                       meta =
                       list(is_de_Morgan_triple = TRUE,
                            N_is_standard = TRUE,
                            T_is_continuous = TRUE,
                            T_is_Archimedean = TRUE,
                            T_generator = function(x) { - log(x) }
                            )
                       )

fuzzy_logic_family_Lukasiewicz <-
function()
    fuzzy_logic_family(name = "Lukasiewicz",
                       N = function(x) 1 - x,
                       T = function(x, y) pmax(x + y - 1, 0),
                       S = function(x, y) pmin(x + y, 1),
                       I = function(x, y) pmin(1 - x + y, 1),
                       meta =
                       list(is_de_Morgan_triple = TRUE,
                            N_is_standard = TRUE,
                            T_is_continuous = TRUE,
                            T_is_Archimedean = TRUE,
                            T_generator = function(x) { 1 - x }
                            )
                       )

## Nilpotent minimum (\min_0).
fuzzy_logic_family_Fodor <-
function()
    fuzzy_logic_family(name = "Fodor",
                       N = function(x) 1 - x,
                       T = function(x, y)
                       ifelse(x + y > 1, pmin(x, y), 0),
                       S = function(x, y)
                       ifelse(x + y < 1, pmax(x, y), 1),
                       I = function(x, y)
                       ifelse(x <= y, 1, pmax(1 - x, y)),
                       meta =
                       list(is_de_Morgan_triple = TRUE,
                            N_is_standard = TRUE,
                            T_is_continuous = FALSE,
                            T_is_Archimedean = FALSE
                            )
                       )

## Frank family (Fodor & Roubens, page 20).
## One parameter family with 0 <= s <= \infty, where 0, 1 and \infty
## give Zadeh, product and Lukasiewicz, respectively.
fuzzy_logic_family_Frank <-
function(s)
{
    if(s < 0)
        stop("Invalid parameter.")
    if(s == 0) fuzzy_logic_family_Zadeh()
    else if(s == 1) fuzzy_logic_family_product()
    else if(s == Inf) fuzzy_logic_family_Lukasiewicz()
    else {
        T <- function(x, y)
            log(1 + (s^x - 1) * (s^y - 1) / (s - 1)) / log(s)
        fuzzy_logic_family(name = "Frank",
                           N = function(x) 1 - x,
                           T = T,
                           S = function(x, y) 1 - T(1 - x, 1 - y),
                           I = function(x, y)
                           ifelse(x <= y, 1,
                                  log(1 + (s - 1) * (s^y - 1) /
                                                    (s^x - 1)) / log(s)),
                           params = list(s = s),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    log((s - 1) / (s^x - 1))
                                }
                                ## Note:
                                ##   f^{-1}(x) = \log_s(1 + (s-1)e^{-x})
                                )
                           )
    }
}

## Hamacher family (Fodor & Roubens, page 21).
## This is a three parameter family of connectives T_\alpha, S_\beta,
## N_\gamma with \alpha >= 0, \beta, \gamma >= -1, which is a de Morgan
## triple iff \gamma > -1 and \alpha = (1 + \beta) / (1 + \gamma).
## Be nice and provide some default values ...
fuzzy_logic_family_Hamacher <-
function(alpha = NULL, beta = 0, gamma = 0)
{
    if(is.null(alpha))
        alpha <- (1 + beta) / (1 + gamma)
    else if((alpha < 0) || (beta < -1) || (gamma < -1))
        stop("Invalid parameter.")
    fuzzy_logic_family(name = "Hamacher",
                       N = function(x) (1 - x) / (1 + gamma * x),
                       T = function(x, y)
                       ifelse(x * y == 0,
                              0,
                              x * y /
                              (alpha + (1 - alpha) * (x + y - x * y))),
                       S = function(x, y)
                       (x + y + (beta - 1) * x * y) /
                        (1 + beta * x * y),
                       I = function(x, y)
                       ifelse(x <= y, 1,
                              ((y * (alpha + (1 - alpha) * x)) /
                               (x + (1 - alpha) * y * (x - 1)))),
                       params =
                       list(alpha = alpha, beta = beta, gamma = gamma),
                       meta =
                       list(is_de_Morgan_triple =
                            alpha == (1 + beta) / (1 + gamma),
                            N_is_standard = (gamma == 0),
                            T_is_continuous = TRUE,
                            T_is_Archimedean = TRUE,
                            T_generator = function(x) {
                                log((alpha + (1 - alpha) * x) / x)
                            }
                            ## Note:
                            ##   f^{-1}(x) = p / (e^x + p - 1)
                            )
                       )
}

## The following "families" are really families of t-norms, which we
## leverage into fuzzy logic families using standard negation and de
## Morgan triplet associated t-conorm.

## Schweizer-Sklar.
fuzzy_logic_family_Schweizer_Sklar <-
function(p)
{
    if(p == -Inf) fuzzy_logic_family_Zadeh()
    else if(p == 0) fuzzy_logic_family_product()
    else if(p == Inf) fuzzy_logic_family_drastic()
    else {
        T <- if(p < 0)
            function(x, y) (x^p + y^p - 1) ^ (1/p)
        else
            function(x, y) pmax(0, (x^p + y^p - 1) ^ (1/p))
        fuzzy_logic_family(name = "Schweizer-Sklar",
                           N = function(x) 1 - x,
                           T = T,
                           S = function(x, y) 1 - T(1 - x, 1 - y),
                           I = function(x, y)
                           ifelse(x <= y, 1, (1 - x^p + y^p) ^ (1/p)),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    (1 - x^p) / p
                                }
                                ## Note:
                                ##   f^{-1}(x) = (1 - px)^{1/p}
                                )
                           )
    }
}

## Yager t-norm.
fuzzy_logic_family_Yager <-
function(p)
{
    if(p < 0)
        stop("Invalid parameter.")
    if(p == 0) fuzzy_logic_family_drastic()
    else if(p == Inf) fuzzy_logic_family_Zadeh()
    else {
        fuzzy_logic_family(name = "Yager",
                           N = function(x) 1 - x,
                           T = function(x, y)
                           pmax(0, 1 - ((1 - x)^p + (1 - y)^p)^(1/p)),
                           S = function(x, y)
                           pmin(1, (x^p + y^p) ^ (1/p)),
                           I = function(x, y)
                           ifelse(x <= y, 1,
                                  1 - ((1 - y)^p - (1 - x)^p)^(1/p)),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    1 - x^p
                                }
                                ## Note:
                                ##   f^{-1}(x) = 1 - x^{1/p}
                                )
                           )
    }
}

## Dombi t-norm.
fuzzy_logic_family_Dombi <-
function(p)
{
    if(p < 0)
        stop("Invalid parameter.")
    if(p == 0) fuzzy_logic_family_drastic()
    else if(p == Inf) fuzzy_logic_family_Zadeh()
    else {
        T <- function(x, y)
            ifelse(x * y == 0,
                   0,
                   1 / (1 + ((1 / x - 1)^p + (1 / y - 1)^p)^(1 / p)))
        fuzzy_logic_family(name = "Dombi",
                           N = function(x) 1 - x,
                           T = T,
                           S = function(x, y) 1 - T(1 - x, 1 - y),
                           I = function(x, y)
                           ifelse(x <= y, 1,
                                  1 / (1 + ((1 / y - 1)^p -
                                            (1 / x - 1)^p)^(1/p))),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    (1 / x - 1)^p
                                }
                                ## Note:
                                ##   f^{-1}(x) = 1/(1 + x^{1/p})
                                )
                           )
    }
}

## Aczel-Alsina t-norm.
fuzzy_logic_family_Aczel_Alsina <-
function(p)
{
    if(p < 0)
        stop("Invalid parameter.")
    if(p == 0) fuzzy_logic_family_drastic()
    else if(p == Inf) fuzzy_logic_family_Zadeh()
    else {
        T <- function(x, y) exp(- (abs(log(x))^p + abs(log(y))^p))
        fuzzy_logic_family(name = "Aczel-Alsina",
                           N = function(x) 1 - x,
                           T = T,
                           S = function(x, y) 1 - T(1 - x, 1 - y),
                           I = function(x, y)
                           ifelse(x <= y, 1,
                                  exp(-((abs(log(y))^p -
                                         abs(log(x))^p))^(1/p))),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    (- log(x))^p
                                }
                                ## Note:
                                ##   f^{-1}(x) = \exp(-x^{1/p})
                                )
                           )
    }
}

## Sugeno-Weber t-norm.
fuzzy_logic_family_Sugeno_Weber <-
function(p)
{
    if(p < -1)
        stop("Invalid parameter.")
    if(p == -1) fuzzy_logic_family_drastic()
    else if(p == Inf) fuzzy_logic_family_product()
    else
        fuzzy_logic_family(name = "Sugeno-Weber",
                           N = function(x) 1 - x,
                           T = function(x, y)
                           pmax(0, (x + y - 1 + p * x * y) / (1 + p)),
                           S = function(x, y)
                           pmin(1, x + y - p * x * y / (1 + p)),
                           I = function(x, y)
                           ifelse(x <= y, 1,
                                  (1 + (1 + p) * y - x) / (1 + p * x)),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE,
                                T_is_continuous = TRUE,
                                T_is_Archimedean = TRUE,
                                T_generator = function(x) {
                                    1 - log(1 + p * x) / log(1 + p)
                                }
                                ## Note
                                ##   f^{-1}(x) = ((1 + p)^{1-x} - 1) / p
                                )
                           )
}

## Dubois-Prade t-norm and t-conorm.

fuzzy_logic_family_Dubois_Prade <-
function(p)
{
    if((p < 0) || (p > 1))
        stop("Invalid parameter.")
    if(p == 0) fuzzy_logic_family_Zadeh()
    else if(p == 1) fuzzy_logic_family_product()
    else {
        T <- function(x, y) x * y / pmax(x, y, p)
        fuzzy_logic_family(name = "Dubois-Prade",
                           N = function(x) 1 - x,
                           T = T,
                           S = function(x, y) 1 - T(1 - x, 1 - y),
                           ## Solve T(x, z) = y for x >= y ...
                           I = function(x, y)
                           ifelse(x <= y, 1, pmax(p / x, 1) * y),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE
                                )
                           )
    }
}

## Yu t-norm and t-conorm.

fuzzy_logic_family_Yu <-
function(p)
{
    if(p < -1)
        stop("Invalid parameter.")
    if(p == -1) fuzzy_logic_family_product()
    else if(p == Inf) fuzzy_logic_family_drastic()
    else {
        fuzzy_logic_family(name = "Yu",
                           N = function(x) 1 - x,
                           T = function(x, y)
                           pmax(0, (1 + p) * (x + y - 1) - p * x * y),
                           S = function(x, y)
                           pmin(1, x + y + p * x * y),
                           params = list(p = p),
                           meta =
                           list(is_de_Morgan_triple = TRUE,
                                N_is_standard = TRUE
                                )
                           )
    }
}


fuzzy_logic_families <-
    list("Zadeh" = fuzzy_logic_family_Zadeh,
         "drastic" = fuzzy_logic_family_drastic,
         "product" = fuzzy_logic_family_product,
         "Lukasiewicz" = fuzzy_logic_family_Lukasiewicz,
         "Fodor" = fuzzy_logic_family_Fodor,
         "Frank" = fuzzy_logic_family_Frank,
         "Hamacher" = fuzzy_logic_family_Hamacher,
         "Schweizer-Sklar" = fuzzy_logic_family_Schweizer_Sklar,
         "Yager" = fuzzy_logic_family_Yager,
         "Dombi" = fuzzy_logic_family_Dombi,
         "Aczel-Alsina" = fuzzy_logic_family_Aczel_Alsina,
         "Sugeno-Weber" = fuzzy_logic_family_Sugeno_Weber,
         "Dubois-Prade" = fuzzy_logic_family_Dubois_Prade,
         "Yu" = fuzzy_logic_family_Yu
         )

## See also e.g. http://en.wikipedia.org/wiki/Construction_of_t-norms
## for more information.

## This has the generators for most Archimedean t-norms, but typically
## not the residual implications: so we use the result that
##   I(x, y) = f^{(-1)}(\max(f(y) - f(x), 0))
## where f is the (additive) generator and f^{(-1)} its pseudoinverse,
## defined as
##   f^{(-1)}(x) = f^{-1}(x) if x \le f(0) and 0 otherwise.
## For the implication, note that x \le y implies f(x) \ge f(y) and
## hence I(x, y) = f^{(-1)}(0) = 1; otherwise,
##   I(x, y) = f^{(-1)}(f(y) - f(x)), x \ge y.

### fuzzy characteristic functions

### * general constructor

charfun_generator <-
function(FUN, height = 1)
{
    ret <- if (is.null(height)) {
        function(...)
            function(x) pmax(0, pmin(FUN(x, ...), 1))
    } else {
        if (height < 0 || height > 1)
            stop("Height must be in the unit interval.")
        function(...)
            function(x) {
                ret <- pmax(0, pmin(FUN(x, ...), 1))
                height * ret / max(ret)
            }
    }
    class(ret) <- "charfun_generator"
    ret
}

is.charfun_generator <-
function(x)
    inherits(x, "charfun_generator")

### * special functions

fuzzy_normal <-
function(mean = NULL, sd = 1, log = FALSE, height = 1, chop = 0)
{
    if (!is.null(height) && (height < 0 || height > 1))
        stop("Height must be in the unit interval.")
    function(x) {
        if (is.null(mean))
            mean <- mean(range(x))
        ret <- dnorm(x, mean = mean, sd = sd, log = log)
        if (!is.null(height))
            ret <- height * ret / max(ret)
        ret[ret <= chop] <- 0
        ret
    }
}
class(fuzzy_normal) <- "charfun_generator"

fuzzy_two_normals <-
function(mean = NULL, sd = c(1,1),
         log = c(FALSE, FALSE), height = 1, chop = 0)
{
    if (!is.null(mean)) {
        if(length(mean) != 2L) stop("Need two mean values.")
        if (mean[2] < mean[1]) {
            sd <- rev(sd)
            log <- rev(log)
        }
    }
    if (!is.null(height) && (height < 0 || height > 1))
        stop("Height must be in the unit interval.")

    sd <- rep_len(sd, length.out = 2)
    log <- rep_len(log, length.out = 2)

    function(x) {
        if (is.null(mean))
            mean <- x[trunc(seq.int(1, length(x), length.out = 4))[2:3]]
        ret <- rep_len(height, length.out = length(x))

        tmp <- dnorm(x[x <= mean[1]], mean = mean[1], sd = sd[1], log = log[1])
        if (!is.null(height))
            tmp <- height * tmp / max(tmp)
        ret[x <= mean[1]] <- tmp

        tmp <- dnorm(x[x >= mean[2]], mean = mean[2], sd = sd[2], log = log[2])
        if (!is.null(height))
            tmp <- height * tmp / max(tmp)
        ret[x >= mean[2]] <- tmp

        ret[ret <= chop] <- 0
        ret
    }
}
class(fuzzy_two_normals) <- "charfun_generator"

fuzzy_bell <-
function(center = NULL, cross = NULL, slope = 4, height = 1, chop = 0)
{
    if (!is.null(height) && (height < 0 || height > 1))
        stop("Height must be in the unit interval.")
    function(x) {
        if (is.null(center))
            center <- mean(range(x))
        if (is.null(cross))
            cross <- trunc(diff(range(x)) / 5)
        ret <- 1 / (1 + abs((x - center) / cross) ^ (2 * slope))
        if (!is.null(height))
            ret <- height * ret / max(ret)
        ret[ret <= chop] <- 0
        ret
    }
}
class(fuzzy_bell) <- "charfun_generator"

fuzzy_sigmoid <-
function(cross = NULL, slope = 0.5, height = 1, chop = 0)
{
    if (!is.null(height) && (height < 0 || height > 1))
        stop("Height must be in the unit interval.")
    function(x) {
        if (is.null(cross))
            cross <- mean(range(x))
        ret <- 1 / (1 + exp(-slope * (x - cross)))
        if (!is.null(height))
            ret <- height * ret / max(ret)
        ret[ret <= chop] <- 0
        ret
    }
}
class(fuzzy_sigmoid) <- "charfun_generator"

fuzzy_trapezoid <-
function(corners = NULL, height = c(1,1), return_base_corners = TRUE)
{
    height <- rep_len(height, length.out = 2)
    if (any(height > 1) || any(height < 0))
        stop("Height(s) must be in the unit interval.")
    if (!is.null(corners) && length(corners) != 4L)
        stop("Need four corner values.")
    function(x) {
        if (is.null(corners))
            corners <- x[trunc(seq.int(1, length(x), length.out = 6))[2:5]]
        ret <- approxfun(corners, c(0, height, 0), rule = 2)(x)
        if (return_base_corners)
            ret[match(corners[c(1,4)], x)] <- .Machine$double.eps
        ret
    }
}
class(fuzzy_trapezoid) <- "charfun_generator"

fuzzy_triangular <-
function(corners = NULL, height = 1, return_base_corners = TRUE)
{
    if (height > 1 || height < 0)
        stop("Height must be in the unit interval.")
    if (!is.null(corners) && length(corners) != 3L)
        stop("Need three corner values.")
    function(x) {
        if (is.null(corners))
            corners <- x[trunc(seq.int(1, length(x), length.out = 5))[2:4]]
        ret <- approxfun(corners, c(0, height, 0), rule = 2)(x)
        if (return_base_corners)
            ret[match(corners[-2], x)] <- .Machine$double.eps
        ret
    }
}
class(fuzzy_triangular) <- "charfun_generator"

fuzzy_cone <-
function(center = NULL, radius = 2, height = 1, return_base_corners = TRUE)
{
    if (height > 1 || height < 0)
        stop("Height must be in the unit interval.")
    function(x) {
        if (is.null(center))
            center <- x[trunc((1 + length(x)) / 2)]
        ret <- approxfun(center + c(-radius, 0, radius),
                         c(0, height, 0), rule = 2)(x)
        if (return_base_corners)
            ret[match(center + c(-radius, radius), x)] <- .Machine$double.eps
        ret
    }
}
class(fuzzy_cone) <- "charfun_generator"

fuzzy_pi3 <-
function(mid = NULL, min = NULL, max = NULL, height = 1,
         return_base_corners = TRUE)
{
    if (height > 1 || height < 0)
        stop("Height must be in the unit interval.")
    function(x) {
        if (is.null(mid))
            mid <- x[trunc((1 + length(x)) / 2)]
        if (is.null(min)) min <- mid - 2
        if (is.null(max)) max <- mid + 2
        ret <- ifelse(x < min | x > max,
               0,
               ifelse(x < mid,
                      height * (1 - (x - mid) ^ 2 / (min - mid) ^ 2),
                      height * (1 - (x - mid) ^ 2 / (max - mid) ^ 2)
                      )
               )
        if (return_base_corners)
            ret[match(c(min, max), x)] <- .Machine$double.eps
        ret
    }
}
class(fuzzy_pi3) <- "charfun_generator"

fuzzy_pi4 <- function (knots, height = 1, return_base_corners = TRUE)
{
    if (height > 1 || height < 0)
        stop("Height must be in the unit interval.")
    if (length(knots) != 4L)
        stop("Need four knots.")
    function(x) {
        ret <- ifelse(x <= knots[1] | x >= knots[4],
                      0,
                      ifelse(x > knots[1] & x <= ((knots[1] + knots[2]) / 2),
                             2  * ((x - knots[1]) / (knots[2] - knots[1]))^2,
                             ifelse(x > ((knots[1] + knots[2])/2) & x < knots[2],
                                    1 - 2*((x - knots[2]) / (knots[2] - knots[1]))^2,
                                    ifelse(x >= knots[2] & x <= knots[3],
                                           height,
                                           ifelse(x > knots[3] & x <= ((knots[3] + knots[4]) / 2),
                                                  1 - 2 * ((x - knots[3]) / (knots[4] - knots[3]))^2,
                                                  2 * ((x - knots[4]) / (knots[4] - knots[3]))^2)))))
        if (return_base_corners)
            ret[match(c(knots[1], knots[4]), x)] <- .Machine$double.eps
        ret
    }
}
class(fuzzy_pi4)<- "charfun_generator"

## * fuzzy set generators for convenience

.expand <-
function(universe = NULL)
{
    if (is.null(universe))
        universe <- sets_options("universe")
    if (is.null(universe))
        universe <- seq.int(0, 20, 0.1)
    as.set(eval(universe))
}

fuzzy_normal_gset <-
function(mean = NULL, sd = 1, log = FALSE, height = 1, chop = 0,
         universe = NULL)
    gset(charfun = fuzzy_normal(mean = mean, sd = sd, log = log,
                                height = height, chop = chop),
         universe = .expand(universe))

fuzzy_two_normals_gset <-
function(mean = NULL, sd = c(1,1), log = c(FALSE, FALSE),
         height = 1, chop = 0, universe = NULL)
    gset(charfun = fuzzy_two_normals(mean = mean, sd = sd, log = log,
                                     height = height, chop = chop),
         universe = .expand(universe))

fuzzy_bell_gset <-
function(center = NULL, cross = NULL, slope = 4, height = 1, chop = 0,
         universe = NULL)
    gset(charfun = fuzzy_bell(center = center, cross = cross, slope = slope,
                              height = height, chop = chop),
         universe = .expand(universe))

fuzzy_sigmoid_gset <-
function(cross = NULL, slope = 0.5, height = 1, chop = 0,
         universe = NULL)
    gset(charfun = fuzzy_sigmoid(cross = cross, slope = slope,
                                 height = height, chop = chop),
         universe = .expand(universe))

fuzzy_trapezoid_gset <-
function(corners = NULL, height = c(1, 1), universe = NULL,
         return_base_corners = TRUE)
    gset(charfun = fuzzy_trapezoid(corners = corners, height = height,
                                   return_base_corners = return_base_corners),
         universe = .expand(universe))

fuzzy_triangular_gset <-
function(corners = NULL, height = 1, universe = NULL,
         return_base_corners = TRUE)
    gset(charfun = fuzzy_triangular(corners = corners, height = height,
                                    return_base_corners = return_base_corners),
         universe = .expand(universe))

fuzzy_cone_gset <-
function(center = NULL, radius = 2, height = 1, universe = NULL,
         return_base_corners = TRUE)
    gset(charfun = fuzzy_cone(center = center, radius = radius,
                              height = height,
                              return_base_corners = return_base_corners),
         universe = .expand(universe))

fuzzy_pi3_gset <-
function(mid = NULL, min = NULL, max = NULL, height = 1, universe = NULL,
         return_base_corners = TRUE)
    gset(charfun = fuzzy_pi3(mid = mid, min = min, max = max, height = height,
                             return_base_corners = return_base_corners),
         universe = .expand(universe))

fuzzy_pi4_gset <-
function(knots, height = 1, universe = NULL, return_base_corners = TRUE)
    gset(charfun = fuzzy_pi4(knots = knots, height = height, return_base_corners = return_base_corners),
         universe = .expand(universe))

### * tuple generator

fuzzy_tuple <-
function(FUN = fuzzy_normal, n = 5, ..., universe = NULL, names = NULL)
{
    universe <- .expand(universe)
    F <- if (is.charfun_generator(FUN))
        function(i) gset(charfun = FUN(i, ...), universe = universe)
    else
        function(i) FUN(i, universe = universe, ...)

    if (length(n) == 1L)
        n <- .get_support(universe)[seq.int(from = 1, to = length(universe),
                                        length.out = n)]
    .structure(as.tuple(lapply(n, F)), names = names)
}

set_power <-
function(x)
{
    if (length(x) < 1L) return(set(set()))
    set_union(set(set()),
              as.set(unlist(lapply(seq_along(x),
                                   function(i)
                                   apply(combn(.as.list(x), i), 2L, as.set)
                                   ),
                            recursive = FALSE)
                     )
              )
}

gset_power <-
function(x)
{
    if (length(x) < 1L) return(set(set()))
    .make_gset_from_list(c(list(gset()),
                           unlist(lapply(seq_along(x),
                                         function(i) gset_combn(x, i)
                                         ),
                                  recursive = FALSE)
                           )
                         )
}

cset_power <-
function(x)
{
    if (length(x) < 1L) return(set(set()))
    .make_gset_from_list(c(list(gset()),
                           unlist(lapply(seq_along(x),
                                         function(i) cset_combn(x, i)
                                         ),
                                  recursive = FALSE)
                           )
                         )
}

### * .cartesian_product

.cartesian_product <-
function(x)
{
    ## Cf. expand.grid().
    out <- vector("list", length(x))
    rep_fac <- 1L
    d <- lengths(x)
    orep <- prod(d)
    for(i in seq_along(x)) {
        nx <- d[i]
        orep <- orep / nx
        out[[i]] <-
            as.list(x[[i]])[rep.int(rep.int(seq_len(nx),
                                            rep.int(rep_fac, nx)), orep)]
        rep_fac <- rep_fac * nx
    }
    out
}

### * .make_list_of_tuples_from_list_of_lists

.make_list_of_tuples_from_list_of_lists <-
function(x)
{
    ret <- unlist(x, recursive = FALSE)
    dim(ret) <- c(length(x[[1L]]), length(x))
    apply(ret, 1L, as.tuple)
}

.make_set_of_tuples_from_list_of_lists <-
function(x)
    as.set(.make_list_of_tuples_from_list_of_lists(x))

### make sure that list elements are not destroyed during set unions

.make_list_elements <-
function(i)
    if (!is.cset(i)) list(i) else i

### as.character replacing empty strings by "_", as needed by environments
.as.character <-
function(x)
{
    x <- as.character(x)
    ind <- !is.na(x)
    if (any(ind)) {
        y <- x[ind]
        y[!nzchar(y)] <- "_"
        y <- substr(y, start = 1, stop = 256)
        while(any((n <- nchar(y, "byte")) > 256))
            y <- substr(y, 1, 256 - ceiling(n / 2))
        x[ind] <- y
    }
    x
}

## hash creator and accessors
.create_hash <-
function(s)
{
    if(length.set(s) < 1L) return(NULL)
    s <- .as.character(s)
    h <- new.env()
    for (i in seq.int(from = length(s), to = 1))
        h[[ s[[i]] ]] <- i
    h
}

### exact_match

.make_matchfun_from_equalityfun <-
function(equalityfun)
{
    equalityfun <- match.fun(equalityfun)
    function(x, table)
    {
        table <- .as.list(table)
        hash <- if (sets_options("hash")) .create_hash(table) else NULL
        l <- length(table)
        FUN <- function(i) {
            from <- if (is.null(hash))
                1
            else
                hash[[ .as.character(list(i)) ]]
            if (is.null(from) || l < 1L) return(NA)
            for (ind in seq.int(from = from, to = l))
                if (equalityfun(table[[ind]], i)) break
            if (ind < l || equalityfun(table[[ind]], i))
                ind else NA
        }
        ret <- lapply(.as.list(x), FUN)
        if (length(ret) < 1L) integer() else unlist(ret)
    }
}

.exact_match <-
    .make_matchfun_from_equalityfun(identical)

### .list_order/sort/unique

.list_order <-
function(x, decreasing = FALSE, ...) {
    .as.character <- function(x, ...) {
## FIXME: also serialize S4 objects
        rec <- unlist(lapply(x, is.recursive))
        x[rec] <- lapply(x[rec], serialize, NULL)
        unlist(lapply(x, paste, collapse = ""))
    }

    len <- length(x)
    if (len < 1L)
        return(integer(0L))

    l <- unlist(lapply(x, length))
    cl <- paste(lapply(x, class))

    num <- (cl == "integer" | cl == "numeric") & (l == 1L)
    numind <- numeric(len)
    numind[num] <- unlist(x[num])

    ch <- character(len)
    if (!all(num)) {
        ch[!num] <- .as.character(x[!num])
        loc <- ""
        suppressWarnings(if (capabilities("iconv")) {
            loc <<- Sys.getlocale("LC_COLLATE")
            on.exit(Sys.setlocale("LC_COLLATE", loc))
            Sys.setlocale("LC_COLLATE", "C")
            ch <- iconv(ch, to = "UTF-8")
        })
    }

    order(l, ## length
          cl, ## class
          numind, ## for numeric mode: numeric value
          ch, ## default: character representation
          decreasing = decreasing, ...)
}

.list_sort <-
function(x, decreasing = FALSE, ...)
    .as.list(x)[.list_order(x, decreasing = decreasing, ...)]

.list_unique <-
function(x)
    .as.list(x)[!duplicated(x)]

.domain_is_numeric <-
function(x)
    all(unlist(lapply(x, is.numeric)) &
        (unlist(lapply(x, length)) == 1L))

### fast structure replacement

.structure <- function(x, ...)
    `attributes<-`(x, c(attributes(x), list(...)))

gset_sum <-
function(...)
{
    l <- list(...)

    ## determine connector
    CON <- if (all(sapply(l, gset_is_crisp, na.rm = TRUE)))
        `+`
    else
        function(x, y) pmin(1, x + y)

    ## compute target support
    support <- do.call(set_union, l)

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, CON)

}

gset_difference <-
function(...)
{
    l <- list(...)

    ## connector
    CON <- function(x, y) pmax(0, x - y)

    ## compute target support
    support <- as.set(l[[1]])
    if (set_is_empty(support))
        return(set())

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, CON)
}

cset_sum <-
function(...)
{
    l <- list(...)

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## determine connector
    CON <- if (all(sapply(l, gset_is_crisp, na.rm = TRUE)))
        `+`
    else
        function(x, y) pmin(1, x + y)

    ## compute target support
    support <- cset(do.call(set_union, l),
                    matchfun = matchfun,
                    orderfun = orderfun)

    ## apply connector
    .make_cset_from_gset_and_orderfun_and_matchfun(
        .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                support,
                                                                CON,
                                                                matchfun),
        orderfun,
        matchfun)

}

cset_difference <-
function(...)
{
    l <- list(...)

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## connector
    CON <- function(x, y) pmax(0, x - y)

    ## compute target support
    support <- cset(as.set(l[[1]]),
                    matchfun = matchfun,
                    orderfun = orderfun)
    if (isTRUE(cset_is_empty(support)))
        return(support)


    ## apply connector
    .make_cset_from_gset_and_orderfun_and_matchfun(
        .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                support,
                                                                CON,
                                                                matchfun),
        orderfun,
        matchfun)
}

set_complement <-
function(x, y)
    .set_complement_using_matchfun(x, y)

.set_complement_using_matchfun <-
function(x, y, matchfun = .exact_match)
{
    if (missing(y))
        return(set())
    y <- .as.list(y)
    ind <- unique(na.omit(matchfun(x, y)))
    .make_set_from_list(if(length(ind)) y[-ind] else y)
}

gset_complement <-
function(x, y = NULL)
{
    if (is.null(y))
        .gset_complement(x)
    else
        gset_intersection(.gset_complement(x,
                         universe = set_union(gset_universe(x), gset_universe(y)),
                         bound = max(gset_bound(x), gset_bound(y))),
                         y
                         )
}

cset_complement <-
function(x, y = NULL)
{
    if (!is.null(y)) {
        matchfun <- .check_matchfun(list(x, y))
        orderfun <- .check_orderfun(list(x, y))
    } else {
        matchfun <- .matchfun(x)
        orderfun <- .orderfun(x)
    }

    Cx <- .make_cset_from_gset_and_orderfun_and_matchfun(
              .gset_complement(x,
                               universe = if(is.null(y)) cset_universe(x) else cset_union(cset_universe(x), cset_universe(y)),
                               bound = if(is.null(y)) cset_bound(x) else max(cset_bound(x), cset_bound(y))),
               orderfun,
               matchfun
    )
    if (is.null(y)) Cx else cset_intersection(Cx, y)
}

.gset_complement <-
function(x, universe = gset_universe(x), bound = gset_bound(x))
{
    if (set_is_empty(universe) || bound < 1L)
        return(set())

    ## efficiency hack: by default, the complement gets the universe
    ## of the original set, *except* if there is a default universe
    ## and the universe attribute of the original set is missing.
    tuniverse <-
        if (!is.null(sets_options("universe")) && is.null(attr(x, "universe")))
            NULL
        else
            universe

    if (isTRUE(gset_is_set(x)) && bound == 1L)
        gset(set_complement(x, universe), universe = tuniverse, bound = 1L)
    else if (gset_is_crisp(x, na.rm = TRUE)) {
        M <- gset(universe, rep.int(bound, length(universe)))
        .structure(.set_bound(.set_universe(gset_difference(M, x), tuniverse), bound),
                  class = c("gset", "cset"))
    } else if (gset_is_fuzzy_set(x, na.rm = TRUE) && bound == 1L)
        .make_gset_from_support_and_memberships(
              universe,
              .N.(.memberships_for_support(x, universe)),
              universe = tuniverse,
              bound = bound
        )
    else {
        connector <- function(x, y) .N.(y)
        M <- gset(universe, rep.int(bound, length(universe)))
        memberships <-
            .apply_connector_to_list_of_gsets_and_support(list(M, x),
                                                          universe,
                                                          connector)
        .make_gset_from_support_and_memberships(
              universe,
              memberships,
              universe = tuniverse,
              bound = bound
        )
    }
}

set_cartesian <-
function(...)
{
    if (nargs() < 2L)
        return(..1)
    l <- list(...)
    if (!all(len <- lengths(l)))
        return(set())
    as.set(.make_list_of_tuples_from_list_of_lists(.cartesian_product(l)))
}

### FIXME: does it make sense to add a na.rm argument here?
### For fuzzy multisets, we are using .T. which has no na.rm ...
gset_cartesian <-
function(...)
{
    ## handle arguments
    if (nargs() < 2L)
        return(..1)
    ll <- list(...)
    if (isTRUE(all(gset_is_set(ll))))
        return(as.gset(set_cartesian(...)))

    ## handle empty sets
    if (any(vapply(ll, gset_cardinality, double(1L)) == 0, na.rm = TRUE))
        return(gset())

    l <- lapply(ll, as.list)

    ## compute cartesian products of support and memberships
    support <- .make_list_of_tuples_from_list_of_lists(.cartesian_product(l))
    memberships <- lapply(l, .get_memberships)
    memberships <- do.call(Map, c("list", .cartesian_product(memberships)))


    ## compute tuple memberships by applying the T-norm to the components
    memberships <-
        if (all(gset_is_crisp(ll), na.rm = TRUE))
            vapply(memberships, function(i) prod(unlist(i)), integer(1L))
        else if (all(gset_is_fuzzy_set(ll), na.rm = TRUE)) {
            vapply(memberships, function(i) Reduce(.T., unlist(i)), double(1L))
        } else {
            lapply(memberships, function(i) {
                ## normalize memberships
                maxlen <- max(vapply(i, gset_cardinality, double(1L), na.rm = TRUE))
                m <- lapply(i, .expand_membership, len = maxlen, rep = FALSE)
                mult <- unlist(do.call(Map, c(list(prod),
                                              lapply(m, .get_memberships)
                                              ))
                               )

                ## compute T-norm
                S <- Reduce(.T., m)
                .make_gset_from_support_and_memberships(as.list(S), mult)
            })
        }
    ## create target set
    .make_gset_from_support_and_memberships(support, memberships)
}

cset_cartesian <-
function(...)
    gset_cartesian(...)

#########################
### Customizable Sets ###
#########################

### generator

cset <-
function(gset,
         orderfun = sets_options("orderfun"),
         matchfun = sets_options("matchfun"))
{
    gset <- as.gset(gset)

    ## recreate gset according to user-specified match-fun
    if (!is.null(matchfun)) {
        uni <- !.duplicated_by_matchfun(gset, matchfun)
        gset <-
            .make_gset_by_support_and_memberships(.as.list(.get_support(gset)[uni]),
                                                  .get_memberships(gset)[uni],
                                                  universe = .get_universe(gset),
                                                  bound = .get_bound(gset)
                                                  )
    }

    ## create cset-object
    .make_cset_from_gset_and_orderfun_and_matchfun(gset,
                                                   orderfun,
                                                   matchfun)
}

## convenience function generator for non-vectorized equality predicates
matchfun <-
function(FUN)
    .make_matchfun_from_equalityfun(FUN)

## Disable numeric subscripting (as sets are "unordered" collections of
## elements).  Note that iterating via for() and lapply() still works,
## the former because this [currently, 2007-09-16] directly uses the
## internal list representation and the latter because we provide an
## as.list() method.

`[.cset` <-
function(x, i = x)
{
    ind <- .lookup_elements(x, i, .matchfun(x))
    cset(gset(.as.list(x)[ind], cset_memberships(x)[ind]),
         .orderfun(x),
         .matchfun(x))
}

`[[.cset` <-
function(x, i)
{
    as.set(x)[[i]]
}

`[<-.cset` <-
function(x, i = x, value)
{
    cset(gset(`[<-`(.as.list(x), .lookup_elements(x, i, .matchfun(x)), value),
                memberships = cset_memberships(x)),
         .orderfun(x),
         .matchfun(x))
}

`[[<-.cset` <-
function(x, i, value)
{
    if (!is.character(i) || length(i) > 1L) i <- list(i)
    if (length(lookup <- .lookup_elements(x, i, .matchfun(x))) < 1L) return(x)

    cset(gset(`[[<-`(.as.list(x), lookup, value),
         memberships = cset_memberships(x)),
         .orderfun(x),
         .matchfun(x))
}

### na.omit

na.omit.cset <-
function(object, ...)
{
    m <- .get_memberships(object)
    if (!is.list(m))
        m[is.na(m)] <- 0
    else
        m <- lapply(m, function(i) {
            m2 <- .get_memberships(i)
            m2[is.na(m2) | is.na(i)] <- 0
            .make_gset_from_support_and_memberships(i, m2)
        })
    .make_gset_from_support_and_memberships(object, m)
}

### Ops-method

Ops.cset <-
function(e1, e2)
{
    if(nargs() == 1L) {
        ## dispatch manually for subclasses
        if (inherits(e1, "set"))
            return(Ops_set(e1, e2, .Generic = .Generic, .Class = .Class))
        if (inherits(e1, "gset"))
            return(Ops_gset(e1, e2, .Generic = .Generic, .Class = .Class))

        if(!(as.character(.Generic) %in% "!"))
            stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                          .Generic, .Class),
                 domain = NA)
        return(cset_complement(e1))
    }

    ## dispatch manually for subclasses
    if (inherits(e1, "set") && inherits(e2, "set"))
        return(Ops_set(e1, e2, .Generic = .Generic, .Class = .Class))
    if (inherits(e1, "gset") && inherits(e2, "gset"))
        return(Ops_gset(e1, e2, .Generic = .Generic, .Class = .Class))

    if(!(as.character(.Generic)
         %in% c("<", "<=", ">", ">=", "==", "!=",
                "&", "|", "*", "+", "-", "^")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    if(as.character(.Generic) == "^") {
        if(is.cset(e1) &&
            ((trunc(e2) != e2) || (e2 < 1L)))
            stop("Cartesian product only defined for positive integers.")
        if(is.cset(e2) && (e1 != 2L))
            stop("Operator not defined.")
    }

    switch(.Generic,
           "+"  = cset_sum(e1, e2),
           "|"  = cset_union(e1, e2),
           "-"  = cset_difference(e1, e2),
           "&"  = cset_intersection(e1, e2),
           "*"  = cset_cartesian(e1, e2),
           "<"  = cset_is_proper_subset(e1, e2),
           "<=" = cset_is_subset(e1, e2),
           ">"  = cset_is_proper_subset(e2, e1),
           ">=" = cset_is_subset(e2, e1),
           "==" = cset_is_equal(e1, e2),
           "!=" = !cset_is_equal(e1, e2),
           "^"  = {
               if(is.cset(e2))
                   cset_power(e2)
               else
                   do.call(cset_cartesian, rep.int(list(e1), e2))}
           )

}

### print methods

print.cset <-
function(x, ...)
    print.gset(x, ...)

print.summary.cset <-
function(x, ...)
    print.summary.gset(x, ...)

### format method

format.cset <-
function(x, ...) {
    FUN <- cset_orderfun(x)
    x <- if (isTRUE(cset_is_set(x)))
        .as.list(x)
    else
        .make_list_of_elements_from_cset(x)
    if (is.function(FUN))
        x <- x[FUN(x)]
    else if(is.integer(FUN) && (length(x) == length(FUN)))
        x <- x[FUN]
    .format_set_or_tuple(x, "{", "}", ...)
}

## summary method

summary.cset <-
function(object, ...)
{
    len <- cset_cardinality(object)
    if (na <- is.na(len))
        len <- length.set(object)
    out <- if (len == 0L)
        gettext("The empty set.")
    else if (len == 1L)
        gettext("A customizable set with 1 element.")
    else if (na)
        gettextf("A customizable set with %g elements.", len)
    else
        gettextf("A customizable set with cardinality %g.", len)
    if(!is.null(attr(object, "matchfun")) && !is.null(attr(object, "orderfun")))
        out <- paste(out, "The match and order functions are user-defined.")
    else if(!is.null(attr(object, "matchfun")))
        out <- paste(out, "The match function is user-defined.")
    else if(!is.null(attr(object, "orderfun")))
        out <- paste(out, "The order function is user-defined.")

    .structure(out, class = "summary.cset")
}

### internal stuff

.make_cset_from_gset_and_orderfun_and_matchfun <-
function(gset, orderfun = NULL, matchfun = NULL)
{
    ## make sure that default orderfun and default matchfun are never stored
    if (identical(orderfun, .list_order))
        orderfun <- NULL
    if (identical(matchfun, .exact_match))
        matchfun <- NULL

    ## promote to gset, if only default-funs are specified
    if (is.null(matchfun) && is.null(orderfun))
        return(gset)

    ## create structure (including overwriting gset-class)
    .structure(gset,
               orderfun = orderfun,
               matchfun = matchfun,
               class = "cset")
}

.duplicated_by_matchfun <-
function(x, matchfun)
    duplicated(.as.list(x)[matchfun(x, x)])

.list_unique_by_matchfun <-
function(x, matchfun)
    .as.list(x)[!.duplicated_by_matchfun(x, matchfun)]

.check_matchfun <-
function(l)
{
    matchfun <- cset_matchfun(l[[1]])
    if (!all(sapply(l, function(i) identical(cset_matchfun(i), matchfun))))
        stop("Need same match functions (or none) for all elements.")
    matchfun
}

.check_orderfun <-
function(l)
{
    orderfun <- cset_orderfun(l[[1]])
    if (!all(sapply(l, function(i) identical(cset_orderfun(i), orderfun))))
        NULL
    else
        orderfun
}

rep.set <-
function(x, ...)
    x

rep.gset <-
function(x, ...)
{
    l <- list(...)
    each <- if (length(l) == 0L)
        1
    else
        l[[1]]
    m <- .get_memberships(x)
    memberships <- if (gset_is_crisp(x, na.rm = TRUE))
        m * each
    else if (gset_is_fuzzy_set(x, na.rm = TRUE))
        lapply(m, rep, each = each)
    else
        lapply(m, function(i) gset(.get_support(i), .get_memberships(i) * each))

    .make_gset_from_support_and_memberships(.get_support(x),
                                           .canonicalize_memberships(memberships))
}

rep.cset <-
function(x, ...)
    rep.gset(x, ...)

########################
### Generalized Sets ###
########################

## The gset class extends 'ordinary' sets by allowing
## membership values for each element. Strictly speaking,
## a generalized set is a pair (A, mu) with support A and
## membership function mu, where mu is a mapping from
## A into 2 ^ ([0, 1] -> N). Special cases are:
## mu : A -> {0, 1} ("ordinary sets")
## mu : A -> [0, 1] ("fuzzy sets")
## mu : A -> 2 ^ {0, 1} ("multi sets")

### Basic stuff (constructors, print/summary methods)

## * gset generator

gset <-
function(support = NULL, memberships = NULL, charfun = NULL,
         elements = NULL, universe = NULL, bound = NULL,
         assume_numeric_memberships = TRUE)
{
    Universe <- universe
    if(is.null(universe))
        universe <- sets_options("universe")

    ### some checks
    if (!is.null(elements) &&
        !(is.null(support) && is.null(memberships) && is.null(charfun)))
        stop("'elements' needs to be specified alone.")
    if (is.null(support) && !is.null(memberships))
        stop("'membership' requires 'support'.")
    if (is.null(universe) && !is.null(charfun))
        stop("'charfun' requires 'universe' (or a default universe).")
    if (!is.null(memberships) && !is.null(charfun))
        stop("Need either 'memberships' and 'support', or 'charfun' and 'universe'.")

    ### element specification:
    ## split support and memberships, and proceed
    if (!is.null(elements)) {
        support <- unlist(elements, recursive = FALSE)
        memberships <- sapply(elements, .get_memberships)
    }

    ### universe & charfun specification:
    ## create memberships from charfun, and proceed
    ## In general, create list of memberships (i.e., fuzzy multisets)
    ## but assume numeric memberships if specified (and by default)
    if (!is.null(universe))
        universe <- if (is.interval(universe))
            as.set(as.numeric(universe))
        else
            as.set(eval(universe))
    if (!is.null(charfun)) {
        if (is.charfun_generator(charfun))
            charfun <- charfun()
        memberships <- if (assume_numeric_memberships) {
            if (.domain_is_numeric(universe))
                charfun(unlist(universe))
            else
                vapply(universe, charfun, double(1L))
        } else
            lapply(universe, charfun)
        support <- universe
        .stop_if_memberships_are_invalid(memberships,
                                         "Membership function invalid.\n  ")
    }

    ## handle empty set
    if (is.null(support))
        return(set())

    ### support & membership specification:
    ## just check memberships
    if (!is.null(memberships))
        .stop_if_memberships_are_invalid(memberships)

    ## check support against universe
    if (!is.null(universe) && !set_is_subset(as.set(support), universe))
        stop("Support must be a subset of the universe.")

    ### canonicalize memberships & create gset
    .make_gset_from_support_and_memberships(support, memberships, Universe,
                                            bound)
}

print.gset <-
function(x, ...)
{
    writeLines(strwrap(format(x, ...), exdent = 1L))
    invisible(x)
}

summary.gset <-
function(object, ...)
{
    len <- gset_cardinality(object)
    if (na <- is.na(len))
        len <- length.set(object)
    out <- if (len == 0L)
        gettext("The empty set.")
    else if (len == 1L)
        gettext("A generalized set with 1 element.")
    else if (na)
        gettextf("A generalized set with %g elements.", len)
    else
        gettextf("A generalized set with cardinality %g.", len)
    .structure(out, class = "summary.gset")
}

print.summary.gset <-
function(x, ...)
{
    writeLines(x)
    invisible(x)
}

format.gset <-
function(x, ...) {
    x <- if (isTRUE(gset_is_set(x)))
        .as.list(x)
    else
        .make_list_of_elements_from_cset(x)
    .format_set_or_tuple(x, "{", "}", ...)
}

### operators

## Disable numeric subscripting (as sets are "unordered" collections of
## elements).  Note that iterating via for() and lapply() still works,
## the former because this [currently, 2007-09-16] directly uses the
## internal list representation and the latter because we provide an
## as.list() method.

`[.gset` <-
function(x, i = x)
{
    ind <- .lookup_elements(x, i)
    gset(.as.list(x)[ind], gset_memberships(x)[ind])
}

`[[.gset` <-
function(x, i)
{
    as.set(x)[[i]]
}

`[<-.gset` <-
function(x, i = x, value)
{
    gset(`[<-`(.as.list(x), .lookup_elements(x, i), value),
         memberships = gset_memberships(x))
}

`[[<-.gset` <-
function(x, i, value)
{
    if (!is.character(i) || length(i) > 1L) i <- list(i)
    if (length(lookup <- .lookup_elements(x, i)) < 1L) return(NULL)
    gset(`[[<-`(.as.list(x), lookup, value),
         memberships = gset_memberships(x))
}

Ops_gset <-
function(e1, e2, .Generic, .Class)
{
    if (missing(e2) == 1L) {
        if(!(as.character(.Generic) %in% "!"))
            stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                          .Generic, .Class),
                 domain = NA)
        return(gset_complement(e1))
    }

    if(!(as.character(.Generic)
         %in% c("<", "<=", ">", ">=", "==", "!=",
                "&", "|", "*", "+", "-", "^")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    if(as.character(.Generic) == "^") {
        if(is.gset(e1) &&
            ((trunc(e2) != e2) || (e2 < 1L)))
            stop("Cartesian product only defined for positive integers.")
        if(is.gset(e2) && (e1 != 2L))
            stop("Operator not defined.")
    }

    switch(.Generic,
           "+"  = gset_sum(e1, e2),
           "|"  = gset_union(e1, e2),
           "-"  = gset_difference(e1, e2),
           "&"  = gset_intersection(e1, e2),
           "*"  = gset_cartesian(e1, e2),
           "<"  = gset_is_proper_subset(e1, e2),
           "<=" = gset_is_subset(e1, e2),
           ">"  = gset_is_proper_subset(e2, e1),
           ">=" = gset_is_subset(e2, e1),
           "==" = gset_is_equal(e1, e2),
           "!=" = !gset_is_equal(e1, e2),
           "^"  = {
               if(is.gset(e2))
                   gset_power(e2)
               else
                   do.call(gset_cartesian, rep.int(list(e1), e2))}
           )

}

### internal stuff

.make_gset_from_list <-
function(list)
    .structure(list, class = c("gset", "cset"))

.make_gset_from_support_and_memberships <-
function(support, memberships, universe = NULL, bound = NULL)
{
    ## simplify memberships:
    if (!is.null(memberships)) {
        ## canonicalize (i.e., make sure that fuzzy multiset memberships
        ## are valid gset objects)
        memberships <- .canonicalize_memberships(memberships)

        ## check length
        if (length.set(memberships) != length.set(support))
            stop("Length of support must match length of memberships.")

        ## for fuzzy multisets, remove elements in membership with
        ## zero support
        if (is.list(memberships)) {
            # find zero elements in list
            z <- lapply(memberships,
                        function(i) sapply(i, `>`, 0) | sapply(i, is.na))
            # empty sets should have length 0
            z <- lapply(z, function(i) if (is.list(i)) 0 else i)
            # filter 0 elements
            memberships <-
                Map(.make_gset_by_support_and_memberships,
                    Map("[", lapply(memberships, .as.list), z),
                    Map("[", lapply(memberships, gset_memberships), z))
        }

        ## compute index of entries with 0 memberships.
        z <- if (is.list(memberships))
            sapply(memberships, gset_cardinality) == 0
        else
            memberships == 0

        ## all zero? return empty set.
        if (!any(is.na(z)) && all(z)) return(set())

        ## remove entries with zero membership.
        if (any(z, na.rm = TRUE)) {
            support <- .as.list(support)[-which(z)]
            memberships <- memberships[-which(z)]
        }
    }

    ## if gset has no or trivial membership:
    if (is.null(memberships) || is.atomic(memberships) &&
        !any(is.na(memberships)) && all(memberships == 1, na.rm = TRUE)) {
        support <- .list_sort(.as.list(support))
        ## if a universe exists, return gset
        if (!is.null(universe))
            return(.make_gset_by_support_and_memberships(support = support,
                                                         memberships = NULL,
                                                         universe = universe))
        else
            return(.make_set_from_list(support))
    }

    ## if gset has fuzzy multiset representation, but
    ## is really only multi or fuzzy, simplify memberships.
    if (is.list(memberships)) {
        tmp <- lapply(memberships, .as.list)
        if (all(lengths(tmp) == 1L)) {
            M <- unlist(memberships)
            if (!any(is.na(M)) && all(M == 1L))
                memberships <- sapply(memberships, .get_memberships)
            else {
                M <- sapply(memberships, .get_memberships)
                if (!any(is.na(M)) && all(M == 1L))
                    memberships <- unlist(memberships)
            }
        }
    }

    ## check memberships against bound, if any.
    bd <- bound
    if (is.null(bd))
        bd <- sets_options("bound")
    if (!is.null(bd) && (max(.multiplicities(memberships), na.rm = TRUE) > bd))
        stop("Memberships must not exceed bound!")

    ## convert support to set. Reorder memberships accordingly, if needed.
    S <- canonicalize_set_and_mapping(x = support, mapping = memberships)

    ## return gset.
    .make_gset_by_support_and_memberships(support = S$set,
                                          memberships = S$mapping,
                                          universe = universe,
                                          bound = bound)
}

.make_gset_by_support_and_memberships <-
function(support, memberships, universe = NULL, bound = NULL)
{
    if (is.null(universe) && (is.null(memberships) ||
        is.atomic(memberships) && !any(is.na(memberships)) &&
                              all(memberships == 1, na.rm = TRUE)))
        .make_set_from_list(support)
    else
        .structure(support,
                   memberships = memberships,
                   universe = universe,
                   bound = bound,
                   class = c("gset", "cset"))
}

.stop_if_memberships_are_invalid <-
function(memberships, errmsg = NULL)
{
    if (is.list(memberships)) {
        for (i in memberships) {
            if (is.gset(i)) {
                if (!gset_is_crisp(i, na.rm = TRUE))
                    stop("For fuzzy multisets, the memberships must be (multi)sets.",
                         call. = FALSE)
                if (!all(sapply(i,
                                function(j) (is.na(j) || is.numeric(j)) && (j >= 0) && (j <= 1)), na.rm = TRUE))
                    stop("For fuzzy multisets, the memberships must be multisets over the unit interval.", call. = FALSE)
            }
        }
    } else {
        if (any(memberships < 0, na.rm = TRUE))
            stop(paste0(errmsg,
                        "Memberships must be positive."), call. = FALSE)
        if (any(memberships > 1, na.rm = TRUE) && any(memberships != trunc(memberships), na.rm = TRUE))
            stop(paste0(errmsg,
                        "Memberships must be either all integer (multisets),\n  or all in the unit interval (fuzzy sets)."), call. = FALSE)
    }
}

.apply_connector_to_list_of_gsets_and_support <-
function(l, support, connector, matchfun = .exact_match,
         enforce_general_case = FALSE)
{
    ## extract memberships according to support
    m <- lapply(l, .memberships_for_support, support, matchfun)

    ## apply connector to memberships
    if (!enforce_general_case && !any(sapply(l, gset_is_fuzzy_multiset)))
        ## for multisets and fuzzy sets, just use normalized memberships
        Reduce(connector, m)
    else {
        ## in all other cases, expand memberships first:

        ## determine max. membership cardinality
        is_list <- sapply(m, is.list)
        FUN <- function(i) if (all(is.na(i))) 0 else max(i, na.rm = TRUE)
        mlen <- max(if (any(is_list)) sapply(m[is_list], sapply,
                                             gset_cardinality),
                    if (any(!is_list)) sapply(m[!is_list], FUN),
                    na.rm = TRUE)

        ## group by support, expand memberships, and apply connector
        lapply(seq_along(support), function(i) {
            Reduce(connector,
                   lapply(m, function(j) .expand_membership(j[[i]],
                                                            len = mlen)))
        })
    }
}

.make_gset_from_list_of_gsets_and_support_and_connector <-
function(l, support, connector, matchfun = .exact_match,
         enforce_general_case = FALSE)
{
    ## compute memberships
    memberships <-
        .apply_connector_to_list_of_gsets_and_support(l, support, connector,
                                                      matchfun,
                                                      enforce_general_case)

    ## create resulting gset
    .make_gset_from_support_and_memberships(support, memberships)

}

.make_list_of_elements_from_cset <-
function(x)
    Map(.make_element_from_support_and_memberships,
        .as.list(x),
        .as.list(.get_memberships(x)))

############
### Sets ###
############

### Basic stuff (constructors, print/summary methods)
set <-
function(...) {
    ret <- .set(...)
    n <- names(ret)
    if (!is.null(n) && any(duplicated(n[n != ""])))
        stop("Labels must be unique.")
    ret
}

.set <- function(...)
    .make_set_from_list(.list_sort(.list_unique(list(...))))

print.set <-
function(x, ...)
{
    writeLines(strwrap(format(x, ...), exdent = 1L))
    invisible(x)
}

summary.set <-
function(object, ...)
{
    len <- length(object)
    out <- if (len == 0L)
        gettext("The empty set.")
    else if (len == 1L)
        gettext("A set with 1 element.")
    else
        gettextf("A set with %d elements.", len)
    .structure(out, class = "summary.set")
}

print.summary.set <-
function(x, ...)
{
    writeLines(x)
    invisible(x)
}

format.set <-
function(x, ...) {
    .format_set_or_tuple(x, "{", "}", ...)
}

Math.set <-
function(x, ...)
    as.set(get(.Generic)(unlist(x), ...))


### operators

Ops_set <-
function(e1, e2, .Generic, .Class)
{
    if (missing(e2)) {
        if(!(as.character(.Generic) %in% "!"))
            stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                          .Generic, .Class),
                 domain = NA)
        return(gset_complement(e1))
    }

    if(!(as.character(.Generic)
         %in% c("<", "<=", ">", ">=", "==", "!=",
                "&", "|", "*", "+", "-", "^")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    if(as.character(.Generic) == "^") {
        if(is.set(e1) &&
            ((trunc(e2) != e2) || (e2 < 1L)))
            stop("Cartesian product only defined for positive integers.")
        if(is.set(e2) && (e1 != 2L))
            stop("Operator not defined.")
    }

    switch(.Generic,
           "+"  = gset_sum(e1, e2),
           "|"  = set_union(e1, e2),
           "-"  = gset_difference(e1, e2),
           "&"  = set_intersection(e1, e2),
           "*"  = set_cartesian(e1, e2),
           "<"  = set_is_proper_subset(e1, e2),
           "<=" = set_is_subset(e1, e2),
           ">"  = set_is_proper_subset(e2, e1),
           ">=" = set_is_subset(e2, e1),
           "==" = set_is_equal(e1, e2),
           "!=" = !set_is_equal(e1, e2),
           "^"  = {
               if(is.set(e2))
                   set_power(e2)
               else
                   do.call(set_cartesian, rep.int(list(e1), e2))
           }
           )

}

### Subscript methods

## Disable numeric subscripting (as sets are "unordered" collections of
## elements).  Note that iterating via for() and lapply() still works,
## the former because this [currently, 2007-09-16] directly uses the
## internal list representation and the latter because we provide an
## as.list() method.

`[.set` <-
function(x, i = x)
{
    .make_set_from_list(.list_sort(.as.list(x)[.lookup_elements(x, i)]))
}

`[[.set` <-
function(x, i)
{
    if (!is.character(i) || length(i) > 1L || !nzchar(i))
        i <- list(i)
    if (length(lookup <- .lookup_elements(x, i)) < 1L) return(NULL)
    .as.list(x)[[lookup]]
}

`[<-.set` <-
function(x, i = x, value)
{
    .make_set_from_list(.list_sort(.list_unique(`[<-`(.as.list(x),
                        .lookup_elements(x, i), value))))
}

`[[<-.set` <-
function(x, i, value)
{
    if (!is.character(i) || length(i) > 1L) i <- list(i)
    if (length(lookup <- .lookup_elements(x, i)) < 1L) return(x)
    .make_set_from_list(.list_sort(.list_unique(`[[<-`(.as.list(x),
                                                       lookup, value))))
}

### internal stuff

.make_set_from_list <-
function(x)
   .structure(x, class = c("set", "gset", "cset"))

.format_set_or_tuple <-
function(x, left, right, ...)
{
    nms <- names(x)
    names(x) <- NULL
    SEP <- rep.int("", length(x))
    if (!is.null(nms))
        SEP[nms != ""] <- " = "
    paste(left,
          if (length(x) > 0)
              paste(nms, SEP, LABELS(as.list(x), ...),
                    sep = "", collapse = ", "),
          right,
          sep = "")
}

.set_subset<-
function(x, i)
    as.set(as.list(x)[i])

`.set_replace`<-
function(x, i, value)
{
    ret <- as.list(x)
    ret[i] <- value
    as.set(ret)
}

.set_subset2<-
function(x, i)
    as.list(x)[[i]]

`.set_replace2`<-
function(x, i, value)
{
    ret <- as.list(x)
    ret[[i]] <- value
    as.set(ret)
}

.lookup_elements <-
function(x, values, matchfun = .exact_match)
{
    n <- names(x)
    ind <- if (!is.null(n) && is.character(values))
        matchfun(values, n)
    else {
        values <- .as.list(values)
        matchfun(values, x)
    }
    ind[!is.na(ind)]
}

#### fuzzy inference

##`%is%` <- function(x, y) gset_charfun(x)(y)
 FUN <- matchfun(function(x,y) isTRUE(all.equal(x,y)))
`%is%` <- function(x, y) cset_charfun(cset(x, matchfun = FUN))(y)

fuzzy_inference <-
function(system, values, implication = c("minimum", "product"))
{
    if (is.character(implication))
        implication <-
            switch(match.arg(implication), minimum = pmin, product = `*`)

    transform_rule <- function(expr) {
        as.call(switch(as.character(expr[[1L]]),
                       "&&" = list(as.name(".T."),
                                   Recall(expr[[2L]]), Recall(expr[[3L]])),
                       "||" = list(as.name(".S."),
                                   Recall(expr[[2L]]), Recall(expr[[3L]])),
                       "!" = list(as.name(".N."), Recall(expr[[2L]])),
                       "%is%" = list(expr[[1L]], as.call(list(as.name("$"),
                                     expr[[2L]], expr[[3L]])),
                                     values[[ as.character(expr[[2L]]) ]]),
                       "(" = list(expr[[1L]], Recall(expr[[2L]])),
                       expr))
    }

    antecedents <-
        sapply(lapply(system$rules,
                      function(i) transform_rule(i$antecedent)),
               eval, system$variables)
    antecedents[is.na(antecedents)] <- 0

    consequents <-
        lapply(system$rules,
               function(i) eval(`[[<-`(i$consequent, 1L, as.name("$")),
                                   system$variables))

    do.call(gset_union,
            Map(gset_transform_memberships,
                consequents,
                rep.int(list(implication), length(antecedents)),
                antecedents))
}

fuzzy_rule <-
function(antecedent, consequent)
{
    .structure(list(antecedent = substitute(antecedent),
                    consequent = substitute(consequent)),
               class = "fuzzy_rule")
}

print.fuzzy_rule <-
function(x, ...)
{
    writeLines(paste(format(x$antecedent), "=>", format(x$consequent)))
    invisible(x)
}

fuzzy_system <-
function(variables, rules)
    .structure(list(variables = variables, rules = rules),
               class = "fuzzy_system")


print.fuzzy_system <-
function(x, ...)
{
    writeLines(sprintf("A fuzzy system consisting of %i variables and %i rules.\n", length(x$variables), length(x$rules)))

    writeLines("Variables:\n")
    writeLines(paste(names(x$variables),
                     "(",
                     lapply(x$variables,
                            function(i) paste(names(i), collapse= ", ")),
                     ")", sep = ""))

    writeLines("\nRules:\n")
    for (i in x$rules) print(i)

    invisible(x)
}

plot.fuzzy_system <-
function(x,
         colfun = function(i) grey.colors(i, start = 0.5, end = 0), ...)
{
    vars <- as.list(x$variables)
    nams <- names(vars)

    n <- length(vars)
    nc <- ceiling(sqrt(n))
    nr <- ceiling(n / nc)
    op <- par(mfrow = c(nr, nc))
    on.exit(par(op))

    for (i in seq_len(n))
        plot(vars[[i]], main = nams[i], col = colfun(length(vars[[i]])))

    invisible(x)
}

fuzzy_partition <-
function(varnames, FUN = fuzzy_normal, universe = NULL, ...)
{
    n <- length(varnames)
    if (is.numeric(varnames)) {
        nam <- names(varnames)
        empty <- !nzchar(nam)
        if (is.null(nam) || any(is.na(empty)) || any(empty))
            stop("Position vector must be named.")
        n <- as.numeric(varnames)
        varnames <- names(varnames)
    }
    .structure(fuzzy_tuple(FUN = FUN, n = n, ...,
                           universe = universe, names = varnames),
               class = c("fuzzy_variable", "tuple"))
}

fuzzy_variable <-
function(...) {
    n <- names(l <- list(...))
    if (is.null(n) || any(!nzchar(n)))
        stop("Argument list must be named.")
    universe <- .expand()
    for (i in seq_along(l))
        if (is.function(l[[i]]))
            l[[i]] <- gset(charfun = l[[i]], universe = universe)
    .structure(as.tuple(l), class = c("fuzzy_variable", "tuple"))
}

print.fuzzy_variable <-
function(x, ...)
{
    writeLines(sprintf("A fuzzy variable with values: %s",
                       paste(names(x), collapse = ", ")))

    invisible(x)

}

plot.fuzzy_variable <-
function(x, ypos = 1, ...)
{
    plot.tuple(x, ...)
    op <- par(xpd = TRUE)
    on.exit(par(op))
    pos = sapply(x, gset_defuzzify, "meanofmax")
    text(pos, ypos, labels = names(pos), pos = 3L)

    invisible(x)
}

.meanfun <-
list(arithmetic = function(x, y) (x + y) / 2,
     geometric = function(x, y) sqrt(x * y),
     harmonic = function(x, y) 2 / (1 / x + 1 / y)
     )

gset_mean <-
function(x, y, type = c("arithmetic", "geometric", "harmonic"))
{
    l <- list(x, y)
    FUN <- .meanfun[[match.arg(type)]]

    ## compute target support
    support <- do.call(set_union, l)

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, FUN,
                                                            enforce_general_case = TRUE)
}

cset_mean <-
function(x, y, type = c("arithmetic", "geometric", "harmonic"))
{
    l <- list(x, y)
    FUN <- .meanfun[[match.arg(type)]]

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## compute target support using correct matchfuns
    support <- cset(do.call(set_union, l),
                    matchfun = matchfun,
                    orderfun = orderfun)

    ## create gset by applying conorm, and then make cset
    .make_cset_from_gset_and_orderfun_and_matchfun(
         .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                 support,
                                                                 FUN,
                                                                 matchfun,
                                                                 TRUE),
         orderfun,
         matchfun)
}

set_cardinality <-
length.gset <-
length.cset <-
length.set <-
function(x)
    length(.as.list(x))

lengths.gset <-
lengths.cset <-
lengths.set <-
function(x, use.names = TRUE)
    lengths(.as.list(x), use.names = use.names)

gset_cardinality <-
function(x, type = c("absolute", "relative"), na.rm = FALSE)
{
    type <- match.arg(type)

    ret <- if (isTRUE(gset_is_set(x)))
        length.set(x)
    else if (gset_is_fuzzy_multiset(x))
        sum(sapply(.get_memberships(x),
                   function(i) sum(as.double(i) * .get_memberships(i),
                                   na.rm = na.rm)),
            na.rm = na.rm)
    else
        sum(.get_memberships(x), na.rm = na.rm)

    if (type == "absolute")
        ret
    else
        ret / length.set(x)
}

cset_cardinality <-
function(x, type = c("absolute", "relative"), na.rm = FALSE)
    gset_cardinality(x, type, na.rm)

str.set <-
function(object, ...)
{
    str(unclass(object), ...)
    writeLines(" - attr(*, \"class\")= chr \"set\"")
}

str.gset <-
function(object, ...)
{
    str(unclass(object), ...)
    writeLines(" - attr(*, \"class\")= chr \"gset\"")
}

str.cset <-
function(object, ...)
{
    str(unclass(object), ...)
    writeLines(" - attr(*, \"class\")= chr \"cset\"")
}

set_intersection <-
function(...)
    .set_intersection(list(...))

.set_intersection <-
function(l, matchfun = .exact_match)
{
    len <- length(l)
    if(len < 1L)
        set()
    else if(len < 2L)
        l[[1L]]
    else if(len < 3L) {
        i <- j <- 1
        if (length.set(l[[2L]]) < length.set(l[[1L]])) i <- 2 else j <- 2
        .make_set_from_list(.as.list(l[[i]])[unique(na.omit(matchfun(l[[j]],
                                                                     l[[i]])))])
    } else
        Recall(c(l[1L], list(Recall(l[-1L]))))
}

gset_intersection <-
function(...)
{
    l <- list(...)

    ## compute target support
    support <- do.call(set_intersection, l)

    ## handle empty support
    if (set_is_empty(support))
        return(set())

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, gset_is_set))))
        return(support)

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, .T.)
}

cset_intersection <-
function(...)
{
    l <- list(...)

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## compute target support using correct matchfuns
    support <- cset(.set_intersection(.as.list(l), matchfun),
                    orderfun, matchfun)

    ## handle empty supprt
    if (set_is_empty(support))
        return(set())

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, cset_is_set))))
        return(support)

    ## create gset by applying conorm, and then make cset
    .make_cset_from_gset_and_orderfun_and_matchfun(
         .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                 support,
                                                                 .T.,
                                                                 matchfun),
         orderfun,
         matchfun)
}

### * converters

### SET CONVERTERS
### We basically have make_set_with_order converters
### which return the resulting set and the original order
### to allow the permutation of some associated meta information
### according to the new order.
### (Examples: memberships of generalized sets and incidences of relations.)
### The as.set converter calls these internally and
### strips the ordering information.

make_set_with_order <-
function(x)
    UseMethod("make_set_with_order")

make_set_with_order.default <-
function(x)
    stop("Not implemented.")

make_set_with_order.NULL <-
function(x)
    make_set_with_order(list())

make_set_with_order.set <-
function(x)
    .make_set_with_order(x)

make_set_with_order.gset <-
function(x)
{
    attr(x, "memberships") <- NULL
    .make_set_with_order(.make_set_from_list(as.list(.get_support(x))))
}

make_set_with_order.cset <-
function(x)
    make_set_with_order(as.gset(x))

make_set_with_order.numeric <-
make_set_with_order.integer <-
function(x)
{
    x <- unique(x)
    O <- order(x)
    .make_set_with_order(.make_set_from_list(as.list(x[O])), O)
}

make_set_with_order.factor <-
make_set_with_order.character <-
make_set_with_order.logical <-
make_set_with_order.ordered <-
make_set_with_order.tuple <-
make_set_with_order.list <-
function(x)
{
    x <- .list_unique(x)
    O <- .list_order(x)
    .make_set_with_order(.make_set_from_list(x[O]), O)
}

make_set_with_order.matrix <-
make_set_with_order.data.frame <-
function(x) {
    x <- unique(x)
    O <- do.call(order, x)
    n <- rownames(x)
    if (is.null(n)) n <- seq_len(nrow(x))
    .make_set_with_order(.make_set_from_list(lapply(split(x, n), as.tuple)), O)
}

.make_set_with_order <-
function(set, order = seq_along(set))
    list(set = set, order = order)

### High-Level converter

as.set <-
function(x)
    make_set_with_order(x)$set

### canonicalizer

canonicalize_set_and_mapping <-
function(x, mapping = NULL, margin = NULL)
{
    x <- make_set_with_order(x)
    if (!is.null(mapping))
        mapping <- if (is.array(mapping) || is.data.frame(mapping)) {
            D <- dim(mapping)
            L <- length(x$set)
            if (is.null(margin))
                margin <- which(D == L)
            permute <- rep.int(list(seq_len(L)), length(D))
            permute[margin] <- rep.int(list(x$order), length(margin))
            do.call("[", c(list(mapping), permute, list(drop = FALSE)))
        } else
            mapping[x$order]

    list(set = x$set, mapping = mapping, order = x$order)
}

###

as.list.set <-
function(x, ...)
    .as.list(x)

as.vector.set <-
    function(x, mode = "any")
        as.vector(.as.list(x), mode = mode)

### gset converters

as.gset <-
function(x)
    UseMethod("as.gset")

as.gset.default <-
function(x)
    gset(x)

as.gset.cset <-
function(x)
{
    attr(x, "orderfun") <- NULL
    attr(x, "matchfun") <- NULL
    class(x) <- if (!is.null(attr(x, "memberships")))
        c("gset")
    else
        c("set", "gset")
    x
}

as.gset.gset <- identity

as.gset.tuple <-
function(x)
    as.gset(as.list(x))

as.gset.numeric <- function(x) .as.gset.atomic(x, as.numeric)
as.gset.character <- function(x) .as.gset.atomic(x, as.character)
as.gset.factor <- function(x) .as.gset.atomic(x, as.factor)
as.gset.ordered <- function(x) .as.gset.atomic(x, as.ordered)
as.gset.integer <- function(x) .as.gset.atomic(x, as.integer)
as.gset.logical <- function(x) .as.gset.atomic(x, as.logical)

.as.gset.atomic <-
function(x, FUN)
{
    tab <- table(x)
    .make_gset_from_support_and_memberships(FUN(c(names(tab), NA)),
                                            c(as.vector(tab), sum(is.na(x))))
}


as.gset.list <-
function(x)
{
    uni <- .list_sort(.list_unique(x))
    count <- table(.exact_match(x, uni))
    .make_gset_from_support_and_memberships(uni, count)
}

as.gset.data.frame <-
as.gset.matrix <-
function(x)
{
    n <- rownames(x)
    if (is.null(n)) n <- seq_len(nrow(x))
    as.gset(lapply(split(x, n), as.tuple))
}

as.list.gset <-
function(x, ...)
    .as.list(x)

as.vector.gset <-
function(x, mode = "any")
    as.vector(.make_list_of_elements_from_cset(x), mode = mode)

### tuple converters

as.tuple <-
function(x)
    UseMethod("as.tuple")

as.tuple.default <-
function(x)
    stop("Not implemented!")

as.tuple.tuple <- identity

as.tuple.numeric <-
as.tuple.factor <-
as.tuple.character <-
as.tuple.integer <-
as.tuple.ordered <-
as.tuple.logical <-
function(x)
    .make_tuple_from_list(as.list(x))

as.tuple.set <-
as.tuple.gset <-
as.tuple.cset <-
as.tuple.list <-
function(x)
    do.call(tuple, x)

as.tuple.data.frame <-
function(x)
{
    ret <- as.list(x)
    attributes(ret) <- NULL
    names(ret) <- colnames(x)
    .make_tuple_from_list(ret)
}

as.list.tuple <-
function(x, ...)
    unclass(x)

as.vector.tuple <-
    function(x, mode = "any")
        as.vector(unclass(x), mode = mode)

### cset converters

as.cset <-
function(x)
    UseMethod("as.cset")

as.cset.default <-
function(x)
    cset(gset(x))

as.cset.cset <-
    identity

as.cset.matrix <-
as.cset.data.frame <-
as.cset.logical <-
function(x)
    as.gset(x)

as.cset.tuple <-
as.cset.numeric <-
as.cset.factor <-
as.cset.character <-
as.cset.integer <-
as.cset.list <-
function(x)
{
    sup <- as.gset(x)
    if(!any(duplicated(x))) {
        o <- .list_order(x)
        i <- seq_along(x)
        i[o] <- i
        cset(sup, orderfun = i)
    } else sup
}

as.cset.ordered <-
function(x)
{
    s <- as.character(x)
    o <- .list_order(s)
    dup <- duplicated(s[o])
    cset(as.gset(x), orderfun = order(x)[o][!dup])
}

as.vector.cset <-
function(x, mode = "any")
    as.vector(.make_list_of_elements_from_cset(x), mode = mode)

as.list.cset <-
function(x, ...)
{
    FUN <- cset_orderfun(x)
    L <- .as.list(x)
    ms <- .get_memberships(L)
    if(is.function(FUN) || is.character(FUN)) {
        order <- do.call(FUN, list(L))
        L <- L[order]
        ms <- ms[order]
    } else if(is.integer(FUN) && (length(L) == length(FUN))) {
        L <- L[FUN]
        ms <- ms[FUN]
    }
    .structure(L, memberships = ms)
}

as.character.cset <-
function(x, ...)
{
    x <- as.list(x)
    fac <- unlist(lapply(x, is.factor))
    x[fac] <- lapply(x[fac], as.character)

    FUN <-
        function(x) paste(paste(list(args(x))),
                          paste(as.character(body(x)),
                                collapse = "\n"), sep = "\n")

    fun <- unlist(lapply(x, is.function))
    x[fun] <- lapply(x[fun], FUN)

    paste(x)
}

### make sure that as.list always works

as.list.function <-
function(x, ...)
    list(x)

### .as.list
### In the current implementation, for (g)gets, it's just unclass ...

.as.list <-
function(x, ...)
    UseMethod(".as.list")

.as.list.default <-
function(x, ...)
    as.list(x, ...)

.as.list.set <-
.as.list.gset <-
    unclass

.as.list.cset <-
function(x, ...)
{
    attr(x, "class") <- NULL
    attr(x, "orderfun") <- NULL
    attr(x, "matchfun") <- NULL
    x
}

### * converters

### SET CONVERTERS
### We basically have make_set_with_order converters
### which return the resulting set and the original order
### to allow the permutation of some associated meta information
### according to the new order.
### (Examples: memberships of generalized sets and incidences of relations.)
### The as.set converter calls these internally and
### strips the ordering information.

make_set_with_order <-
function(x)
    UseMethod("make_set_with_order")

make_set_with_order.default <-
function(x)
    stop("Not implemented.")

make_set_with_order.NULL <-
function(x)
    make_set_with_order(list())

make_set_with_order.set <-
function(x)
    .make_set_with_order(x)

make_set_with_order.gset <-
function(x)
{
    attr(x, "memberships") <- NULL
    .make_set_with_order(.make_set_from_list(as.list(.get_support(x))))
}

make_set_with_order.cset <-
function(x)
    make_set_with_order(as.gset(x))

make_set_with_order.numeric <-
make_set_with_order.integer <-
function(x)
{
    x <- unique(x)
    O <- order(x)
    .make_set_with_order(.make_set_from_list(as.list(x[O])), O)
}

make_set_with_order.factor <-
make_set_with_order.character <-
make_set_with_order.logical <-
make_set_with_order.ordered <-
make_set_with_order.tuple <-
make_set_with_order.list <-
function(x)
{
    x <- .list_unique(x)
    O <- .list_order(x)
    .make_set_with_order(.make_set_from_list(x[O]), O)
}

make_set_with_order.matrix <-
make_set_with_order.data.frame <-
function(x) {
    x <- unique(x)
    O <- do.call(order, x)
    n <- rownames(x)
    if (is.null(n)) n <- seq_len(nrow(x))
    .make_set_with_order(.make_set_from_list(lapply(split(x, n), as.tuple)), O)
}

.make_set_with_order <-
function(set, order = seq_along(set))
    list(set = set, order = order)

### High-Level converter

as.set <-
function(x)
    make_set_with_order(x)$set

### canonicalizer

canonicalize_set_and_mapping <-
function(x, mapping = NULL, margin = NULL)
{
    x <- make_set_with_order(x)
    if (!is.null(mapping))
        mapping <- if (is.array(mapping) || is.data.frame(mapping)) {
            D <- dim(mapping)
            L <- length(x$set)
            if (is.null(margin))
                margin <- which(D == L)
            permute <- rep.int(list(seq_len(L)), length(D))
            permute[margin] <- rep.int(list(x$order), length(margin))
            do.call("[", c(list(mapping), permute, list(drop = FALSE)))
        } else
            mapping[x$order]

    list(set = x$set, mapping = mapping, order = x$order)
}

###

as.list.set <-
function(x, ...)
    .as.list(x)

as.vector.set <-
    function(x, mode = "any")
        as.vector(.as.list(x), mode = mode)

### gset converters

as.gset <-
function(x)
    UseMethod("as.gset")

as.gset.default <-
function(x)
    gset(x)

as.gset.cset <-
function(x)
{
    attr(x, "orderfun") <- NULL
    attr(x, "matchfun") <- NULL
    class(x) <- if (!is.null(attr(x, "memberships")))
        c("gset")
    else
        c("set", "gset")
    x
}

as.gset.gset <- identity

as.gset.tuple <-
function(x)
    as.gset(as.list(x))

as.gset.numeric <- function(x) .as.gset.atomic(x, as.numeric)
as.gset.character <- function(x) .as.gset.atomic(x, as.character)
as.gset.factor <- function(x) .as.gset.atomic(x, as.factor)
as.gset.ordered <- function(x) .as.gset.atomic(x, as.ordered)
as.gset.integer <- function(x) .as.gset.atomic(x, as.integer)
as.gset.logical <- function(x) .as.gset.atomic(x, as.logical)

.as.gset.atomic <-
function(x, FUN)
{
    tab <- table(x)
    .make_gset_from_support_and_memberships(FUN(c(names(tab), NA)),
                                            c(as.vector(tab), sum(is.na(x))))
}


as.gset.list <-
function(x)
{
    uni <- .list_sort(.list_unique(x))
    count <- table(.exact_match(x, uni))
    .make_gset_from_support_and_memberships(uni, count)
}

as.gset.data.frame <-
as.gset.matrix <-
function(x)
{
    n <- rownames(x)
    if (is.null(n)) n <- seq_len(nrow(x))
    as.gset(lapply(split(x, n), as.tuple))
}

as.list.gset <-
function(x, ...)
    .as.list(x)

as.vector.gset <-
function(x, mode = "any")
    as.vector(.make_list_of_elements_from_cset(x), mode = mode)

### tuple converters

as.tuple <-
function(x)
    UseMethod("as.tuple")

as.tuple.default <-
function(x)
    stop("Not implemented!")

as.tuple.tuple <- identity

as.tuple.numeric <-
as.tuple.factor <-
as.tuple.character <-
as.tuple.integer <-
as.tuple.ordered <-
as.tuple.logical <-
function(x)
    .make_tuple_from_list(as.list(x))

as.tuple.set <-
as.tuple.gset <-
as.tuple.cset <-
as.tuple.list <-
function(x)
    do.call(tuple, x)

as.tuple.data.frame <-
function(x)
{
    ret <- as.list(x)
    attributes(ret) <- NULL
    names(ret) <- colnames(x)
    .make_tuple_from_list(ret)
}

as.list.tuple <-
function(x, ...)
    unclass(x)

as.vector.tuple <-
    function(x, mode = "any")
        as.vector(unclass(x), mode = mode)

### cset converters

as.cset <-
function(x)
    UseMethod("as.cset")

as.cset.default <-
function(x)
    cset(gset(x))

as.cset.cset <-
    identity

as.cset.matrix <-
as.cset.data.frame <-
as.cset.logical <-
function(x)
    as.gset(x)

as.cset.tuple <-
as.cset.numeric <-
as.cset.factor <-
as.cset.character <-
as.cset.integer <-
as.cset.list <-
function(x)
{
    sup <- as.gset(x)
    if(!any(duplicated(x))) {
        o <- .list_order(x)
        i <- seq_along(x)
        i[o] <- i
        cset(sup, orderfun = i)
    } else sup
}

as.cset.ordered <-
function(x)
{
    s <- as.character(x)
    o <- .list_order(s)
    dup <- duplicated(s[o])
    cset(as.gset(x), orderfun = order(x)[o][!dup])
}

as.vector.cset <-
function(x, mode = "any")
    as.vector(.make_list_of_elements_from_cset(x), mode = mode)

as.list.cset <-
function(x, ...)
{
    FUN <- cset_orderfun(x)
    L <- .as.list(x)
    ms <- .get_memberships(L)
    if(is.function(FUN) || is.character(FUN)) {
        order <- do.call(FUN, list(L))
        L <- L[order]
        ms <- ms[order]
    } else if(is.integer(FUN) && (length(L) == length(FUN))) {
        L <- L[FUN]
        ms <- ms[FUN]
    }
    .structure(L, memberships = ms)
}

as.character.cset <-
function(x, ...)
{
    x <- as.list(x)
    fac <- unlist(lapply(x, is.factor))
    x[fac] <- lapply(x[fac], as.character)

    FUN <-
        function(x) paste(paste(list(args(x))),
                          paste(as.character(body(x)),
                                collapse = "\n"), sep = "\n")

    fun <- unlist(lapply(x, is.function))
    x[fun] <- lapply(x[fun], FUN)

    paste(x)
}

### make sure that as.list always works

as.list.function <-
function(x, ...)
    list(x)

### .as.list
### In the current implementation, for (g)gets, it's just unclass ...

.as.list <-
function(x, ...)
    UseMethod(".as.list")

.as.list.default <-
function(x, ...)
    as.list(x, ...)

.as.list.set <-
.as.list.gset <-
    function(x, ...) unclass(x)

.as.list.cset <-
function(x, ...)
{
    attr(x, "class") <- NULL
    attr(x, "orderfun") <- NULL
    attr(x, "matchfun") <- NULL
    x
}

set_outer <-
function(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
{
    ## convenience
    nx <- deparse(substitute(X))
    if(missing(Y)) {
        Y <- X
        ny <- nx
    } else if(is.function(Y) || is.character(Y)) {
        FUN <- Y
        Y <- X
        ny <- nx
    } else ny <- deparse(substitute(Y))

    FUN <- match.fun(FUN)

    ## convert to lists
    xlabs <- LABELS(X, quote = quote)
    ylabs <- LABELS(Y, quote = quote)
    X <- as.list(X)
    Y <- as.list(Y)

    ## loop
    xrep <- rep.int(X, times = (ylen <- length(Y)))
    yrep <- rep(Y, each = (xlen <- length(X)))
    ret <- mapply(FUN, xrep, yrep, MoreArgs = list(...), SIMPLIFY = FALSE)

    ## simplify if sensible
    if(SIMPLIFY && all(sapply(ret, is.atomic)))
        ret <- unlist(ret, recursive = FALSE)

    ## make matrix
    dim(ret) <- c(xlen, ylen)
    dimnames(ret) <- list(xlabs, ylabs)
    ret
}

cset_outer <-
gset_outer <-
function(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
    set_outer(X = X, Y = Y, FUN = FUN, ..., SIMPLIFY = SIMPLIFY, quote = quote)

### * plot function
plot.gset <-
function(x, type = NULL, ylim = NULL,
         xlab = "Universe", ylab = "Membership Grade", ...)
{
    if (cset_is_empty(x, na.rm = TRUE))
        return(invisible(x))

    if (is.null(type))
        type <- if(cset_is_set_or_fuzzy_set(x, na.rm = TRUE) &&
                   .domain_is_numeric(x))
            "l"
        else
            "barplot"

    if (!cset_is_fuzzy_multiset(x) && (type != "barplot"))
        return(plot(tuple(x), ...))

    if (is.null(ylim) && !cset_is_multiset(x, na.rm = TRUE))
        ylim <- c(0,1)

    m <- cset_memberships(x)
    if (cset_is_fuzzy_multiset(x)) {
        maxlen <- max(lengths(m), na.rm = TRUE)
        m <- sapply(m, .expand_membership, len = maxlen)
    }
    barplot(m,
            ylim = ylim,
            names.arg = LABELS(x),
            beside = TRUE,
            ...)

    invisible(x)
}

plot.cset <-
function(x, ...)
    plot.gset(x, ...)

plot.set <-
function(x, ...)
{
    if (all(sapply(x, is.cset)))
        plot(as.tuple(x), ...)
    else
        plot.gset(x, ...)
}

plot.tuple <-
function(x, type = "l", ylim = NULL,
         xlab = "Universe", ylab = "Membership Grade", col = 1,
         continuous = TRUE, ...)
{
    l <- as.list(x)

    ## expand charfun generators
    for (i in seq_along(l))
        if (is.charfun_generator(l[[i]]))
            l[[i]] <- gset(charfun = l[[i]](), universe = .expand(NULL))

    ## check for fuzzy multisets
    ind <- sapply(l, cset_is_fuzzy_multiset)
    if (all(ind))
        stop("Cannot plot tuple of fuzzy multisets.")
    if (any(ind)) {
        warning("All fuzzy multisets ignored.")
        l <- l[!ind]
    }

    ## ylim
    ylim <- if (is.null(ylim) &&
                !any(sapply(l, cset_is_multiset, na.rm = TRUE)))
        c(0, 1)
    else
        c(0, max(unlist(lapply(l, .get_memberships)), na.rm = TRUE))

    ## find common domain
    names(l) <- NULL ## remove labels
    universe <- do.call(set_union,
                        c(lapply(l, cset_support),
                          lapply(lapply(l, cset_universe), as.set))
                        )
    NUMMODE <- .domain_is_numeric(universe)

    ## prepare plot region
    m <- if (NUMMODE) {
        universe <- cset(universe, matchfun = match)
        sort(unique(unlist(universe)))
    } else
        seq_along(universe)

    plot(m, rep.int(0, length(universe)),
         ylim = ylim,
         type = "n",
         axes = FALSE,
         xlab = xlab,
         ylab = ylab,
         ...)
    axis(2)
    if (NUMMODE)
        axis(1, labels = TRUE)
    else
        axis(1, at = m, labels = LABELS(universe))

    ## plot sets
    lines(as.tuple(l), type = type, col = col, universe = universe, ...)
    invisible(x)
}

plot.charfun_generator <-
function(x, universe = NULL, ...)
    plot(gset(charfun = x(), universe = .expand(universe)), ...)

## lines

lines.tuple <-
function(x, col = 1, universe = NULL, ...)
{
    l <- as.list(x)
    col <- rep_len(col, length.out = length(x))

    ## expand charfun generators
    for (i in seq_along(l))
        if (is.charfun_generator(l[[i]]))
            l[[i]] <- gset(l[[i]](), universe = .expand(NULL))

    ## check for fuzzy multisets
    ind <- sapply(l, cset_is_fuzzy_multiset)
    if (all(ind))
        stop("Cannot plot tuple of fuzzy multisets.")
    if (any(ind)) {
        warning("All fuzzy multisets ignored.")
        l <- l[!ind]
    }

    ## try to deduce universe
    names(l) <- NULL ## remove labels
    universe <- do.call(set_union,
                        c(lapply(l, cset_support),
                          lapply(lapply(l, cset_universe), as.set))
                        )

    ## call workhorse
    for (i in seq_along(l))
        lines(l[[i]], col = col[i], universe = universe, ...)

    invisible(x)
}

lines.set <-
function(x, ...)
{
    if (all(sapply(x, is.cset)))
        lines(as.tuple(x), ...)
    else
        lines.gset(x, ...)
}

lines.cset <-
function(x, ...)
    lines.gset(x, ...)

lines.gset <-
function(x, type = "l", col = 1, continuous = TRUE, universe = NULL, ...)
{
    universe <- .expand(universe)
    if (.domain_is_numeric(universe)) {
        ndom <- sort(unique(unlist(cset(universe, matchfun = match))))
        y <- rep.int(0, length(ndom))
        s <- cset(x, matchfun = match)
        .match <- matchfun(function(x, y) isTRUE(all.equal(x, y)))
        matches <- .match(unlist(s), ndom)
        isna <- is.na(matches)
        if(all(isna)) return(invisible(x))
        y[matches[!isna]] <- cset_memberships(s)[!isna]

        ## remove discontinuities
        if (continuous) {
            sup <- range(which(y > 0))
            zeros <- (y == 0) &
            seq_along(y) >= sup[1] &
                  seq_along(y) <= sup[2]
            y <- y[!zeros]
            ndom <- ndom[!zeros]
        }

        ## make sure that first and last values of support
        ## connect straight to bottom
        if (!type %in% c("p", "h")) {
            .insert <-
                function(X, i, value)
                    c(X[1:(i-1)], value, X[i:length(X)])

            ind <- min(which(y > 0))
            if (ind > 1L) {
                y <- .insert(y, ind, 0)
                ndom <- .insert(ndom, ind, ndom[ind])
            }

            ind <- max(which(y > 0))
            if (ind < length(y)) {
                y <- .insert(y, ind + 1, 0)
                ndom <- .insert(ndom, ind + 1, ndom[ind])
            }
        }

        lines(ndom, y, col = col, type = type, ...)
    } else {
        lines(.exact_match(cset_support(x), as.list(universe)),
              cset_memberships(x),
              col = col,
              type = type,
              ...)
    }

    invisible(x)
}

lines.charfun_generator <-
function(x, universe = NULL, ...)
    lines(gset(charfun = x(), universe = .expand(universe)), ...)

## set predicates

.is.listonly  <-
function(x) is.list(x) && is.null(attributes(x))

.is_list_of_sets <-
function(x) .is.listonly(x) && all(vapply(x, is.set, logical(1L)))

is.set <-
function(x)
    inherits(x, "set")

set_is_empty <-
function(x)
{
    if (is.cset(x))
        set_cardinality(x) < 1L
    else if (.is_list_of_sets(x))
        vapply(x, set_cardinality, integer(1L)) < 1L
    else stop("Argument must be a set or a list of sets.")
}

.vectorize_set_op <-
function(x, y, predicate)
{
    if (is.set(x) && is.set(y))
        predicate(x, y)
    else if(.is.listonly(x) && is.set(y))
        Vectorize(predicate)(x, list(y))
    else if(is.set(x) && .is.listonly(y))
        Vectorize(predicate)(list(x), y)
    else if(.is_list_of_sets(x) && .is_list_of_sets(y))
        Vectorize(predicate)(x, y)
    else stop("Predicate only implemented for sets or list of sets.")
}

set_is_subset <-
function(x, y)
{
    .help <- function(a, b) set_is_empty(set_complement(b, a))
    .vectorize_set_op(x, y, .help)

}

set_is_proper_subset <-
function(x, y)
{
    .help <- function(a, b) set_is_empty(set_complement(b, a)) && length(a) != length(b)
    .vectorize_set_op(x, y, .help)

}

set_is_equal <-
function(x, y)
{
    .help <- function(a, b)
        identical(a, b) ||
        ((length(a) == length(b))
         && (length(set_intersection(a,b)) == length(a)))
    .vectorize_set_op(x, y, .help)
}

set_contains_element <-
function(x, e)
{
    if (is_element(e) || is.tuple(e) || is.gset(e))
        e <- list(e)
    if (is.set(x))
        x <- list(x)
    if(!.is_list_of_sets(x))
        stop("Predicate only implemented for a set or a list of sets.")

   .help <- function(x, e) {
       if(set_is_empty(x))
           return(FALSE)
       if(is_element(e))
           e <- e[[1]]
       !is.na(.exact_match(list(e), x))
    }

    Vectorize(.help)(x, e)
}

## gset-predicates

.is_list_of_gsets <-
function(x) .is.listonly(x) && all(vapply(x, is.gset, logical(1L)))

is_element <-
function(e)
    inherits(e, "gset_element")

is.gset <-
function(x)
    inherits(x, c("gset", "set"))

gset_is_empty <-
function(x, na.rm = FALSE)
{
    if (is.cset(x))
        gset_cardinality(x, na.rm = na.rm) == 0
    else if (.is_list_of_gsets(x))
        vapply(x, gset_cardinality, double(1L), na.rm = na.rm) == 0
    else stop("Argument must be a gset or a list of gsets.")
}

.vectorize_gset_op <-
function(x, y, predicate)
{
    if(is.gset(x) && is.gset(y))
        predicate(x, y)
    else if(.is.listonly(x) && is.gset(y))
        Vectorize(predicate)(x, list(y))
    else if(is.gset(x) && .is.listonly(y))
        Vectorize(predicate)(list(x), y)
    else if(.is_list_of_gsets(x) && .is_list_of_gsets(y))
        Vectorize(predicate)(x, y)
    else stop("Arguments must be gsets or list of gsets.")
}

gset_is_subset <-
function(x, y, na.rm = FALSE)
{
    .help <-
        function(a, b) set_is_subset(gset_support(a), gset_support(b)) &&
    all(unlist(.apply_connector_to_list_of_gsets_and_support(list(a, b),
                                                             .get_support(a),
                                                             `<=`)),
        na.rm = na.rm)
    .vectorize_gset_op(x, y, .help)

}

gset_is_proper_subset <-
function(x, y, na.rm = FALSE)
{
    .help <- function(x, y) gset_is_subset(x, y, na.rm = na.rm) &&
        cset_cardinality(x, na.rm = na.rm) != cset_cardinality(y, na.rm = na.rm)
    .vectorize_gset_op(x, y, .help)
}

gset_is_equal <-
function(x, y, na.rm = FALSE)
{
    .help <- function(a, b)
        identical(a, b) ||
        cset_cardinality(a, na.rm = na.rm) ==
            cset_cardinality(b, na.rm = na.rm) &&
            cset_cardinality(gset_intersection(a, b), na.rm = na.rm) ==
            cset_cardinality(a, na.rm = na.rm)
    .vectorize_gset_op(x, y, .help)

}

gset_contains_element <-
function(x, e) {
    if (is.tuple(e) || is.gset(e))
        e <- list(list(e))
    if (is_element(e))
        e <- list(gset(elements = list(e)))
    if (is.gset(x))
        x <- list(x)
    if(!.is_list_of_gsets(x))
        stop("Predicate only implemented for a gset or a list of gsets.")

    .help <- function(x, e) gset_is_subset(as.gset(e), x)

    Vectorize(.help)(x, e)
}

.vectorize_unary_gset_op <-
function(x, predicate)
{
    if (is.cset(x))
        predicate(x)
    else if (.is_list_of_csets(x))
        vapply(x, predicate, logical(1L))
    else stop("Predicate only implemented for a [cg]set or a list of [cg]sets.")
}

gset_is_multiset <-
function(x, na.rm = FALSE)
{
    .help <- function(x) {
        m <- .get_memberships(x)
        !is.list(m) && (na.rm && any(m > 1, na.rm = TRUE) ||
                        !na.rm && any(m > 1) && !any(m < 1, na.rm = TRUE))
    }
    .vectorize_unary_gset_op(x, .help)
}

gset_is_crisp <-
gset_is_set_or_multiset <-
function(x, na.rm = FALSE)
{
    .help <- function(x) {
        m <- .get_memberships(x)
        !is.list(m) && (na.rm && all(m >= 1, na.rm = TRUE) ||
                        !na.rm && all(m >= 1) && !any(m < 1, na.rm = TRUE))
    }
    .vectorize_unary_gset_op(x, .help)

}

gset_is_fuzzy_set <-
function(x, na.rm = FALSE)
{
    .help <- function(x) {
        m <- .get_memberships(x)
        !is.list(m) && (na.rm && any(m < 1, na.rm = TRUE) ||
                        !na.rm && any(m < 1) && !any(m > 1, na.rm = TRUE))
    }
    .vectorize_unary_gset_op(x, .help)
}

gset_is_set_or_fuzzy_set <-
function(x, na.rm = FALSE)
{
    .help  <- function(x) {
        m <- .get_memberships(x)
        !is.list(m) && (na.rm && all(m <= 1, na.rm = TRUE) ||
                        !na.rm && all(m <= 1) && !any(m > 1, na.rm = TRUE))
    }
    .vectorize_unary_gset_op(x, .help)

}

gset_is_fuzzy_multiset <-
function(x)
{
    .help <- function(x) is.list(.get_memberships(x))
    .vectorize_unary_gset_op(x, .help)

}

gset_is_set <-
function(x, na.rm = FALSE)
{
    .help <- function(x) {
        m <- .get_memberships(x)
        !is.list(m) && all(m == 1, na.rm = na.rm)
    }
    .vectorize_unary_gset_op(x, .help)

}

gset_has_missings <-
function(x)
{
    .help <- function(x) {
        M <- .get_memberships(x)
        if (is.list(M))
            any(is.na(unlist(M))) || any(is.na(unlist(lapply(M, .get_memberships))))
        else
            any(is.na(M))
    }
    .vectorize_unary_gset_op(x, .help)
}

## cset-predicates

.is_list_of_csets <-
function(x) .is.listonly(x) && all(vapply(x, is.cset, logical(1L)))

is.cset <-
function(x)
    inherits(x, c("cset", "gset", "set"))

cset_is_empty <-
function(x, na.rm = FALSE)
{
    if(is.cset(x))
        cset_cardinality(x, na.rm = na.rm) == 0
    else
        all(vapply(x, cset_cardinality, double(1L), na.rm = na.rm) == 0)
}

.vectorize_cset_op <-
function(x, y, predicate)
{
    if(is.cset(x) && is.cset(y))
        predicate(x, y)
    else if(.is.listonly(x) && is.cset(y))
        Vectorize(predicate)(x, list(y))
    else if(is.cset(x) && .is.listonly(y))
        Vectorize(predicate)(list(x), y)
    else if(.is_list_of_csets(x) && .is_list_of_csets(y))
        Vectorize(predicate)(x, y)
    else stop("Comparison not implemented.")
}

cset_is_subset <-
function(x, y, na.rm = FALSE)
{
    .help <-
        function(a, b) {
            .check_matchfun(list(a, b))
            set_is_empty(.set_complement_using_matchfun(cset_support(b),
                                                       cset_support(a),
                                                       cset_matchfun(a))) &&
    all(unlist(.apply_connector_to_list_of_gsets_and_support(list(a, b),
                                                             .get_support(a),
                                                             `<=`,
                                                             cset_matchfun(a))),
        na.rm = na.rm
        )
    }
    .vectorize_cset_op(x, y, .help)

}

cset_is_proper_subset <-
function(x, y, na.rm = FALSE)
{
    .help <- function(a, b)
         cset_is_subset(a, b, na.rm = na.rm) &&
             cset_cardinality(x, na.rm = na.rm) !=
             cset_cardinality(y, na.rm = na.rm)
    .vectorize_cset_op(x, y, .help)
}

cset_is_equal <-
function(x, y, na.rm = FALSE)
{
    .help <- function(a, b)
        identical(a, b) ||
        cset_cardinality(a, na.rm = na.rm) ==
            cset_cardinality(b, na.rm = na.rm) &&
    cset_cardinality(cset_intersection(a, b), na.rm = na.rm) ==
        cset_cardinality(a, na.rm = na.rm)
    .vectorize_cset_op(x, y, .help)

}

cset_contains_element <-
function(x, e)
{
    if(cset_is_empty(x))
        return(FALSE)
    matchfun <- cset_matchfun(x)
    x <- .make_list_of_elements_from_cset(x)
    e <- e(e)
    if(is.tuple(e) || is.cset(e) || is_element(e))
        e <- list(e)
    ind <- matchfun(e, x)
    if(is.na(ind)) return(FALSE)
    .get_memberships(x[[ind]]) == .get_memberships(e[[1]])
}

cset_is_multiset <-
function(x, na.rm = FALSE)
    gset_is_multiset(x, na.rm = na.rm)

cset_is_crisp <-
cset_is_set_or_multiset <-
function(x, na.rm = FALSE)
    gset_is_crisp(x, na.rm = na.rm)

cset_is_fuzzy_set <-
function(x, na.rm = FALSE)
    gset_is_fuzzy_set(x, na.rm = na.rm)

cset_is_set_or_fuzzy_set <-
function(x, na.rm = FALSE)
    gset_is_set_or_fuzzy_set(x, na.rm = na.rm)

cset_is_fuzzy_multiset <-
function(x)
    gset_is_fuzzy_multiset(x)

cset_is_set <-
function(x, na.rm = FALSE)
    gset_is_set(x, na.rm = na.rm)

cset_has_missings <-
function(x)
    gset_has_missings(x)

## contains_element operator dispatch

"%e%" <-
function(e, x)
{
    if(is.set(x) || .is_list_of_sets(x))
        set_contains_element(x, e)
    else if(is.gset(x) || .is_list_of_gsets(x))
        gset_contains_element(x, e)
    else if(is.cset(x) || .is_list_of_csets(x))
        cset_contains_element(x, e)
    else if(is.interval(x))
        interval_contains_element(x, e)
    else
        stop("Predicate undefined.")
}

set_combn <-
function(x, m)
{
    if (m == 0)
        set()
    else
        do.call(set, apply(combn(.as.list(x), m), 2L, as.set))
}

gset_combn <-
function(x, m)
{
    if (m == 0L)
        gset()
    else {
        support <- apply(combn(.as.list(x), m), 2L, as.set)
        memberships <- unlist(apply(combn(.get_memberships(x), m), 2L, list),
                              recursive = FALSE)
        do.call(set, Map(gset, support, memberships))
    }
}

cset_combn <-
function(x, m)
{
    ## get matchfun and orderfun
    matchfun <- .matchfun(x)
    orderfun <- .orderfun(x)
    if (is.integer(orderfun))
        orderfun <- NULL

    if (m == 0L)
        gset()
    else {
        support <- apply(combn(.as.list(x), m), 2L, as.set)
        memberships <- unlist(apply(combn(.get_memberships(x), m), 2L, list),
                              recursive = FALSE)
        do.call(set,
                lapply(Map(gset, support, memberships),
                       cset, orderfun, matchfun)
                       )
    }
}

gset_product <-
function(...)
{
    l <- list(...)

    ## compute target support
    support <- do.call(set_intersection, l)
    if (set_is_empty(support))
        return(set())

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, gset_is_set))))
        return(support)

    ## apply connector
    .make_gset_from_list_of_gsets_and_support_and_connector(l, support, `*`)
}

cset_product <-
function(...)
{
    l <- list(...)

    ## check matchfun and orderfun
    matchfun <- .check_matchfun(l)
    orderfun <- .check_orderfun(l)

    ## compute target support using correct matchfuns
    support <- cset(.set_intersection(.as.list(l), matchfun),
                    orderfun, matchfun)
    if (isTRUE(cset_is_empty(support)))
        return(set())

    ## handle the "ordinary set" case
    if (isTRUE(all(sapply(l, cset_is_set))))
        return(support)

    ## create gset by applying conorm, and then make cset
    .make_cset_from_gset_and_orderfun_and_matchfun(
         .make_gset_from_list_of_gsets_and_support_and_connector(l,
                                                                 support,
                                                                 `*`,
                                                                 matchfun),
         orderfun,
         matchfun)
}

cut.gset <-
function(x, level = 1, type = c("alpha", "nu"), strict = FALSE, ...)
{
    type <- match.arg(type)

    support <- .as.list(.get_support(x))
    memberships <- .get_memberships(x)

    if (type == "alpha") {
        if (gset_is_crisp(x, na.rm = TRUE))
            return(x)

        if (!gset_is_fuzzy_multiset(x)) {
            ind <- if (strict)
                memberships > level
            else
                memberships >= level

            if (length(ind) > 0L)
                .make_gset_from_support_and_memberships(support, as.double(ind))
            else
                set()
        } else {
            M <- sapply(memberships, if (strict)
                        function(i) sum(as.double(unlist(i) > level) * .get_memberships(i))
            else
                        function(i) sum(as.double(unlist(i) >= level) * .get_memberships(i))
                        )
            ind <- is.na(M) | M > 0
            .make_gset_from_support_and_memberships(support[ind], M[ind])
        }
    } else {
        if (gset_is_set_or_fuzzy_set(x, na.rm = TRUE)) {
            if (level > 1)
                set()
            else
                x
        } else if (!gset_is_fuzzy_multiset(x)) {
            ind <- if (strict)
                memberships > level
            else
                memberships >= level

            if (length(ind) > 0L)
                .make_gset_from_support_and_memberships(support, as.double(ind))
            else
                set()

        } else {
            M <- sapply(memberships, function(i) {
                m <- .expand_membership(i)
                if (isTRUE(m[level + strict] == 1))
                    1
                else if (any(is.na(m)))
                    NA
                else if (length(m) < level + strict)
                    0
                else m[level + strict]
            })
            .make_gset_from_support_and_memberships(support, M)
        }
    }
}

cut.cset <-
function(x, level = 1, type = c("alpha", "nu"), strict = FALSE, ...)
    cut.gset(x, level = level, ...)

all.equal.set <-
function(target, current, ...)
{
    if (gset_is_equal(target, current, na.rm = TRUE))
        return(TRUE)

    msg <- NULL
    if (data.class(target) != data.class(current)) {
        msg <- paste("target is ", data.class(target), ", current is ",
                     data.class(current), sep = "")
    } else {
        if (isTRUE((lt <- length(target)) != (lc <- length(current))))
            msg <- paste("cardinality of target is ", lt, ", of current is ",
                         lc, sep = "")
        if (isTRUE((lc <- length(gset_difference(current, target))) > 0))
            msg <- c(msg, paste(lc, "elements of current are not in target"))
        if (isTRUE((lt <- length(gset_difference(target, current))) > 0))
            msg <- c(msg, paste(lt, "elements of target are not in current"))
    }
    msg
}

all.equal.gset <-
function(target, current, ...)
{
    if (gset_is_equal(target, current, na.rm = TRUE))
        return(TRUE)
    if (isTRUE(all.equal(as.list(target), as.list(current)))
        && isTRUE(all.equal(gset_memberships(target), gset_memberships(current))))
        return(TRUE)

    if (data.class(target) != data.class(current))
        paste("target is ", data.class(target), ", current is ",
              data.class(current), sep = "")
    else if (gset_is_fuzzy_set(target, na.rm = TRUE) &&
             gset_is_multiset(current, na.rm = TRUE))
        "target is a fuzzy set, current is a multiset"
    else if (gset_is_fuzzy_set(current, na.rm = TRUE) &&
             gset_is_multiset(target, na.rm = TRUE))
        "current is a fuzzy set, target is a multiset"
    else if (!isTRUE(all.equal(lt <- gset_cardinality(target, na.rm = TRUE),
             lc <- gset_cardinality(current, na.rm = TRUE))))
        paste("cardinality of target is ", lt, ", of current is ",
              lc, sep = "")
    else
        "target and current have different memberships"
}

all.equal.cset <-
function(target, current, ...)
{
    if (cset_is_equal(target, current, na.rm = TRUE))
        return(TRUE)

    msg <- all.equal.gset(target, current, ...)
    if (isTRUE(msg)) msg <- ""
    if (!identical(.orderfun(target), .orderfun(current)))
        msg <- c(msg, "order functions of target and current differ")
    if (!identical(.matchfun(target), .matchfun(current)))
        msg <- c(msg, "match functions of target and current differ")
    msg
}

set_symdiff <-
function(...)
{
    len <- length(l <- list(...))
    SD <- function(x, y) set_union(set_complement(x, y), set_complement(y, x))

    if (len < 1L)
        gset()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(SD, l)
}

gset_symdiff <-
function(...)
{
    len <- length(l <- list(...))
    SD <- function(x, y) gset_union(gset_complement(x, y), gset_complement(y, x))
    if (len < 1L)
        gset()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(SD, l)
}

cset_symdiff <-
function(...)
{
    len <- length(l <- list(...))
    SD <- function(x, y) cset_union(cset_complement(x, y), cset_complement(y, x))
    if (len < 1L)
        set()
    else if (len < 2L)
        l[[1L]]
    else
        Reduce(SD, l)
}

"%D%" <-
function(x, y)
{
    if (!is.cset(x)) x <- as.cset(x)
    if (!is.cset(y)) y <- as.cset(y)
    if (is.set(x) && is.set(y))
        set_symdiff(x, y)
    else if (is.gset(x) && is.gset(y))
        gset_symdiff(x, y)
    else if (is.cset(x) && is.cset(y))
        cset_symdiff(x, y)
    else
        stop("Operator only defined for (g,c)sets.")
}

cset_defuzzify <-
gset_defuzzify <-
function(x, method = c("meanofmax", "smallestofmax", "largestofmax", "centroid"))
    switch(match.arg(method),
           centroid = mean(x),
           smallestofmax = min(gset_peak(x)),
           meanofmax = mean(gset_peak(x)),
           largestofmax = max(gset_peak(x))
           )

### Membership transformations

gset_transform_memberships <-
function(x, FUN, ...)
{
    F <- function(x) pmax(0, pmin(1, FUN(x, ...)))
    m <- if (cset_is_set_or_multiset(x, na.rm = TRUE))
        lapply(.get_memberships(x), function(i) gset(F(1), i))
    else if (!cset_is_fuzzy_multiset(x))
        F(.get_memberships(x))
    else
        lapply(.get_memberships(x), F)
    .set_memberships(x, m)
}

cset_transform_memberships <-
function(x, FUN, ...)
    gset_transform_memberships(x, FUN, ...)

cset_concentrate <-
gset_concentrate <-
function(x)
    gset_transform_memberships(x, function(i) i * i)

cset_dilate <-
gset_dilate <-
function(x)
    gset_transform_memberships(x, sqrt)

cset_normalize <-
gset_normalize <-
function(x, height = 1)
{
    if (height < 0 || height > 1)
        stop("Height must be in the unit interval.")
    gset_transform_memberships(x, function(i) height * i /
                               max(.get_memberships(x), na.rm = TRUE))
}

## internal functions to handle memberships

.expand_membership <-
function(x, decreasing = TRUE, len = NA, rep = TRUE)
{
    if (is.null(x))
        x <- 0
    if (!is.atomic(x))
        x <- .as.list(x)
    M <- .get_memberships(x)

    ## create vector from gset
    ret <- if (rep) {
        if (is.list(x))
            rep.int(as.numeric(x), M)
        else if (length(x) == 1L && !is.na(x) && x > 1)
            rep.int(1, x)
        else
            x
    } else {
        as.numeric(x)
    }

    ## optionally, fill up with 0s
    if (!is.na(len)) {
        ret <- c(ret, rep.int(0, len - length(ret)))
        if (!rep)
            M <- c(M, rep.int(0, len - length(M)))
    }

    ## optionally, sort values
    if (!is.null(decreasing)) {
        O <- order(ret, decreasing = decreasing)
        ret <- ret[O]
        if (!rep) M <- M[O]
    }

    if (rep)
        ret
    else
        .structure(ret, memberships = M)
}

.memberships_for_support <-
function(x, support, matchfun = .exact_match)
{
    m <- matchfun(support, .get_support(x))
    tmp <- if (all(is.na(m)))
        m
    else
        .get_memberships(x)[m]
    tmp[is.na(m)] <- if (is.list(tmp)) list(set(0)) else 0
    tmp
}

.canonicalize_memberships <-
function(memberships)
{
    if (is.list(memberships))
        lapply(memberships, as.gset)
    else
        memberships
}

library(sets)

# section 3 examples

s <- set(1L, 2L, 3L)
s2 <- as.set(2:4)

test_that("as.set", {
  expect_equal(s2, set(2L, 3L, 4L))
})

test_that("empty set", {
  expect_true(set_is_empty(set()))
})

test_that("less than or equal to", {
  expect_lte(set(1), set(1,2))
})

test_that("length", {
  expect_equal(length(s), 3)
})

test_that("union", {
  expect_equal(s | set("a"), set("a", 1L, 2L, 3L))
})

test_that("intersection", {
  expect_equal(s & s2, set(2L, 3L))
})

test_that("symmetric difference", {
  expect_equal(s %D% s2, set(1L, 4L))
})

test_that("Cartesian product", {
  expect_equal(s * s2,
               set(tuple(1L, 2L), tuple(1L, 3L), tuple(1L, 4L),
                   tuple(2L, 2L), tuple(2L, 3L), tuple(2L, 4L),
                    tuple(3L, 2L), tuple(3L, 3L), tuple(3L, 4L)))
  expect_equal(s ^ 2L,
               set(tuple(1L, 1L), tuple(1L, 2L), tuple(1L, 3L),
                   tuple(2L, 1L), tuple(2L, 2L), tuple(2L, 3L),
                   tuple(3L, 1L), tuple(3L, 2L), tuple(3L, 3L)))
})

test_that("power set", {
  expect_equal(2 ^ s,
               set(set(), set(1L), set(2L), set(3L), set(1L, 2L),
                   set(1L, 3L), set(2L, 3L), set(1L, 2L, 3L)))
})

test_that("relative complement", {
  expect_equal(set_complement(set(1), set(1, 2, 3)), set(2, 3))
})

test_that("absolute complement", {
  expect_equal(!set(1), set())
})

test_that("subsets", {
  expect_equal(set_combn(s, 2L),
               set(set(1L, 2L), set(1L, 3L), set(2L, 3L)))
})

test_that("closure and reduction", {
  cl <- closure(set(set(1), set(2), set(3)), "union")
  expect_equal(cl, set(set(1), set(2), set(3), set(1, 2), set(1, 3),
                       set(2, 3), set(1, 2, 3)))
  expect_equal(reduction(cl, "union"), set(set(1), set(2), set(3)))
})

test_that("sum and range", {
  expect_equal(sum(s), 6)
  expect_equal(range(s), c(1, 3))
})

test_that("subscripting", {
  s2 <- set(1,2, c, list(1,2))
  s2[[c]] <- "foo"
  s2[[list(1, 2)]] <- "bar"
  expect_equal(s2[list("foo", 1)], set("foo", 1))
})

test_that("sapply", {
  expect_equal(sapply(s, sqrt), c(1, sqrt(2), sqrt(3)))
})

test_that("set_outer", {
  expect_true(all(set_outer(set(1, 2), set(1, 2, 3), "/") ==
                    matrix(c(1, 2, 1/2, 1, 1/3, 2/3), ncol = 3)))
})

# section 4 examples

test_that("defining generalized sets", {
  X <- c("A", "B", "C")
  expect_equal(gset(support = X), set("A", "B", "C"))
  expect_equal(gset(support = X, universe = LETTERS[1:10]),
               set("A", "B", "C"))
  ms <- c(0.1, 0.3, 1)
  expect_equal(gset(support = X, memberships = ms),
               gset(elements =
                      list(e("A", 0.1), e("B", 0.3), e("C", 1))))
  f <- function(x) switch(x, A = 0.1, B = 0.2, C = 1, 0)
  expect_equal(gset(universe = X, charfun = f),
               gset(elements =
                      list(e("A", 0.1), e("B", 0.2), e("C", 1))))
})

test_that("generalized set operations", {
  X <- gset(c("A", "B", "C"), 4:6)
  Y <- gset(c("B", "C", "D"), 1:3)
  expect_equal(X | Y, gset(c("A", "B", "C", "D"), c(4, 5, 6, 3)))
  expect_equal(X & Y, gset(c("B", "C"), c(1, 2)))
  expect_equal(X + Y, gset(c("A", "B", "C", "D"), c(4, 6, 8, 3)))
  expect_equal(X - Y, gset(c("A", "B", "C"), c(4, 4, 4)))
  expect_equal(gset_product(X, Y), gset(c("B", "C"), c(5, 12)))
})

test_that("generalized set complements", {
  expect_equal(!gset(1, 0.3), gset(1, 0.7))
  X <- gset("a", universe = letters[1:3])
  expect_equal(!X, set("b", "c"))
  expect_equal(!!X, set("a"))
  expect_equal(!gset(1L, 2, universe = 1:3, bound = 3),
               gset(c(1L, 2L, 3L), c(1, 3, 3)))
})

test_that("generalized set means", {
  x <- gset(1:3, 1:3/3)
  y <- gset(1:2, 1:2/2)
  expect_equal(gset_mean(x, y),
               gset(c(1L, 2L, 3L), c(5/12, 5/6, 1/2)))
  expect_equal(gset_mean(x, y, "harmonic"),
               gset(c(1L, 2L), c(0.4, 0.8)))
  expect_equal(gset_mean(set(1), set(1, 2)),
               gset(c(1, 2), c(1, 0.5)))
})

test_that("transform memberships", {
  x <- gset(1:10, 1:10/10)
  expect_equal(gset_transform_memberships(x, pmax, 0.5),
               gset(1:10, c(rep(0.5, 5), 0.6, 0.7, 0.8, 0.9, 1)))
  x <- gset(1, 2)
  expect_equal(rep(x, 0.5), set(1))
  y <- gset(1:2, 1:2/2)
  expect_equal(gset_dilate(y), gset(1:2, c(1 / sqrt(2), 1)))
  expect_equal(gset_concentrate(y), gset(1:2, c(0.25, 1)))
  expect_equal(gset_normalize(y, 0.5), gset(1:2, c(0.25, 0.5)))
})

test_that("fuzzy logic", {
  x <- 1:10 / 10
  y <- rev(x)
  expect_equal(.S.(x, y),
               c(1.0, 0.9, 0.8, 0.7, 0.6, 0.6, 0.7, 0.8, 0.9, 1.0))
  fuzzy_logic("Fodor")
  expect_equal(.S.(x,y), rep(1, 10))
  X <- gset(c("A", "B", "C"), c(0.3, 0.5, 0.8))
  Y <- gset(c("B", "C", "D"), c(0.1, 0.3, 0.9))
  fuzzy_logic("Zadeh")
  expect_equal(X & Y, gset(c("B", "C"), c(0.1, 0.3)))
  expect_equal(X | Y,
               gset(c("A", "B", "C", "D"), c(0.3, 0.5, 0.8, 0.9)))
  expect_equal(gset_complement(X, Y), gset(c("B", "C", "D"), c(0.1, 0.2, 0.9)))
  fuzzy_logic("Fodor")
  expect_equal(X & Y, gset("C", 0.3))
  expect_equal(X | Y,
               gset(c("A", "B", "C", "D"), c(0.3, 0.5, 1, 0.9)))
  expect_equal(gset_complement(X, Y), gset("D", 0.9))
  expect_equal(cut(X, 0.5), set("B", "C"))
})

\documentclass[article,nojss]{jss}
\usepackage{amsfonts,enumerate}

\pdfsuppressptexinfo=-1

\author{David Meyer\\FH Technikum Wien \And
        Kurt Hornik\\WU Wirtschaftsuniversit{\"a}t Wien}
\Plainauthor{David Meyer, Kurt Hornik}

\title{Generalized and Customizable Sets in \proglang{R}}
\Plaintitle{Generalized and Customizable Sets in R}


\Abstract{
  This introduction to the \proglang{R} package \pkg{sets} is a (slightly)
  modified version of \cite{sets:meyer+hornik:2009}, published in the
  \emph{Journal of Statistical Software}.

  We present data structures and algorithms for sets and some
  generalizations thereof (fuzzy sets, multisets, and fuzzy multisets)
  available for {\sf R} through the \pkg{sets} package. Fuzzy
  (multi-)sets are based on dynamically bound fuzzy logic families.
  Further extensions include user-definable iterators and matching
  functions.
}

\Keywords{\proglang{R}, set, fuzzy logic, multiset, fuzzy set}
\Plainkeywords{R, set, fuzzy logic, multiset, fuzzy set}

%\VignetteIndexEntry{Generalized and Customizable Sets in R}
%\VignetteDepends{sets,proxy}
%\VignetteKeywords{R, set, fuzzy logic, multiset, fuzzy set}
%\VignettePackage{sets}

\Address{
  David Meyer\\
  Department of Information Systems and Operations\\
  WU Wirtschaftsuniversit\"at Wien\\
  Augasse 2--6\\
  1090 Wien, Austria
  E-mail: \email{David.Meyer@wu.ac.at}\\
  URL: \url{http://wi.wu.ac.at/~meyer/}\\

  Kurt Hornik\\
  Department of Statistics and Mathematics\\
  WU Wirtschaftsuniversit\"at Wien\\
  Augasse 2--6\\
  1090 Wien, Austria
  E-mail: \email{Kurt.Hornik@wu.ac.at}\\
  URL: \url{http://statmath.wu.ac.at/~hornik/}\\
}

%% need no \usepackage{Sweave.sty}
\SweaveOpts{engine=R,eps=FALSE,results=verbatim,fig=FALSE,echo=TRUE}
\setkeys{Gin}{width=0.7\textwidth}

<<preliminaries,echo=FALSE,results=hide>>=
library("sets")
options(prompt = "R> ", continue = "+  ", width = 77)
@

\newcommand{\var}[1]{\textit{\texttt{#1}}}
\newcommand{\data}[1]{\texttt{#1}}
\newcommand{\class}[1]{`\texttt{#1}'}

\newcommand{\comment}[1]{\begin{quote} COMMENT. #1\end{quote}}

%% \code without `-' ligatures
\def\nohyphenation{\hyphenchar\font=-1 \aftergroup\restorehyphenation}
\def\restorehyphenation{\hyphenchar\font=`-}
{\catcode`\-=\active%
  \global\def\code{\bgroup%
    \catcode`\-=\active \let-\codedash%
    \Rd@code}}
\def\codedash{-\discretionary{}{}{}}
\def\Rd@code#1{\texttt{\nohyphenation#1}\egroup}

%% FIXME
\newcommand{\samp}[1]{\normalfont\texttt{#1}}

\newcommand{\eqn}[1]{$#1$}
\newcommand{\codefun}[1]{\code{#1()}}
\newcommand{\codefunind}[1]{\codefun{#1}\index{\texttt{#1}}}
\newcommand{\codeind}[1]{\code{#1}\index{\texttt{#1}}}
\newcommand{\sQuote}[1]{`{#1}'}
\newcommand{\dQuote}[1]{``{#1}''}

\begin{document}

\section{Introduction}

Only few will deny the importance of sets and set theory, building the
fundamentals of modern mathematics. For theory-building typically
axiomatic approaches \citep[e.g.,][]{sets:zermelo:1908, sets:fraenkel:1922}
are used. However, even
the primal, ``naive'' concept of sets representing ``collections
of distinct objects'' \citep{sets:cantor:1895} discarding order
and count information seems both natural and practical.
The main operation being ``is-element-of'', sets alone are of limited
\emph{practical} use---they most of the times serve as basic building blocks for
more complex structures such as relations and generalized sets. A
common way is to consider pairs $(X, m)$ with set $X$ (``universe'')
and membership function
$m : X \rightarrow D$ mapping each member to its ``grade''. The subset
of $X$ of elements with non-zero membership is called ``support''.
In \emph{multisets}, elements may appear more than once, i.e., $D =
\mathbb{N}_0$ ($m$ is also called the multiplicity function). There are many
applications in computer science and other disciplines
\citep[for a survey, see, e.g.,][]{sets:singh+ibrahim+yohanna:2007}.
In statistics, multisets appear as frequency tables. \emph{Fuzzy sets}
have become quite popular since their introduction by
\cite{sets:zadeh:1965}. Here, the membership function maps into the
unit interval. An interesting characteristic of fuzzy sets is that
the actual behavior of set operations depends on the underlying
fuzzy logic employed, which can be chosen according to domain-specific
needs. Fuzzy sets are actively used in fields such as machine
learning, engineering, medical science, and artificial intelligence
\citep{sets:dubois+prade+yager:1996}.
\emph{Fuzzy multisets} \citep{sets:yager:1986}
combine both approaches by allowing
each element to map to more than one fuzzy membership grade, i.e., $D$ is
the power set of multisets over the unit interval. Examples for the
application of fuzzy multisets can be found in the field of information
retrieval \citep[e.g.,][]{sets:matthe+caluwe+tre:2006}.

The use of sets and variants thereof is common in modern general purpose
programming languages:  \proglang{Java} and \proglang{C++}
provide corresponding abstract data types (ADTs) in their class
libraries, \proglang{Pascal} and \proglang{Python}
offer sets as native data type.
Indeed, since set elements are order-invariant and unique,
lookup mechanisms can be implemented very efficiently (for example via
hashing, resulting in nearly constant run-time complexity, compared to linear
search, requiring $n/2$ steps on the average for $n$ elements).
Surprisingly enough, sets are not standard in many
mathematical programming environments such as \proglang{MATLAB} and
\proglang{Mathematica}, and also \proglang{R}. Although the two latter offer
set operations such as union and intersection, these are applied to
linearly indexable structures (lists and vectors, respectively),
\emph{interpreting} them as sets. When it comes to \proglang{R},
this emulation is far from complete, and occasionally leads to inconsistent
behavior. First of all, the existing infrastructure has no clear concept
of how to compare elements, leading to possibly confusing
results when different data types are involved in
computations:
<<>>=
s <- list(1, "1")
union(s, s)
intersect(s, s)
@
The reason is that most of the existing operations rely on
\codefun{match} which automatically performs type conversions
disturbing in this context.
Also, quite a few other basic operations such as
the Cartesian product, the power set, the subset predicate, etc., are
missing, let alone more specialized operations such as
the closure under union or intersection.
Then, the current facilities do not make use of a class
system, making extensions hard (if not impossible).  Another consequence
is that no distinction can be made between sequences (ordered
collections of objects) and sets (unordered collections of objects),
which is key for the definition of complex structures
where both concepts are combined such as relations.
Also, there is no support in base \proglang{R} for extensions such as fuzzy sets
or multisets.

A few extension packages available from the Comprehensive \proglang{R}
Archive Network deal
with fuzzy concepts: Package \pkg{fuzzyFDR}
\citep{sets:lewin:2007}
calculates fuzzy decision rules for multiple testing, but does not
provide any explicit data structures for fuzzy sets.
The main functions in \pkg{fso}
\citep{sets:roberts:2007} for fuzzy set ordination compute and
return, among other information, membership values represented by
numeric matrices for some variables of the the input data.
\pkg{fuzzyRankTests} \citep{sets:geyer:2007} provides statistical tests
based on
fuzzy $p$ values and fuzzy confidence intervals, the latter being
returned as two separate numeric vectors for values and memberships. The
\pkg{gcl} package \citep{sets:vinterbo:2007}
infers fuzzy rules from the input data,
encapsulated in a classifying function returned by the training
function. The rules are composed of triangular fuzzy sets, represented
by triples describing the triangles' corner points for which the
memberships become $(0,1,0)$, respectively. Similarly, the \pkg{FKBL}
package for fuzzy knowledge base learning \citep{sets:alvarez:2007}
uses sequences of triangular fuzzy sets, defined by a vector of corner
points. Finally, \pkg{fuzzyOP} \citep{sets:aklan+altindas+macit:2008}
provides support for fuzzy
numbers: A set of $n$ numbers is represented by a $k \times 2n$
numeric matrix, where two consecutive columns represent (at most $k$)
supporting points and memberships,
respectively, of the corresponding piecewise linear
membership function. If some numbers have fewer supporting points than
others, the remaining cells are filled with missing values (\code{NA}s).

The \pkg{sets} package
\citep{sets:meyer+hornik:2008} presented here
provides a flexible and customizable basic infrastructure for
\emph{finite} sets and the generalizations mentioned above, including
basic operations for fuzzy logic. Apart from complementing the data
structures implemented in
base \proglang{R}, extension packages like the ones mentioned above
could gain in flexibility from building on a common infrastructure,
facilitating data exchange and leveraging synergies.

The remainder of the paper is
structured as follows. In Section~\ref{sec:design}, we discuss the
design rationale of data structures and core algorithms.
Section~\ref{sec:sets} introduces the most important set
operations. Section~\ref{sec:gsets} starts with constructors and
specific methods for generalized sets, followed by a more focused
presentation of the fuzzy logic infrastructure, and of functionality for
handling and visualizing membership information. Section~\ref{sec:csets}
shows how generalized sets can further be customized by specifying
user-definable matching functions and
iterators.  Section~\ref{sec:examples} presents three examples before
Section~\ref{sec:conclusion} concludes.

\section{Design issues}
\label{sec:design}

There are many ways of implementing sets. Choice and efficiency largely
depend on the domain range (i.e., the number of possible values for each
element). If the domain is relatively small, i.e. in the range of integral
data types such as \code{byte}, \code{integer}, \code{word} etc., the
probably most efficient representation is an array of bits
representing the domain elements like in \proglang{Pascal}
\citep{sets:wirth:1983}.
Operations such as union and
intersection can then straightforwardly be implemented
using logical \code{OR} and \code{AND}, respectively.
This approach obviously fails for intractably large domains
(e.g., strings or recursive objects). Without
further application knowledge, one needs to resort to generic
container ADTs with efficient element access such as hash tables or
search trees (for unique elements).
Operations can then be implemented following the classical
element-based definitions: Union by
inserting all elements of the smaller set into the
larger one; intersection by creating a new set with
all elements of the smaller set also contained in the larger one;
etc.

Clearly, set comparison must be permutation invariant.
Some care is needed for nested sets. Assume, e.g., the comparison of
$A = \{1, \{2, 3\}\}$ and $B = \{1, \{3, 2\}\}$ which clearly are identical.
To implement set equality, a matching operator would be used to
check if all elements of $A$ are
contained in $B$. If elements were internally stored in this order during
creation, the objects representing
$\{2, 3\}$ and $\{3, 2\}$ would be different. Comparing two set elements
for equality would thus require to recursively compare all elements
down the nested structures, which can quickly become infeasible
computationally.
We avoid this by using a canonical ordering during set creation,
guaranteeing that identical sets have identical
physical representation as well. We chose to sort
elements using the natural order for numeric values, the
Unicode character representation for strings, and the serialization
byte sequence (as strings) for other objects.
Eventually, the ordered elements are stored in a list.

For the \pkg{sets} package, further limitations are imposed by the
extensions presented in Sections~\ref{sec:gsets} and
\ref{sec:csets}: Generalized sets require, for each element,
the membership information, and we also support user-defined, high-level
matching functions for comparing elements. Since operations defined
for generalized sets basically operate on the memberships,
it seems appropriate to store these as (generic) vectors in the same
order than the corresponding elements.
Thus, memberships of
separate sets can simply be combined element-wise.

Many operations (e.g., testing for equality, subsetting, intersection,
etc.) are based on matching elements of the sets involved.  This is
implemented by inserting the elements of the larger one into a hash
table (we use hashed environments), and to look up the elements of the
smaller set in this table \citep[p.~391]{sets:knuth:1973}.  As hash key,
we use the elements' character representation. Since different objects
might map to the same hash key, we actually store the \emph{indexes} of
the list elements, and match the actual objects using a simple linear
search. (Note that since the element list is sorted, elements with same
representation are grouped, so the search will typically be fast.)

The implementation is based on \proglang{R}'s \code{S}3 class system, allowing
the definition of generic functions, dispatching appropriate methods
depending on the first argument's class information.
Objects for sets, generalized sets, and customizable sets have
classes \class{set}, \class{gset}, and \class{cset}, respectively, with
\class{set} inheriting from \class{gset} in turn inheriting from
\class{cset}.
Suitable operators (such as \code{\&} for intersection)
are then ``overloaded'' to dispatch the right internal function
corresponding to the operands' classes by defining corresponding
methods for group generics.
Additionally, all operations can directly be accessed using the
corresponding name combined with a \samp{set\_}, \samp{gset\_}, or
\samp{cset\_} prefix to give the user
the choice of up- or downcasts when objects of different class levels
are involved in one computation. For example, consider the union of the
set $\{1\}$ and the fuzzy set $\{2/0.5\}$: using the generic operator
will give an error since the operands' classes differ. The user needs,
in fact, resolve the semantic ambiguity by explicitly choosing the
intended operation: If the result should be a
generalized (fuzzy) set, \codefun{gset\_union} should be used.  To make
the result a set (stripping membership information), s/he employs
\codefun{set\_union} instead.

\section{Sets}
\label{sec:sets}

The basic constructor for creating sets is the \codefunind{set}
function accepting any number of \proglang{R}~objects as arguments.
<<sets1,keep.source=TRUE>>=
s <- set(1L, 2L, 3L)
print(s)
@
For elements that are not sets or atomic vectors of length 1, the
print method for sets will use labels indicating the class (and
length for vectors):
<<>>=
set("test", c, set("a", 2.5), list(1, 2))
@
Mainly for cosmetic reasons, there is also a tuple class that can be
used for vectors:
<<>>=
set(1, pair(1,2), tuple(1L, 2L, 3L))
@
In addition, there is a generic \codefunind{as.set}
function coercing suitable objects to sets.
<<>>=
s2 <- as.set(2:4)
print(s2)
@
There are some basic predicate functions (and corresponding generic operators)
defined for the (in)equality (\code{!=}, \code{==}),
(proper) subset (\code{<}, \code{<=}), (proper) superset (\code{>},
\code{>=}), and element-of (\code{\%e\%}) operations:

<<setpredicates>>=
set_is_empty(set())
set(1) <= set(1,2)
@
Note that all predicate functions are vectorized for convenience:
<<>>=
1:4 %e% set(1L, 2L, 3L)
@
The sequence \code{1:4} as \emph{one} element would be looked up by
using \code{list(1:4)} on the left-hand side.
The class-specific functions dispatched by the generic operators
are \codefun{set\_contains\_element},
\codefun{set\_is\_equal}, etc.
Other than these predicate functions, one can use
\codefun{length} for the cardinality:
<<>>=
length(s)
@
\codefun{c} and \code{|} for the union, \code{\&} for the intersection,
\code{\%D\%} for the symmetric difference:
<<>>=
s | set("a")
s & s2
s %D% s2
@
\code{*} and \code{\^{ }$n$} for the ($n$-fold)
Cartesian product (yielding a set of $n$-tuples):
<<>>=
s * s2
s ^ 2L
@
and \code{2\^} for the power set:
<<>>=
2 ^ s
@
The class-specific functions
\codefunind{set\_union}, \codefunind{set\_intersection}, and
\codefunind{set\_symdiff} accept more than two arguments.\footnote{The
  $n$-ary symmetric difference of a collection of sets consists of all
  elements contained in an odd number of the sets in the collection.}
It is also possible to compute the \emph{relative} complement of a set $X$ in
$Y$, basically removing the elements of $X$ from $Y$:
<<>>=
set_complement(set(1), set(1,2,3))
@
Note, however, that for sets (as opposed to \emph{generalized} sets),
the concept of a ``universe'' is not necessarily required, and
therefore the \emph{absolute} complement of a set not a well-defined operation.
In the \pkg{sets} package, objects of class
\class{set} are special cases of generalized sets. To stay faithful to
the simplicity of the original set concept,
we \emph{define} a \class{set} object's
universe to be identical to the set itself.
The \emph{absolute} complement of a \class{set} object
is therefore always the empty set:
<<>>>=
!set(1)
@

\codefunind{set\_combn} returns the set of all subsets of specified
length:
<<<>>=
## subsets
set_combn(s, 2L)
@
\codefun{closure} and \codefun{reduction} compute the closure
and reduction under union or intersection for a set \emph{family}
(i.e., a set of sets):
<<>>=
cl <- closure(set(set(1), set(2), set(3)), "union")
print(cl)
reduction(cl, "union")
@
The \codefun{Summary} group methods will also work if defined for the
elements:
<<setsummary>>=
sum(s)
range(s)
@

Because set elements are unordered, it is not allowed to use positional
subscripting. However, sets can be subset and
elements be replaced by using the elements as index themselves:
<<>>=
s2 <- set(1,2, c, list(1,2))
print(s2)
s2[[c]] <- "foo"
s2[[list(1, 2)]] <- "bar"
print(s2)
s2[list("foo", 1)]
@
Further, iterations over \emph{all}
elements can be carried out
using \codefun{for} and \codefun{lapply}/\codefun{sapply}:
<<iterators>>=
sapply(s, sqrt)
for (i in s) print(i)
@
Note that \codefun{for} only
works because the underlying \proglang{C} code ignores
the class information, and directly
processes the low-level list representation instead. This will be
replaced by a more intelligent \dQuote{foreach} mechanism as soon as
it exists in base \proglang{R}. \codefun{sapply} and \codefun{lapply} call the generic
\codefun{as.list} function before iterating over the elements. Since a
corresponding method exists for sets objects, this is
\dQuote{safer} than using \codefun{for}.

Using \codefunind{set\_outer},
it is possible to apply a function on all factorial combinations of the
elements of two sets.  If only one set is specified, the function is
applied on all pairs of this set.
<<setouter>>=
set_outer(set(1,2), set(1,2,3), "/")
@
\section{Generalized sets}
\label{sec:gsets}

There are several extensions of sets such as \emph{fuzzy sets}
and \emph{multisets}. Both can be be seen as special cases of
\emph{fuzzy multisets}. All have in common that they are defined on some
\emph{universe} of elements, and that each element maps to some
membership information.
We present how generalized sets are constructed, and
demonstrate the effect of choosing different fuzzy logic families.

\subsection{Constructors and specific methods}

Generalized sets are created using the \codefun{gset} function. The
required arguments depend on whether membership information is specified
extensionally (listing members) or intensionally
(giving a rule for membership):

\begin{enumerate}
\item Extensional specification:
  \begin{enumerate}[a)]
  \item Specify support and memberships as separate vectors. If
    memberships are omitted, they are assumed to be 1.
  \item Specify a set of elements along with their individual membership
    grades, using the element function (\codefun{e}).
  \end{enumerate}
\item Intensional specification: Specify universe and membership function.
\end{enumerate}

\noindent Note that for efficiency reasons, \codefun{gset} will not
store elements with zero memberships grades, and the specification of a
universe is only required with membership functions.
For convenience (and storage efficiency), a
default universe can be defined using \codefun{sets\_options}.
Set-specific universes supersede the default universe, if any.
If no universe (general or specific) is defined, the support of a set will be
interpreted as its universe. For multisets, the definition of a
(general or set-specific) universe can be complemented by a maximum
multiplicity or \emph{bound}.

\noindent Without membership information, \codefun{gset} creates a set (the support is converted to a set internally):
<<keep.source=TRUE>>=
X <- c("A", "B", "C")
gset(support = X)
@
Note, however, that unlike for \class{set} objects, it is possible to
define a universe that differs from (i.e., is a proper superset of)
the support:
<<>>=
gset(support = X, universe = LETTERS[1:10])
@
A multiset requires an integer membership vector:
<<>>=
multi <- 1:3
gset(support = X, memberships = multi)
@
For fuzzy sets, the memberships need to be out of the unit interval:
<<>>=
ms <- c(0.1, 0.3, 1)
gset(support = X, memberships = ms)
@
Alternatively to separate support/membership specification, each
elements can be paired with its membership value using \codefun{e}:
<<>>=
gset(elements = list(e("A", 0.1), e("B", 0.2), e("C", 0.3)))
@
Fuzzy sets can, additionally, be created using a membership function,
applied to a specified (or the default) universe:
<<>>=
f <- function(x) switch(x, A = 0.1, B = 0.2, C = 1, 0)
gset(universe = X, charfun = f)
@
For fuzzy multisets,
the membership argument expects a list of membership grades,
either specified as vectors, or as multisets:
<<keep.source=TRUE>>=
ms2 <- list(c(0.1, 0.3, 0.4), c(1, 1),
  gset(support = ms, memberships = multi))
gset(support = X, memberships = ms2)
@
\codefun{gset\_cardinality} returns the (relative) cardinality of a
generalized set, computed as the sum (mean) of all memberships.
\codefun{gset\_support}, \codefun{gset\_memberships}, \codefun{gset\_height}
and \codefun{gset\_core} can be used to retrieve support,
memberships, height (maximum membership degree),
and the core (elements with membership 1), respectively, of a
generalized set. \codefun{gset\_charfun} returns a (point-wise defined)
characteristic function for a given gset.
Note that in general, this will be different
from the characteristic function possibly used for the creation.

As for sets, the usual operations
such as union and intersection are available:
<<>>=
X <- gset(c("A", "B", "C"), 4:6)
Y <- gset(c("B", "C", "D"), 1:3)

X | Y
X & Y
@
Additionally, the product (\codefun{gset\_product}), sum (\code{+}),
and difference (\code{-}) of sets are defined,
which multiply, add, and subtract
multiplicities (or memberships for fuzzy sets):
<<>>=
X + Y
X - Y
gset_product(X, Y)
@
For fuzzy (multi-)sets, not only the relative, but also the
absolute complement (\code{!}) is defined:
<<>>=
!gset(1, 0.3)
X <- gset("a", universe = letters[1:3])
!X
!!X
!gset(1L, 2, universe = 1:3, bound = 3)
@

\codefun{gset\_mean} creates a new set by averaging corresponding
memberships using the arithmetic, geometric or harmonic mean.
Note that missing elements have 0 membership degree:
<<>>=
x <- gset(1:3, 1:3/3)
y <- gset(1:2, 1:2/2)
gset_mean(x, y)
gset_mean(x, y, "harmonic")

gset_mean(set(1), set(1, 2))
@

The membership vector of a generalized set can be transformed via
\linebreak
\codefun{gset\_transform\_memberships}, applying any
\emph{vectorized} function to the memberships:
<<>>=
x <- gset(1:10, 1:10/10)
gset_transform_memberships(x, pmax, 0.5)
@
Note that for multisets, an element's
membership (multiplicity) $m$ is interpreted as a one-vector of length
$m$, yielding possibly unexpected results:
<<>>=
x <- gset(1, 2)
gset_transform_memberships(x, `*`, 0.5)
@
For multisets, the \codefun{rep} function is a more natural choice
for membership transformations:
<<>>=
rep(x, 0.5)
@
In addition, three convenience functions are defined for fuzzy (multi-)sets:
\linebreak
\codefun{gset\_concentrate} and
\codefun{gset\_dilate} apply the square and the square root function,
and \codefun{gset\_normalize} normalizes the memberships to a specified
maximum:
<<>>=
gset_dilate(y)
gset_concentrate(y)
gset_normalize(y, 0.5)
@
(Note that these functions clearly have no effect for multisets.)

\subsection{Fuzzy logic and fuzzy sets}

For fuzzy (multi-)sets, the user can choose the logic underlying the
operations using the \codefun{fuzzy\_logic} function.  Fuzzy logics are
represented as named lists with four components \code{N}, \code{T},
\code{S}, and \code{I} containing the corresponding functions for
negation, conjunction (``t-norm''), disjunction (``t-conorm''), and
(residual) implication \citep{sets:klement+mesiar+pap:2000}.  The fuzzy
logic is selected by calling \codefun{fuzzy\_logic} with a character
string specifying the fuzzy logic ``family'', and optional
parameters. The exported functions \codefun{.N.}, \codefun{.T.},
\codefun{.S.}, and \codefun{.I.} reflect the currently selected
bindings. Available families include:  \code{"Zadeh"} (default),
\code{"drastic"}, \code{"product"}, \code{"Lukasiewicz"},
\code{"Fodor"}, \code{"Frank"}, \code{"Hamacher"},
\code{"Schweizer-Sklar"}, \code{"Yager"}, \code{"Dombi"},
\code{"Aczel-Alsina"}, \code{"Sugeno-Weber"}, \linebreak
\code{"Dubois-Prade"}, and
\code{"Yu"} (see Appendix~\ref{sec:fuzzyfamilies}).  A call to
\codefun{fuzzy\_logic} without arguments returns the current logic.

<<fuzzy>>=
x <- 1:10 / 10
y <- rev(x)
.S.(x, y)
fuzzy_logic("Fodor")
.S.(x, y)
@
Fuzzy set operations automatically use the active fuzzy logic setting:
<<keep.source=TRUE>>=
X <- gset(c("A", "B", "C"), c(0.3, 0.5, 0.8))
print(X)
Y <- gset(c("B", "C", "D"), c(0.1, 0.3, 0.9))
print(Y)
@
First, we try the Zadeh logic (default):
<<>>=
fuzzy_logic("Zadeh")
X & Y
X | Y
gset_complement(X, Y)
@
The results are different by switching to the Fodor logic:
<<>>=
fuzzy_logic("Fodor")
X & Y
X | Y
gset_complement(X, Y)
@
<<fuzzyoff,echo=FALSE,results=hide>>=
fuzzy_logic("Zadeh")
@
The \codefun{cut} method for generalized sets
\dQuote{filters} all elements with memberships
exceeding a specified level ($\alpha$-cuts)---the result, thus,
is a crisp (multi)set:
<<>>=
cut(X, 0.5)
@
The method can also be used for $\nu$-cuts, selecting elements
according to their multiplicity.

\subsection{Characteristic functions and their visualization}
\label{sec:plots}

The \pkg{sets} package provides several generators of
characteristic functions to be used as templates for the creation of
fuzzy sets, including the following shapes:
gaussian curve (\codefun{fuzzy\_normal}),
double gaussian curve (\codefun{fuzzy\_two\_normals}),
bell curve (\codefun{fuzzy\_bell}),
sigmoid curve (\codefun{fuzzy\_sigmoid}),
$\Pi$-like curves (\codefun{fuzzy\_pi3}, \codefun{fuzzy\_pi4}),
trapezoid
\linebreak
(\codefun{fuzzy\_trapezoid}), and triangle (\codefun{fuzzy\_triangular},
\codefun{fuzzy\_cone}). For
example, a fuzzy normal function and a corresponding fuzzy set
are created using:
<<>>=
N <- fuzzy_normal(mean = 0, sd = 1)
N(-3:3)

gset(charfun = N, universe = -3:3)
@
For convenience, we
also provide wrappers that directly generate corresponding sets,
given a specified universe:
<<>>=
fuzzy_normal_gset(universe = -3:3)
@
(If no universe is specified, the default universe is used; if this is
also missing, the universe currently defaults to \code{seq(0, 20, 0.1)}.)
It is also possible to create function generators for characteristic
functions from other functions (such as distribution functions):
<<>>=
fuzzy_poisson <- charfun_generator(dpois)
gset(charfun = fuzzy_poisson(10), universe = seq(0, 20, 2))
@
\codefun{fuzzy\_tuple} generates a sequence (tuple) of
sets based on any of the generating functions (except
\codefun{fuzzy\_trapezoid} and \codefun{fuzzy\_triangular}). The chosen
generating function is called with different
values (chosen along the universe) passed to the
first argument, thus varying the position or the resulting graph:
<<>>=
## creating a series of fuzzy normal sets
fuzzy_tuple(fuzzy_normal, 5)
@
(\verb|<<gset(201)>>| denotes an object of class \class{gset} with
201 elements---the size of the default universe).
The \pkg{sets} package provides support for visualizing the
membership functions of generalized
sets, and in particular fuzzy sets.
For (fuzzy) multisets, the plot method
produces a (grouped) barplot for
the membership vector (see Figure~\ref{fig:plots}, top left):
<<plot1,fig=FALSE>>=
## a fuzzy multiset
X <- gset(c("A", "B"), list(1:2/2, 0.5))
plot(X)
@
Characteristic function generators can directly be plotted using a
default universe (see Figure~\ref{fig:plots}, top right):
<<plot2,fig=FALSE>>=
plot(fuzzy_bell)
@
There is a plot method for tuples for visualizing a sequence
of sets (see Figure~\ref{fig:plots}, bottom left):
<<plot3,fig=FALSE>>=
## creating a sequence of sets
plot(fuzzy_tuple(fuzzy_cone, 10), col = gray.colors(10))
@
Plots of several sets can be superposed using the line method (see
Figure~\ref{fig:plots}, bottom right):
<<plot4,fig=FALSE,keep.source=TRUE>>=
x <- fuzzy_normal_gset()
y <- fuzzy_trapezoid_gset(corners = c(5, 10, 15, 17), height = c(0.7, 1))
plot(tuple(x, y), lty = 3)
lines(x | y, col = 2)
lines(gset_mean(x, y), col = 3, lty = 2)
@

\setkeys{Gin}{width=0.48\textwidth}
\begin{figure}[t!]
\begin{center}
<<plot1fig,fig=TRUE,echo=FALSE,height=4,width=4.5>>=
<<plot1>>
@
\hfill
<<plot2fig,fig=TRUE,echo=FALSE,height=4,width=4.5>>=
<<plot2>>
@
\hfill
<<plot3fig,fig=TRUE,echo=FALSE,height=4,width=4.5>>=
<<plot3>>
@
\hfill
<<plot4fig,fig=TRUE,echo=FALSE,height=4,width=4.5>>=
<<plot4>>
@
\caption{Membership plots for fuzzy sets. Top left: grouped barplot
  for a fuzzy multiset. Top right: graph of a bell curve. Bottom left:
sequence of triangular functions. Bottom right: two combinations of a
normal and a trapezoid function (dotted lines: basic shapes;
solid (red) line: union; dashed (green) line: arithmetic mean).}
\label{fig:plots}
\end{center}
\end{figure}

Finally, we note that the \pkg{sets} package provides basic
infrastructure for fuzzy inference
\linebreak
(\codefun{fuzzy\_inference}), involving the definition of
linguistic variables (\codefun{fuzzy\_variable}, \codefun{fuzzy\_partition})
and fuzzy rules (\codefun{fuzzy\_rule}) to build a fuzzy
system
\linebreak
(\codefun{fuzzy\_system}). A few methods for defuzzification
(\codefun{gset\_defuzzify}) are also provided.

\newpage

\section{User-definable extensions}
\label{sec:csets}

We added \emph{customizable sets}
extending generalized sets in two ways: First, users
can control the way elements are matched, i.e., define equivalence
classes of elements. Second, arbitrary iteration orders can be specified.

\subsection{Matching functions}

By default, sets and generalized sets use
\codefun{identical} to match elements which is maximally
restrictive. Note that this differs from the behavior of
\proglang{R}'s equality operator or
\codefun{match} which perform implicit type conversions
and thus confound, e.g., \code{1}, \code{1L} and \code{"1"}. In the
following example, note that on most computer systems, $3.3-2.2$ will not be
identical to $1.1$ due to numerical issues.
<<keep.source=TRUE>>=
x <- set("1", 1L, 1, 3.3 - 2.2, 1.1)
print(x)

y <- set(1, 1.1, 2L, "2")
print(y)
1L %e% y

x | y
@
Customizable sets can be created using the \codefun{cset} constructor,
specifying the generalized set and some matching function.
<<>>=
X <- cset(x, matchfun = match)
print(X)
Y <- cset(y, matchfun = match)
print(Y)
1L %e% Y
X | Y
@
Matching functions take two
vector arguments, say, \code{x} and \code{table},
with \code{table} being a vector where the elements of \code{x} are looked up.
The function should be vectorized in the \code{x}, i.e. return the
first matching position for each element of \code{x}.
In order to make use of non-vectorized
predicates such as \codefun{all.equal}, the \pkg{sets} package
provides \codefun{matchfun} to generate one:
<<>>=
FUN <- matchfun(function(x, y) isTRUE(all.equal(x, y)))
X <- cset(x, matchfun = FUN)
print(X)
Y <- cset(y, matchfun = FUN)
print(Y)
1L %e% Y

X | Y
@
\codefunind{sets\_options} can be used to conveniently switch the
default match and/or order function if a number of \class{cset} objects need
to be created:
<<keep.source=TRUE>>=
sets_options("matchfun", match)
cset(x)
cset(y)

cset(1:3) <= cset(c("1", "2", "3"))
@
<<echo=FALSE>>=
### restore package defaults
sets_options("matchfun", NULL)
@

\subsection{Iterators}

In addition to specifying matching functions, it is possible to
change the order in which
iterators such as \codefun{as.list} (but not \codefun{for}---see end
of Section~\ref{sec:sets})
process the elements. Note that the behavior of \codefun{as.list}
influences the labeling and print methods for
customizable sets. Sets and generalized sets
use the canonical internal ordering for iterations. With
customizable sets, a ``natural'' ordering of elements can be kept
by specifying either a permutation vector or an order function:

<<keep.source=TRUE>>=
cset(letters[1:5], orderfun = 5:1)

FUN <- function(x) order(as.character(x), decreasing = TRUE)
Z <- cset(letters[1:5], orderfun = FUN)
print(Z)
as.character(Z)
@
Note that converters for ordered factors keep the order:
<<keep.source=TRUE>>=
o <- ordered(c("a", "b", "a"), levels = c("b", "a"))
as.set(o)
as.gset(o)
as.cset(o)
@
Converters for other data types will use the order information
only if elements are unique:
<<keep.source=TRUE>>=
as.cset(c("A", "quick", "brown", "fox"))
as.cset(c("A", "quick", "brown", "fox", "quick"))
@

\section{Examples}
\label{sec:examples}

In the following, we present two examples for the use of multisets
and fuzzy multisets.

\subsection{Multisets}

Multisets are frequent in statistics since they can be seen as
frequency tables of some objects. Using the \pkg{sets} package, a
\dQuote{generalized} table can easily be constructed from a list of \proglang{R}
objects using the \codefun{as.gset} coercion function. Assume, e.g., that one
samples a number of fourfold tables given the margins using \codefun{r2dtable}:
<<>>=
set.seed(4711)
l <- r2dtable(1000, r = 1:2, c = 2:1)
@
Since the sum of the first row (and second column) are constrained to
$1$, the top left cell entry can only be $0$ or $1$. Also, given the marginals,
there is only one degree of freedom in fourfold tables,
so the value of this first cell determines the others, and thus
only two possible tables exist:
<<>>=
l[1:2]
@
To count them, we can simply use \codefun{as.gset} that will construct a
multiset from the list:
<<>>=
s <- as.gset(l)
print(s)
@
Replace the matrices by the first cells' values:
<<>>=
for (i in s) s[[i]] <- i[1]
print(s)
@
The estimated probabilities of having $0$ or $1$ in the first cell can thus
be obtained by:
<<>>=
gset_memberships(s) / 1000
@
The probability for $0$
clearly corresponds to the p value of the corresponding Fisher test:
<<>>=
fisher.test(l[[1]])$p.value
@

\subsection{Fuzzy multisets}

Fuzzy multisets can be used to represent objects appearing several
times with different membership grades (e.g., weights, degrees of
credibility, \dots). \cite{sets:mizutani+inokuchi+miyamoto:2008}
describe an interesting application of fuzzy multisets to text mining:
The occurrences of some terms of interest (``neural network'',
``fuzzy'', ``image'') in titles, abstracts, and keywords
of 30 documents on fuzzy theory are represented by fuzzy
multisets, with varying memberships depending on whether a term occurs
in the title (degree 1), the keywords (degree 0.6), and/or the
abstract (degree 0.2).
<<>>=
data("fuzzy_docs")
print(fuzzy_docs[8:9])
@
This information is then used to compute distances between documents,
and ultimately to compare several (non-linear) clustering methods
regarding their abilities of recovering the true underlying
structure. In fact, it is known that the first 12 documents are
related to neural networks, and the remaining 18 to image
processing.

In the following, we will perform simple hierarchical
clustering. We start by computing a distance matrix for the 30
documents. The \pkg{sets} package implements the
Jaccard dissimilarity, defined for two generalized sets $X$ and $Y$ as
$1 - |X \cap Y| / |X \cup Y|$ where $|\cdot|$
denotes the cardinality for generalized sets.
A corresponding dissimilarity matrix can be obtained using,
e.g., the \pkg{proxy} package \citep{sets:meyer+buchta:2008}:
<<>>=
library("proxy")
d <- dist(fuzzy_docs, gset_dissimilarity)
@
We then apply Ward's clustering method:
<<>>=
cl1 <- hclust(d, "ward")
@
resulting in a clustering depicted in Figure~\ref{fig:cluster1}.
Clearly, the neural network-related documents (\#1--\#12) are separated from
the image processing papers:
<<>>=
labs1 <- cutree(cl1, 2)
print(labs1)
@

\setkeys{Gin}{width=0.85\textwidth}
\begin{figure}[h]
\begin{center}
<<clusterplot,fig=TRUE,echo=FALSE,height=5.5,width=7>>=
plot(cl1)
@
\caption{Dendrogram for the \texttt{fuzzy\_docs} data, clustered using
  Ward's method on Jaccard distances computed from fuzzy multisets.}
\label{fig:cluster1}
\end{center}
\end{figure}

Note that for this data, using
different weightings for terms in titles, keywords and
abstracts are key to recover the subgroups.
Naive text mining approaches operate on ``classical''
term-document-matrices only, counting term occurrences without further
information on their relevance:
<<keep.source=TRUE>>=
tdm <- set_outer(c("neural networks", "fuzzy", "image"),
  fuzzy_docs, `%in%`)
print(tdm)
@
By again computing Jaccard distances
<<>>=
d <- dist(tdm, "Jaccard", by_rows = FALSE)
@
and the corresponding hierarchical clustering,
visualized in Figure~\ref{fig:cluster2},
<<>>=
cl2 <- hclust(d, "ward")
@
we can see that the ``standard'' text mining approach fails to
correctly assign four documents (\#9--\#12):
<<>>=
labs2 <- cutree(cl2, 2)
print(labs2)
table(labs1, labs2)
@

\setkeys{Gin}{width=0.85\textwidth}
\begin{figure}[h]
\begin{center}
<<clusterplot2,fig=TRUE,echo=FALSE,height=5.5,width=7>>=
plot(cl2)
@
\caption{Dendrogram for the \texttt{fuzzy\_docs} data, using Ward's
  method on a term-document-matrix generated for the data.}
\label{fig:cluster2}
\end{center}
\end{figure}

\section{Conclusion}
\label{sec:conclusion}

In this paper, we described the \pkg{sets} package for \proglang{R}, providing
infrastructure for sets and generalizations thereof such as fuzzy
sets, multisets and fuzzy multisets. The fuzzy variants make use of a
dynamic fuzzy logic infrastructure offering several fuzzy logic
families. Generalized sets are further extended to allow for
user-defined iterators and matching functions. Current work focuses on
data structures and algorithms for relations,
an important application of sets.

\bibliography{sets}

\begin{appendix}

\section{Available fuzzy logic families}
\label{sec:fuzzyfamilies}

Let us
refer to \eqn{N(x) = 1 - x} as the \emph{standard} negation, and,
for a t-norm \eqn{T}, let \eqn{S(x, y) = 1 - T(1 - x, 1 - y)} be the
\emph{dual} (or complementary) t-conorm.  Available specifications and
corresponding families are as follows, with the standard negation used
unless stated otherwise.

\begin{description}
  \item[\code{"Zadeh"}]{Zadeh's logic with \eqn{T = \min} and
    \eqn{S = \max}.  Note that the minimum t-norm, also known as the
    G{\"o}del t-norm, is the pointwise largest t-norm, and that the
    maximum t-conorm is the smallest t-conorm.}
  \item[\code{"drastic"}]{The drastic logic with t-norm
    \eqn{T(x, y) = y} if \eqn{x = 1}, \eqn{x} if \eqn{y = 1}, and 0
    otherwise, and complementary t-conorm \eqn{S(x, y) = y} if
    \eqn{x = 0}, \eqn{x} if \eqn{y = 0}, and 1 otherwise.  Note that
    the drastic t-norm and t-conorm are the smallest t-norm and
    largest t-conorm, respectively.}
  \item[\code{"product"}]{The family with the product t-norm
    \eqn{T(x, y) = xy} and dual t-conorm \eqn{S(x, y) = x + y - xy}.}
  \item[\code{"Lukasiewicz"}]{The \L{}ukasiewicz logic with t-norm
    \eqn{T(x, y) = \max(0, x + y - 1)} and dual t-conorm
    \eqn{S(x, y) = \min(x + y, 1)}.}
  \item[\code{"Fodor"}]{The family with Fodor's \emph{nilpotent
      minimum} t-norm given by \eqn{T(x, y) = \min(x, y)} if
    \eqn{x + y > 1}, and 0 otherwise, and the dual t-conorm given by
  \eqn{S(x, y) = \max(x, y)} if \eqn{x + y < 1}, and 1 otherwise.}
  \item[\code{"Frank"}]{The family of Frank t-norms \eqn{T_p},
    \eqn{p \ge 0}, which gives the Zadeh, product and \L{}ukasiewicz
    t-norms for \eqn{p = 0}, 1, and \eqn{\infty}, respectively,
    and otherwise is given by
    \eqn{T(x, y) = \log_p (1 + (p^x - 1) (p^y - 1) / (p - 1))}.}
  \item[\code{"Hamacher"}]{The three-parameter family of Hamacher,
    with negation \eqn{N_\gamma(x) = (1 - x) / (1 + \gamma x)},
    t-norm  \eqn{T_\alpha(x, y) = xy / (\alpha + (1 - \alpha)(x + y - xy))},
    and t-conorm
    \eqn{S_\beta(x, y) = (x + y + (\beta - 1) xy) / (1 + \beta xy)},
    where \eqn{\alpha \ge 0} and \eqn{\beta, \gamma \ge -1}.  This
    gives a deMorgan triple (for which $N(S(x, y)) = T(N(x), N(y))$
    iff \eqn{\alpha = (1 + \beta) / (1 + \gamma)}.}
\end{description}

The following parametric families are obtained by combining the
corresponding families of t-norms with the standard negation and
complementary t-conorm.

\begin{description}
\item[\code{"Schweizer-Sklar"}]{The Schweizer-Sklar family
    \eqn{T_p}, \eqn{-\infty \le p \le \infty}, which
    gives the Zadeh (minimum), product and drastic t-norms for
    \eqn{p = -\infty}, \eqn{0}, and \eqn{\infty},
    respectively, and otherwise is given by
    \eqn{T_p(x, y) = \max(0, (x^p + y^p - 1)^{1/p})}.}
\item[\code{"Yager"}]{The Yager family \eqn{T_p}, \eqn{p \ge 0},
    which gives the drastic and minimum t-norms for \eqn{p = 0}
    and \eqn{\infty}, respectively, and otherwise is given by
    \eqn{T_p(x, y) = \max(0, 1 - ((1-x)^p + (1-y)^p)^{1/p})}.}
\item[\code{"Dombi"}]{The Dombi family \eqn{T_p}, \eqn{p \ge 0},
    which gives the drastic and minimum t-norms for \eqn{p = 0}
    and \eqn{\infty}, respectively, and otherwise is given by
    \eqn{T_p(x, y) = 0} if \eqn{x = 0} or \eqn{y = 0}, and
    \eqn{T_p(x, y) = 1 / (1 + ((1/x - 1)^p + (1/y - 1)^p)^{1/p})} if
    both \eqn{x > 0} and \eqn{y > 0}.}
\item[\code{"Aczel-Alsina"}]{The family of t-norms \eqn{T_p},
    \eqn{p \ge 0}, introduced by Acz{\'e}l and Alsina, which gives the
    drastic and minimum t-norms for \eqn{p = 0} and
    \eqn{\infty}, respectively, and otherwise is given by
    \eqn{T_p(x, y) = \exp(-(|\log(x)|^p + |\log(y)|^p)^{1/p})}.}
\item[\code{"Sugeno-Weber"}]{The family of t-norms \eqn{T_p},
    \eqn{-1 \le p \le \infty}, introduced by Weber
    with dual t-conorms introduced by Sugeno, which gives the
    drastic and product t-norms for \eqn{p = -1} and
    \eqn{\infty}, respectively, and otherwise is given by
    \eqn{T_p(x, y) = \max(0, (x + y - 1 + pxy) / (1 + p))}.}
\item[\code{"Dubois-Prade"}]{The family of t-norms \eqn{T_p},
    \eqn{0 \le p \le 1}, introduced by Dubois and Prade, which gives
    the minimum and product t-norms for \eqn{p = 0} and \eqn{1},
    respectively, and otherwise is given by
    \eqn{T_p(x, y) = xy / \max(x, y, p)}.}
\item[\code{"Yu"}]{The family of t-norms \eqn{T_p}, \eqn{p \ge -1},
    introduced by Yu, which gives the product and drastic t-norms for
    \eqn{p = -1} and \eqn{\infty}, respectively, and otherwise is
    given by \eqn{T(x, y) = \max(0, (1 + p) (x + y - 1) - p x y)}.}
\end{description}

\end{appendix}

\end{document}

\name{fuzzydocs}
\title{Documents on Fuzzy Theory}
\alias{fuzzy_docs}
\description{
  Occurence of three terms (neural networks, fuzzy, and image)
  in 30 documents retrieved
  from a Japanese article data base on fuzzy theory and systems.
}
\usage{
data("fuzzy_docs")
}
\format{
  \code{fuzzy_docs} is a list of 30 fuzzy multisets, representing the
  occurrence of the terms \dQuote{neural networks}, \dQuote{fuzzy}, and
  \dQuote{image} in each document. Each term appears with up to
  three membership values representing weights,
  depending on whether the term occurred
  in the abstract (0.2), the keywords section (0.6), and/or the title
  (1). The first 12 documents concern neural networks, the remaining 18
  image processing. In the reference, various clustering methods have
  been employed to recover the two groups in the data set.
}
\source{
  K. Mizutani, R. Inokuchi, and S. Miyamoto (2008),
  Algorithms of Nonlinear Document Clustering Based on Fuzzy Multiset
  Model,
  \emph{International Journal of Intelligent Systems},
  \bold{23}, 176--198.
}
\examples{
data(fuzzy_docs)

## compute distance matrix using Jaccard dissimilarity
d <- as.dist(set_outer(fuzzy_docs, gset_dissimilarity))

## apply hierarchical clustering (Ward method)
cl <- hclust(d, "ward")

## retrieve two clusters
cutree(cl, 2)

## -> clearly, the clusters are formed by docs 1--12 and 13--30,
## respectively.

}
\keyword{datasets}

\name{fuzzyinference}
\alias{fuzzy_rule}
\alias{fuzzy_variable}
\alias{fuzzy_partition}
\alias{fuzzy_system}
\alias{fuzzy_inference}
\alias{\%is\%}
\title{Fuzzy inference}
\description{Basic infrastructure for building and using fuzzy inference
  systems.}  
\usage{
fuzzy_inference(system, values, implication = c("minimum", "product"))
fuzzy_rule(antecedent, consequent)
fuzzy_system(variables, rules)
fuzzy_partition(varnames, FUN = fuzzy_normal, universe = NULL, \dots)
fuzzy_variable(\dots)
x \%is\% y
}
\arguments{
  \item{\dots}{For \code{fuzzy_variable}: named list of fuzzy sets (or
    membership functions from which the fuzzy sets will be created using the
    default universe). For \code{fuzy_partition}: further arguments passed to \code{FUN}.}
  \item{antecedent, consequent}{parts of an inference rule (see
    details).}
  \item{variables}{Set or tuple of fuzzy variables (note that
    tuples must be used if two variables have the same definition).}
  \item{rules}{Set of rules.}
  \item{system}{A fuzzy system.}
  \item{values}{Named list of input values to the system. The names must
    match the labels of the variable set.}
  \item{implication}{A vectorized function taking two arguments, or a
    character string indicating the parallel minimum or the product
    function.}
  \item{varnames}{Either a character vector of fuzzy category labels, to
    be used with the default locations, or
    a named numeric vector of locations.}
  \item{FUN}{Function generator for membership functions to be used for
    the fuzzy partition.}
  \item{universe}{Universal set used for computing the memberships
    grades.}
  \item{x}{The name of a fuzzy variable.}
  \item{y}{The name of a category, belonging to a fuzzy variable.}
}
\details{
  These functions can be used to create simple fuzzy inference machines
  based on fuzzy (\dQuote{linguistic}) variables and fuzzy rules. This
  involves five steps:

  \enumerate{
    \item Fuzzification of the input variables.
    \item Application of fuzzy operators (AND, OR, NOT) in the
      antecedents of some given rules.
    \item Implication from the antecedent to the consequent.
    \item Aggregation of the consequents across the rules.
    \item Defuzzification of the resulting fuzzy set.
  }
  
  Implication is based on either the minimum or the product.
  The evaluation of the logical expressions in the antecedents, as
  well as the aggregation of the evaluation result for each single rule,
  depends on the fuzzy logic currently set.
}
\value{
  For \code{fuzzy_inference}: a generalized set. For
  \code{fuzzy_rule} and \code{fuzzy_system}: an object of class
  \code{fuzzy_rule} and \code{fuzzy_system}, respectively.
  For \code{fuzzy_variable} and \code{fuzzy_partition}: an object of
  class \code{fuzzy_variable}, inheriting from \code{tuple}.
}
\seealso{
  \code{\link{set}} and \code{\link{gset}} for the
  set types, \code{\link{fuzzy_tuple}} for available fuzzy functions,
  and \code{\link{fuzzy_logic}} on the behavior of the implemented fuzzy
  operators. 
}
\examples{
## set universe
sets_options("universe", seq(from = 0, to = 25, by = 0.1))

## set up fuzzy variables
variables <-
set(service =
    fuzzy_partition(varnames =
                    c(poor = 0, good = 5, excellent = 10),
                    sd = 1.5),
    food =
    fuzzy_variable(rancid =
                   fuzzy_trapezoid(corners = c(-2, 0, 2, 4)),
                   delicious =
                   fuzzy_trapezoid(corners = c(7, 9, 11, 13))),
    tip =
    fuzzy_partition(varnames =
                    c(cheap = 5, average = 12.5, generous = 20),
                    FUN = fuzzy_cone, radius = 5)
    )

## set up rules
rules <-
set(
    fuzzy_rule(service \%is\% poor || food \%is\% rancid,
               tip \%is\% cheap),
    fuzzy_rule(service \%is\% good,
               tip \%is\% average),
    fuzzy_rule(service \%is\% excellent || food \%is\% delicious,
               tip \%is\% generous)
    )

## combine to a system
system <- fuzzy_system(variables, rules)
print(system)
plot(system) ## plots variables

## do inference
fi <- fuzzy_inference(system, list(service = 3, food = 8))

## plot resulting fuzzy set
plot(fi)

## defuzzify
gset_defuzzify(fi, "centroid")

## reset universe
sets_options("universe", NULL)
}
\keyword{math}

\name{similarity}
\alias{set_similarity}
\alias{gset_similarity}
\alias{cset_similarity}
\alias{set_dissimilarity}
\alias{gset_dissimilarity}
\alias{cset_dissimilarity}
\title{Similarity and Dissimilarity Functions}
\description{Similarities and dissimilarities for (generalized) sets.}
\usage{
set_similarity(x, y, method = "Jaccard")
gset_similarity(x, y, method = "Jaccard")
cset_similarity(x, y, method = "Jaccard")

set_dissimilarity(x, y,
                  method = c("Jaccard", "Manhattan", "Euclidean",
                             "L1", "L2"))
gset_dissimilarity(x, y,
                   method = c("Jaccard", "Manhattan", "Euclidean",
                              "L1", "L2"))
cset_dissimilarity(x, y,
                   method = c("Jaccard", "Manhattan", "Euclidean",
                              "L1", "L2"))
}
\arguments{
  \item{x, y}{Two (generalized/customizable) sets.}
  \item{method}{Character string specifying the proximity method (see below).}
}
\details{
  For two generalized sets \eqn{X} and \eqn{Y}, the
  \code{Jaccard} similarity is \eqn{|X \cap Y| / |X \cup Y|}{|X intersect Y| /
    |X U Y|} where \eqn{|\cdot|}{|.|} denotes the cardinality for
  generalized sets (sum of memberships). The \code{Jaccard} 
  dissimilarity is 1 minus the similarity.

  The \code{L1} (or \code{Manhattan}) and \code{L2} (or
  \code{Euclidean}) 
  dissimilarities are defined as
  follows. For two fuzzy multisets \eqn{A} and \eqn{B} on a
  given universe \eqn{X} with elements \eqn{x}, let
  \eqn{M_A(x)} and \eqn{M_B(x)} be functions returning the memberships of an
  element \eqn{x} in sets \eqn{A} and \eqn{B}, respectively. The
  memberships are returned in \emph{standard form},
  i.e. as an infinite vector of decreasing membership
  values, e.g. \eqn{(1, 0.3, 0, 0, \dots)}{(1, 0.3, 0, 0, ...)}.
  Let \eqn{M_A(x)_i} and \eqn{M_B(x)_i} denote the \eqn{i}th components of these
  membership vectors. Then the L1 distance is defined as:
  \deqn{d_1(A, B) = \sum_{x \in X}\sum_{i=1}{\infty}|M_A(x)_i -
  M_B(x)_i|}{d1(A, B) = sum_x sum_i |M_A(x)_i - M_B(x)_i|} and the L2
  distance as: \deqn{d_2(A, B) = \sqrt{\sum_{x \in
  X}\sum_{i=1}{\infty}|M_A(x)_i - M_B(x)_i|^2}}{d2(A, B) = sqrt{sum_x sum_i |M_A(x)_i - M_B(x)_i|^2}}
  
}
\value{
  A numeric value (similarity or dissimilarity, as specified).
}
\seealso{
  \code{\link{set}}.
}
\source{
  T. Matthe, R. De Caluwe, G. de Tre, A. Hallez, J. Verstraete, M. Leman,
  O. Cornelis, D. Moelants, and J. Gansemans (2006),
  Similarity Between Multi-valued Thesaurus
  Attributes: Theory and Application in Multimedia Systems,
  \emph{Flexible Query Answering Systems},
  Lecture Notes in Computer Science, Springer, 331--342.
		  
  K. Mizutani, R. Inokuchi, and S. Miyamoto (2008),
  Algorithms of Nonlinear Document Clustering Based on Fuzzy Multiset
  Model,
  \emph{International Journal of Intelligent Systems},
  \bold{23}, 176--198.
}
\examples{
A <- set("a", "b", "c")
B <- set("c", "d", "e")
set_similarity(A, B)
set_dissimilarity(A, B)

A <- gset(c("a", "b", "c"), c(0.3, 0.7, 0.9))
B <- gset(c("c", "d", "e"), c(0.2, 0.4, 0.5))
gset_similarity(A, B, "Jaccard")
gset_dissimilarity(A, B, "Jaccard")
gset_dissimilarity(A, B, "L1")
gset_dissimilarity(A, B, "L2")

A <- gset(c("a", "b", "c"), list(c(0.3, 0.7), 0.1, 0.9))
B <- gset(c("c", "d", "e"), list(0.2, c(0.4, 0.5), 0.8))
gset_similarity(A, B, "Jaccard")
gset_dissimilarity(A, B, "Jaccard")
gset_dissimilarity(A, B, "L1")
gset_dissimilarity(A, B, "L2")
}
\keyword{math}

\name{interval}
\alias{as.character.interval}
\alias{as.double.interval}
\alias{as.interval}
\alias{as.interval.default}
\alias{as.interval.interval}
\alias{as.interval.list}
\alias{as.interval.NULL}
\alias{as.interval.numeric}
\alias{as.interval.set}
\alias{as.interval.tuple}
\alias{as.list.interval}
\alias{integers}
\alias{integers2reals}
\alias{interval}
\alias{interval_complement}
\alias{interval_contains_element}
\alias{interval_difference}
\alias{interval_division}
\alias{interval_domain}
\alias{interval_intersection}
\alias{interval_is_bounded}
\alias{interval_is_closed}
\alias{interval_is_countable}
\alias{interval_is_degenerate}
\alias{interval_is_empty}
\alias{interval_is_equal}
\alias{interval_is_less_than_or_equal}
\alias{interval_is_greater_than_or_equal}
\alias{interval_is_less_than}
\alias{interval_is_greater_than}
\alias{interval_is_finite}
\alias{interval_is_half_bounded}
\alias{interval_is_left_bounded}
\alias{interval_is_left_closed}
\alias{interval_is_left_open}
\alias{interval_is_left_unbounded}
\alias{interval_is_proper}
\alias{interval_is_proper_subinterval}
\alias{interval_is_right_bounded}
\alias{interval_is_right_closed}
\alias{interval_is_right_open}
\alias{interval_is_right_unbounded}
\alias{interval_is_subinterval}
\alias{interval_is_unbounded}
\alias{interval_is_uncountable}
\alias{interval_measure}
\alias{interval_power}
\alias{interval_product}
\alias{interval_sum}
\alias{interval_symdiff}
\alias{interval_union}
\alias{is.interval}
\alias{max.interval}
\alias{mean.interval}
\alias{min.interval}
\alias{naturals}
\alias{naturals0}
\alias{range.interval}
\alias{reals}
\alias{reals2integers}
\alias{prod.interval}
\alias{sum.interval}
\alias{\%..\%}
\alias{\%<\%}
\alias{\%>\%}
\alias{\%<=\%}
\alias{\%>=\%}
\title{Intervals}
\description{Interval class for countable and uncountable numeric sets.}
\usage{
interval(l=NULL, r=l,
         bounds=c("[]", "[)", "(]", "()", "[[", "]]", "][",
                  "open", "closed", "left-open", "right-open",
                  "left-closed", "right-closed"),
         domain=NULL)

reals(l=NULL, r=NULL,
      bounds=c("[]", "[)", "(]", "()", "[[", "]]", "][",
               "open", "closed", "left-open", "right-open",
               "left-closed", "right-closed"))
integers(l=NULL, r=NULL)
naturals(l=NULL, r=NULL)
naturals0(l=NULL, r=NULL)
l \%..\% r

interval_domain(x)

as.interval(x)
integers2reals(x, min=-Inf, max=Inf)
reals2integers(x)

interval_complement(x, y=NULL)
interval_intersection(\dots)
interval_symdiff(\dots)
interval_union(\dots)

interval_difference(\dots)
interval_division(\dots)
interval_product(\dots)
interval_sum(\dots)

is.interval(x)
interval_contains_element(x, y)
interval_is_bounded(x)
interval_is_closed(x)
interval_is_countable(\dots)
interval_is_degenerate(x)
interval_is_empty(x)
interval_is_equal(x, y)
interval_is_less_than_or_equal(x, y)
interval_is_less_than(x, y)
interval_is_greater_than_or_equal(x, y)
interval_is_greater_than(x, y)
interval_is_finite(x)
interval_is_half_bounded(x)
interval_is_left_bounded(x)
interval_is_left_closed(x)
interval_is_left_open(\dots)
interval_is_left_unbounded(x)
interval_measure(x)
interval_is_proper(\dots)
interval_is_proper_subinterval(x, y)
interval_is_right_bounded(x)
interval_is_right_closed(x)
interval_is_right_open(\dots)
interval_is_right_unbounded(x)
interval_is_subinterval(x, y)
interval_is_unbounded(x)
interval_is_uncountable(x)
interval_power(x, n)
x \%<\% y
x \%>\% y
x \%<=\% y
x \%>=\% y
}
\arguments{
  \item{x}{For \code{as.interval()} and \code{is.interval()}: an \R
    object. For all other functions: an interval object (or any other \R object
    coercible to one).}
  \item{y}{An interval object (or any other \R object coercible to
    one).}
  \item{min, max}{Integers defining the range to be coerced.}
  \item{l, r}{Numeric values defining the bounds of the interval. For
    integer domains, these will be rounded.}
  \item{bounds}{Character string specifying whether the interval is
    open, closed, or left/right-open/closed. Symbolic shortcuts such as
    \code{"()"} or \code{"]["} for an open interval, etc., are also accepted.}
  \item{domain}{Character string specifying the domain of the interval:
    \code{"R"}, \code{"Z"}, \code{"N"}, and \code{"N0"} for the reals,
    integers, positive integers and non-negative integers,
    respectively. If unspecified, the domain will be guessed from the
    mode of the numeric values specifying the bounds.}
  \item{n}{Integer exponent.}
  \item{\dots}{Interval objects (or other \R objects coercible to
    interval objects).}
}
\details{
  An interval object represents a multi-interval, i.e., a union of
  disjoint, possibly unbounded (i.e., infinite)
  ranges of numbers---either the extended reals, or sequences of
  integers. The usual set operations (union, complement, intersection)
  and predicates (equality, (proper) inclusion) are implemented. If
  (numeric) sets and interval objects are mixed, the result will be an
  interval object. Some basic interval arithmetic operations
  (addition, subtraction, multiplication, division, power) as well
  mathematical functions (\code{log}, \code{log2}, \code{log10}, \code{exp},
  \code{abs}, \code{sqrt}, \code{trunc}, \code{round}, \code{floor},
  \code{ceiling}, \code{signif}, and the trigonometric functions)
  are defined. Note that the rounding functions will discretize the
  interval. 

  Coercion methods for the \code{as.numeric}, \code{as.list}, and
  \code{as.set} generics are implemented. \code{reals2integers()}
  discretizes a real multi-interval. \code{integers2reals()} returns a
  multi-interval of corresponding (degenerate) real intervals.

  The summary functions \code{min}, \code{max}, \code{range},
  \code{sum}, \code{mean} and \code{prod} are implemented and work
  on the interval bounds.

  \code{sets_options()} allows to change the style of open bounds
  according to the ISO 31-11 standard using reversed brackets instead of
  round parentheses (see examples).
}
\value{For the predicates: a logical value. For all other functions:
  an interval object.}
\seealso{
  \code{\link{set}} and \code{\link{gset}} for \emph{finite} (generalized) sets.
}
\examples{
#### * general interval constructor

interval(1,5)
interval(1,5, "[)")
interval(1,5, "()")

## ambiguous notation -> use alternative style
sets_options("openbounds", "][")
interval(1,5, "()")
sets_options("openbounds", "()")

interval(1,5, domain = "Z")
interval(1L, 5L)

## degenerate interval
interval(3)

## empty interval
interval()

#### * reals
reals()
reals(1,5)
reals(1,5,"()")
reals(1) ## half-unbounded

## (auto-)complement
!reals(1,5)
interval_complement(reals(1,5), reals(2, Inf))

## combine/c(reals(2,4), reals(3,5))
reals(2,4) | reals(3,5)

## intersection
reals(2,4) & reals(3,5)

## overlapping intervals
reals(2,4) & reals(3,5)
reals(2,4) & reals(4,5,"(]")

## non-overlapping
reals(2,4) & reals(7,8)
reals(2,4) | reals(7,8)
reals(2,4,"[)") | reals(4,5,"(]")

## degenerated cases
reals(2,4) | interval()
c(reals(2,4), set())

reals(2,4) | interval(6)
c(reals(2,4), set(6), 9)

## predicates
interval_is_empty(interval())
interval_is_degenerate(interval(4))
interval_is_bounded(reals(1,2))
interval_is_bounded(reals(1,Inf)) ## !! FALSE, because extended reals
interval_is_half_bounded(reals(1,Inf))
interval_is_left_bounded(reals(1,Inf))
interval_is_right_unbounded(reals(1,Inf))
interval_is_left_closed(reals(1,Inf))
interval_is_right_closed(reals(1,Inf)) ## !! TRUE

reals(1,2) <= reals(1,5)
reals(1,2) < reals(1,2)
reals(1,2) <= reals(1,2,"[)")
reals(1,2,"[)") < reals(1,2)

#### * integers
integers()
naturals()
naturals0()

3 \%..\% 5
integers(3, 5)
integers(3, 5) | integers(6,9)
integers(3, 5) | integers(7,9)

interval_complement(naturals(), integers())

naturals() <= naturals0()
naturals0() <= integers()

## mix reals and integers
c(reals(2,5), integers(7,9))
interval_complement(reals(2,5), integers())
interval_complement(integers(2,5), reals())

try(interval_complement(integers(), reals()), silent = TRUE)
## infeasible --> error

integers() <= reals()
reals() <= integers()

### interval arithmetic
x <- interval(2,4)
y <- interval(3,6)
x + y
x - y
x * y
x / y

## summary functions
min(x, y)
max(y)
range(y)
mean(y)
}
\keyword{math}

\name{outer}
\alias{set_outer}
\alias{gset_outer}
\alias{cset_outer}
\alias{tuple_outer}
\title{Outer Product of Sets (Tuples)}
\description{Outer \dQuote{product} of (g)sets (tuples).}
\usage{
set_outer(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
gset_outer(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
cset_outer(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
tuple_outer(X, Y, FUN = "*", ..., SIMPLIFY = TRUE, quote = FALSE)
}
\arguments{
  \item{X, Y}{Set (tuple) objects or vectors. If \code{Y} is omitted, \code{X}
    will be used instead. In this case, \code{FUN} can also be specified
    as \code{Y} for convenience.}
  \item{FUN}{A function or function name (character string).}
  \item{SIMPLIFY}{Logical. If \code{TRUE} and all return values of
    \code{FUN} are atomic and of length 1, the result will be an atomic
    matrix; otherwise, a recursive one (a list with \code{dim}
    attribute).}
  \item{quote}{logical indicating whether the character strings used for
    the row and column names of the returned matrix should
    be quoted.}
  \item{\dots}{Additional arguments passed to the \code{FUN}.}
}
\details{
  This function applies \code{FUN} to all pairs of elements specified in
  \code{X} and \code{Y}.  Basically intended as a replacement for
  \code{\link[base]{outer}} for sets (tuples), it will also accept any vector for
  \code{X} and \code{Y}.  The return value will be a matrix of dimension
  \code{length(X)} times \code{length(Y)}, atomic or recursive
  depending on the complexity of \code{FUN}'s return type and the
  \code{SIMPLIFY} argument.
}
\seealso{
  \code{\link{set}}, \code{\link{tuple}}, \code{\link[base]{outer}}.
}
\examples{
set_outer(set(1,2), set(1,2,3), "/")
X <- set_outer(set(1,2), set(1,2,3), pair)
X[[1,1]]
Y <- set_outer(set(1,2), set(1,2,3), set)
Y[[1,1]]
set_outer(2 ^ set(1,2,3), set_is_subset)

tuple_outer(pair(1,2), triple(1,2,3))
tuple_outer(1:5, 1:4, "^")
}
\keyword{math}

\name{tuple}
\alias{tuple}
\alias{singleton}
\alias{pair}
\alias{triple}
% \alias{quadruple}
% \alias{quintuple}
% \alias{sextuple}
% \alias{septuple}
% \alias{octuple}
% \alias{nonuple}
% \alias{decuple}

\alias{as.tuple}
\alias{is.tuple}
\alias{tuple_is_ntuple}
\alias{tuple_is_singleton}
\alias{tuple_is_pair}
\alias{tuple_is_triple}
% \alias{tuple_is_quadruple}
% \alias{tuple_is_quintuple}
% \alias{tuple_is_sextuple}
% \alias{tuple_is_septuple}
% \alias{tuple_is_octuple}
% \alias{tuple_is_nonuple}
% \alias{tuple_is_decuple}

\title{Tuples}
\description{Creation and manipulation of tuples.}
\usage{
tuple(\dots)
as.tuple(x)
is.tuple(x)
singleton(\dots)
pair(\dots)
triple(\dots)
tuple_is_singleton(x)
tuple_is_pair(x)
tuple_is_triple(x)
tuple_is_ntuple(x, n)
}
\arguments{
  \item{x}{An \R object.}
  \item{n}{A non-negative integer.}
  \item{\dots}{Possibly named \R objects (for \code{singleton},
    \code{pair}, and
    \code{triple} exactly one, two, and three, respectively.)}
}
\details{
  These functions represent basic infrastructure for handling tuples of
  general (\R) objects.  Class \code{tuple} is used in particular to
  correctly handle cartesian products of sets.  Although tuple objects
  should behave like \dQuote{ordinary} vectors, some operations might
  yield unexpected results since tuple objects are in fact list objects
  internally. The \code{Summary} methods do work if
  defined for the set elements. The \code{\link{mean}} and
  \code{\link[stats]{median}}
  methods try to convert the object to a numeric vector before calling
  the default methods.
}
\seealso{
  \code{\link{set}}.
}
\examples{
## Constructor.
tuple(1,2,3, TRUE)
triple(1,2,3)
pair(Name = "David", Height = 185)
tuple_is_triple(triple(1,2,3))
tuple_is_ntuple(tuple(1,2,3,4), 4)

## Converter.
as.tuple(1:3)

## Operations.
c(tuple("a","b"), 1)
tuple(1,2,3) * tuple(2,3,4)
rep(tuple(1,2,3), 2)
min(tuple(1,2,3))
sum(tuple(1,2,3))
}
\keyword{math}

\name{plot}
\alias{plot.gset}
\alias{plot.set}
\alias{plot.cset}
\alias{plot.tuple}
\alias{plot.charfun_generator}
\alias{lines.gset}
\alias{lines.set}
\alias{lines.cset}
\alias{lines.tuple}
\alias{lines.charfun_generator}
\title{Plot functions for generalized sets}
\description{Plot and lines functions for (tuples of) generalized sets and
  function generators of characteristic functions.}
\usage{
\method{plot}{gset}(x, type = NULL, ylim = NULL,
         xlab = "Universe", ylab = "Membership Grade", \dots)
\method{plot}{cset}(x, \dots)
\method{plot}{set}(x, \dots)
\method{plot}{tuple}(x, type = "l", ylim = NULL,
         xlab = "Universe", ylab = "Membership Grade", col = 1,
         continuous = TRUE, \dots)
\method{plot}{charfun_generator}(x, universe = NULL, \dots)

\method{lines}{gset}(x, type = "l", col = 1, continuous = TRUE,
         universe = NULL, \dots)
\method{lines}{cset}(x, \dots)
\method{lines}{set}(x, \dots)
\method{lines}{tuple}(x, col = 1, universe = NULL, \dots)
\method{lines}{charfun_generator}(x, universe = NULL, \dots)
}
\arguments{
  \item{x}{For a method for class \var{foo}, an object of class
    \var{foo}.}
  \item{type}{Same as the \code{type} argument of
    \code{\link[graphics]{plot}}.  For \code{plot.gset}
    and \code{plot.cset}, \code{"barplot"} can also be used.}
  \item{universe}{Universal set used for setting up the plot region. By
    default, this is deduced from the object(s) to be plotted.}
  \item{col}{Character or integer vector specifying the color of the
    object(s) to be plotted.}
  \item{continuous}{Logical indicating whether zero membership degrees
    \dQuote{inside} the graph should be ignored.}
  \item{xlab, ylab}{Character labels for the axes.}
  \item{ylim}{Double vector of length 2 defining the range of the y axis.}
  \item{\dots}{Further arguments passed to the default plot methods.}
}
\value{
  The main argument (invisibly).
}

\seealso{
  \code{\link{set}}, \code{\link{gset}}, and \code{\link{tuple}} for the
  set types, and \code{\link{fuzzy_normal}} for available characteristic
  functions.
}
\examples{
## basic plots
plot(gset(1:3, 1:3/3))
plot(gset(1:3, 1:3/3, universe = 0:4))
plot(gset(c("a", "b"), list(1:2/2, 0.3)))

## characteristic functions
plot(fuzzy_normal)
plot(tuple(fuzzy_normal, fuzzy_bell), col = 1:2)
plot(fuzzy_pi3_gset(min = 2, max = 15))

## superposing plots using lines()
x <- fuzzy_normal_gset()
y <- fuzzy_trapezoid_gset(corners = c(5, 10, 15, 17), height = c(0.7, 1))
plot(tuple(x, y))
lines(x | y, col = 2)
lines(x & y, col = 3)

## another example using gset_mean
x <- fuzzy_two_normals_gset(sd = c(2, 1))
y <- fuzzy_trapezoid_gset(corners = c(5, 9, 11, 15))
plot(tuple(x, y))
lines(tuple(gset_mean(x, y),
            gset_mean(x, y, "geometric"),
            gset_mean(x, y, "harmonic")),
      col = 2:4)

## creating a sequence of sets
plot(fuzzy_tuple(fuzzy_cone, 10), col = gray.colors(10))
}
\keyword{math}

\title{Fuzzy logic}
\encoding{UTF-8}
\name{fuzzy}
\alias{.N.}
\alias{.T.}
\alias{.S.}
\alias{.I.}
\alias{fuzzy_logic}
\description{Fuzzy Logic}
\usage{
fuzzy_logic(new, \dots)
.N.(x)
.T.(x, y)
.S.(x, y)
.I.(x, y)
}
\arguments{
  \item{x, y}{Numeric vectors.}
  \item{new}{A character string specifying one of the available fuzzy
    logic \dQuote{families} (see details).}
  \item{\dots}{optional parameters for the selected family.}
}
\details{
  A call to \code{fuzzy_logic()} without arguments returns the currently
  set fuzzy logic, i.e., a named list with four
  components \code{N}, \code{T}, \code{S}, and \code{I} containing the
  corresponding functions for negation, conjunction
  (\dQuote{t-norm}), disjunction (\dQuote{t-conorm}), and residual
  implication (which may not be available).

  The package provides several fuzzy logic \emph{families}.
  A concrete fuzzy logic is selected 
  by calling \code{fuzzy_logic} with a character
  string specifying the family name, and optional parameters.  Let us
  refer to \eqn{N(x) = 1 - x} as the \emph{standard} negation, and,
  for a t-norm \eqn{T}, let \eqn{S(x, y) = 1 - T(1 - x, 1 - y)} be the
  \emph{dual} (or complementary) t-conorm.  Available specifications and
  corresponding families are as follows, with the standard negation used
  unless stated otherwise.
  
  \describe{
    \item{\code{"Zadeh"}}{Zadeh's logic with \eqn{T = \min} and
      \eqn{S = \max}.  Note that the minimum t-norm, also known as the
      Gödel t-norm, is the pointwise largest t-norm, and that the
      maximum t-conorm is the smallest t-conorm.}
    \item{\code{"drastic"}}{the drastic logic with t-norm
      \eqn{T(x, y) = y} if \eqn{x = 1}, \eqn{x} if \eqn{y = 1}, and 0
      otherwise, and complementary t-conorm \eqn{S(x, y) = y} if
      \eqn{x = 0}, \eqn{x} if \eqn{y = 0}, and 1 otherwise.  Note that
      the drastic t-norm and t-conorm are the smallest t-norm and
      largest t-conorm, respectively.}
    \item{\code{"product"}}{the family with the product t-norm
      \eqn{T(x, y) = xy} and dual t-conorm \eqn{S(x, y) = x + y - xy}.} 
    \item{\code{"Lukasiewicz"}}{the Lukasiewicz logic with t-norm
      \eqn{T(x, y) = \max(0, x + y - 1)} and dual t-conorm
      \eqn{S(x, y) = \min(x + y, 1)}.}
    \item{\code{"Fodor"}}{the family with Fodor's \emph{nilpotent
	minimum} t-norm given by \eqn{T(x, y) = \min(x, y)} if
      \eqn{x + y > 1}, and 0 otherwise, and the dual t-conorm given by
    \eqn{S(x, y) = \max(x, y)} if \eqn{x + y < 1}, and 1 otherwise.}
    \item{\code{"Frank"}}{the family of Frank t-norms \eqn{T_p},
      \eqn{p \ge 0}, which gives the Zadeh, product and Lukasiewicz
      t-norms for \eqn{p = 0}, 1, and \eqn{\infty}{Inf}, respectively,
      and otherwise is given by 
      \eqn{T(x, y) = \log_p (1 + (p^x - 1) (p^y - 1) / (p - 1))}.}
    \item{\code{"Hamacher"}}{the three-parameter family of Hamacher,
      with negation \eqn{N_\gamma(x) = (1 - x) / (1 + \gamma x)},
      t-norm
      \eqn{T_\alpha(x, y) = xy / (\alpha + (1 - \alpha)(x + y - xy))},
      and t-conorm
      \eqn{S_\beta(x, y) = (x + y + (\beta - 1) xy) / (1 + \beta xy)},
      where \eqn{\alpha \ge 0} and \eqn{\beta, \gamma \ge -1}.  This
      gives a deMorgan triple (for which \eqn{N(S(x, y)) = T(N(x), N(y))}
      iff \eqn{\alpha = (1 + \beta) / (1 + \gamma)}.  The parameters can
      be specified as \code{alpha}, \code{beta} and \code{gamma},
      respectively.  If \eqn{\alpha} is not given, it is taken as
      \eqn{\alpha = (1 + \beta) / (1 + \gamma)}. 
      The default values for \eqn{\beta} and \eqn{\gamma} are 0, so that
      by default, the product family is obtained.}
  }

  The following parametric families are obtained by combining the
  corresponding families of t-norms with the standard negation.

  \describe{
    \item{\code{"Schweizer-Sklar"}}{the Schweizer-Sklar family
      \eqn{T_p}, \eqn{-\infty \le p \le \infty}{-Inf <= p <= Inf}, which
      gives the Zadeh (minimum), product and drastic t-norms for
      \eqn{p = -\infty}{p = -Inf}, \eqn{0}, and \eqn{\infty}{Inf},
      respectively, and otherwise is given by
      \eqn{T_p(x, y) = \max(0, (x^p + y^p - 1)^{1/p})}.}
    \item{\code{"Yager"}}{the Yager family \eqn{T_p}, \eqn{p \ge 0},
      which gives the drastic and minimum t-norms for \eqn{p = 0}
      and \eqn{\infty}{Inf}, respectively, and otherwise is given by
      \eqn{T_p(x, y) = \max(0, 1 - ((1-x)^p + (1-y)^p)^{1/p})}.}
    \item{\code{"Dombi"}}{the Dombi family \eqn{T_p}, \eqn{p \ge 0},
      which gives the drastic and minimum t-norms for \eqn{p = 0}
      and \eqn{\infty}{Inf}, respectively, and otherwise is given by
      \eqn{T_p(x, y) = 0} if \eqn{x = 0} or \eqn{y = 0}, and
      \eqn{T_p(x, y) = 1 / (1 + ((1/x - 1)^p + (1/y - 1)^p)^{1/p})} if
      both \eqn{x > 0} and \eqn{y > 0}.}
    \item{\code{"Aczel-Alsina"}}{the family of t-norms \eqn{T_p},
      \eqn{p \ge 0}, introduced by Aczél and Alsina, which gives the
      drastic and minimum t-norms for \eqn{p = 0} and
      \eqn{\infty}{Inf}, respectively, and otherwise is given by
      \eqn{T_p(x, y) = \exp(-(|\log(x)|^p + |\log(y)|^p)^{1/p})}.}
    \item{\code{"Sugeno-Weber"}}{the family of t-norms \eqn{T_p},
      \eqn{-1 \le p \le \infty}{-1 <= p <= Inf}, introduced by Weber
      with dual t-conorms introduced by Sugeno, which gives the
      drastic and product t-norms for \eqn{p = -1} and
      \eqn{\infty}{Inf}, respectively, and otherwise is given by
      \eqn{T_p(x, y) = \max(0, (x + y - 1 + pxy) / (1 + p))}.}
    \item{\code{"Dubois-Prade"}}{the family of t-norms \eqn{T_p},
      \eqn{0 \le p \le 1}, introduced by Dubois and Prade, which gives
      the minimum and product t-norms for \eqn{p = 0} and \eqn{1},
      respectively, and otherwise is given by
      \eqn{T_p(x, y) = xy / \max(x, y, p)}.}
    \item{\code{"Yu"}}{the family of t-norms \eqn{T_p}, \eqn{p \ge -1},
      introduced by Yu, which gives the product and drastic t-norms for
      \eqn{p = -1} and \eqn{\infty}{Inf}, respectively, and otherwise is
      given by \eqn{T(x, y) = \max(0, (1 + p) (x + y - 1) - p x y)}.}
  }

  By default, the Zadeh logic is used.

  \code{.N.}, \code{.T.}, \code{.S.}, and \code{.I.} are dynamic
  functions, i.e., wrappers that call the corresponding function of the
  current fuzzy logic.  Thus, the behavior of code using these
  functions will change according to the chosen logic.
}
\references{
  C. Alsina, M. J. Frank and B. Schweizer (2006),
  \emph{Associative Functions: Triangular Norms and Copulas}.
  World Scientific. ISBN 981-256-671-6.
  
  J. Dombi (1982),
  A general class of fuzzy operators, the De Morgan class of fuzzy
  operators and fuzziness measures induced by fuzzy operators,
  \emph{Fuzzy Sets and Systems} \bold{8}, 149--163.

  J. Fodor and M. Roubens (1994),
  \emph{Fuzzy Preference Modelling and Multicriteria Decision Support}.
  Kluwer Academic Publishers, Dordrecht.

  D. Meyer and K. Hornik (2009),
  Generalized and customizable sets in R,
  \emph{Journal of Statistical Software} \bold{31}(2), 1--27.
  \doi{10.18637/jss.v031.i02}.

  B. Schweizer and A. Sklar (1983),
  \emph{Probabilistic Metric Spaces}.
  North-Holland, New York.  ISBN 0-444-00666-4.
}
\examples{
x <- c(0.7, 0.8)
y <- c(0.2, 0.3)

## Use default family ("Zadeh")
.N.(x)
.T.(x, y)
.S.(x, y)
.I.(x, y)

## Switch family and try again
fuzzy_logic("Fodor")
.N.(x)
.T.(x, y)
.S.(x, y)
.I.(x, y)
}
\keyword{math}

\name{labels}
\alias{LABELS}
\alias{LABEL}
\alias{LABEL.character}
\title{Labels from objects}
\description{
  Creates \dQuote{nice} labels from objects.
}
\usage{
LABELS(x, max_width = NULL, dots = "...", unique = FALSE,
       limit = NULL, \dots)
LABEL(x, limit = NULL, \dots)
\method{LABEL}{character}(x, limit = NULL, quote = sets_options("quote"), \dots)
}
\arguments{
  \item{x}{For \code{LABELS}, a vector of \R objects (if the object is
    not a vector, it is converted using \code{as.list}).
    For \code{LABEL}, an \R object.}
  \item{max_width}{Integer vector (recycled as needed) specifying the
    maximum label width for each component of \code{x}.
    If \code{NULL}, there is no limit, otherwise, the label will be
    truncated to \code{max_width}.}
  \item{dots}{A character string appended to a truncated label.
    If \code{NULL}, nothing is appended.}
  \item{unique}{Logical indicating whether
    \code{\link[base]{make.unique}} should be called on the final
    result.}
  \item{limit}{Maximum length of vectors or sets to be represented as
    is. Longer elements will be replaced by a label.}
  \item{quote}{Should character strings be quoted, or not?
    (default: \code{TRUE})}
  \item{\dots}{Optional arguments passed to the \code{LABEL} methods.}
}
\value{
  A character vector of labels generated from the supplied object(s).
  \code{LABELS} first checks whether the object has names and uses these
  if any; otherwise, \code{LABEL} is called for each element to generate
  a \dQuote{short} representation.

  \code{LABEL} is generic to allow user extensions.
  The current methods return the result of \code{\link{format}} if the
  argument is of length 1 (for objects of classes \code{\link{set}} and
  \code{\link{tuple}}: by default of length 5), and create a simple class
  information otherwise.
}
\examples{
LABELS(list(1, "test", X = "1", 1:5))
LABELS(set(X = as.tuple(1:20), "test", list(list(list(1,2)))))
LABELS(set(pair(1,2), set("a", 2), as.tuple(1:10)))
LABELS(set(pair(1,2), set("a", 2), as.tuple(1:10)), limit = 11)
}
\keyword{math}

\name{canonicalize set and mapping}
\alias{canonicalize_set_and_mapping}
\title{Canonicalize set and mapping}
\description{Helper function that canonicalizes set elements, and
  possibly reorders a given mapping accordingly.}
\usage{
canonicalize_set_and_mapping(x, mapping = NULL, margin = NULL)
}
\arguments{
  \item{x}{An object to be transformed into a set.}
  \item{mapping}{A list, array or data frame representing a mapping of the set.}
  \item{margin}{Margins to be reordered (ignored if \code{mapping} is a
    list). If \code{NULL}, all margins with the same length of \code{x}
    will be used.}
}
\details{
  This helper function can be used when a set is to be created from some
  object \code{x}, and another object contains some meta-information
  on the set elements in the same order than the elements
  of \code{x}. The set creation can cause the input elements to be
  permuted. By the use of this function, the meta information can be kept in
  sync with the result of iterating over the associated set.
}
\value{
  A list with three named components:
  \item{set}{The set created from \code{x}.}
  \item{mapping}{\code{mapping}, possibly reordered to match the order
    of \code{set}.}
  \item{order}{The order used for rearranging the mapping.}
}
\seealso{
  \code{\link{set}}.
}
\examples{
L <- list(c, "a", 3)
M1 <- list("a","b","c")
M2 <- matrix(1:9, ncol = 3)
canonicalize_set_and_mapping(L, M1)
canonicalize_set_and_mapping(L, M2)
canonicalize_set_and_mapping(L, M2, 1)
}
\keyword{math}

\name{cset}
\alias{cset}
\alias{as.cset}
\alias{is.cset}
\alias{cset_is_empty}
\alias{cset_is_subset}
\alias{cset_is_proper_subset}
\alias{cset_is_equal}
\alias{cset_contains_element}
\alias{cset_core}
\alias{cset_peak}
\alias{cset_universe}
\alias{cset_bound}
\alias{cset_height}
\alias{cset_union}
\alias{cset_sum}
\alias{cset_difference}
\alias{cset_product}
\alias{cset_mean}
\alias{cset_intersection}
\alias{cset_symdiff}
\alias{cset_complement}
\alias{cset_power}
\alias{cset_cartesian}
\alias{cset_combn}
\alias{cset_cardinality}
\alias{cset_charfun}
\alias{cset_memberships}
\alias{cset_transform_memberships}
\alias{cset_concentrate}
\alias{cset_dilate}
\alias{cset_normalize}
\alias{cset_defuzzify}
\alias{cset_support}
\alias{cset_is_crisp}
\alias{cset_is_fuzzy_set}
\alias{cset_is_multiset}
\alias{cset_is_set_or_fuzzy_set}
\alias{cset_is_set_or_multiset}
\alias{cset_is_fuzzy_multiset}
\alias{cset_is_set}
\alias{cset_has_missings}
\alias{length.cset}
\alias{lengths.cset}
\alias{matchfun}
\alias{mean.cset}
\alias{median.cset}
\alias{cut.cset}
\alias{cset_orderfun}
\alias{cset_matchfun}
\alias{cset_orderfun<-}
\alias{cset_matchfun<-}

\title{Customizable sets}
\description{Creation and manipulation of customizable sets.}
\usage{
cset(gset,
     orderfun = sets_options("orderfun"),
     matchfun = sets_options("matchfun"))
cset_support(x)
cset_core(x, na.rm = FALSE)
cset_peak(x, na.rm = FALSE)
cset_height(x, na.rm = FALSE)
cset_memberships(x, filter = NULL)
cset_universe(x)
cset_bound(x)

cset_transform_memberships(x, FUN, \dots)
cset_concentrate(x)
cset_dilate(x)
cset_normalize(x, height = 1)
cset_defuzzify(x,
               method = c("meanofmax", "smallestofmax",
                          "largestofmax", "centroid"))

matchfun(FUN)

cset_orderfun(x)
cset_matchfun(x)
cset_orderfun(x) <- value
cset_matchfun(x) <- value

as.cset(x)
is.cset(x)

cset_is_empty(x, na.rm = FALSE)
cset_is_subset(x, y, na.rm = FALSE)
cset_is_proper_subset(x, y, na.rm = FALSE)
cset_is_equal(x, y, na.rm = FALSE)
cset_contains_element(x, e)

cset_is_set(x, na.rm = FALSE)
cset_is_multiset(x, na.rm = FALSE)
cset_is_fuzzy_set(x, na.rm = FALSE)
cset_is_set_or_multiset(x, na.rm = FALSE)
cset_is_set_or_fuzzy_set(x, na.rm = FALSE)
cset_is_fuzzy_multiset(x)
cset_is_crisp(x, na.rm = FALSE)
cset_has_missings(x)

cset_cardinality(x, type = c("absolute", "relative"), na.rm = FALSE)
cset_union(\dots)
cset_mean(x, y, type = c("arithmetic", "geometric", "harmonic"))
cset_product(\dots)
cset_difference(\dots)
cset_intersection(\dots)
cset_symdiff(\dots)
cset_complement(x, y)
cset_power(x)
cset_cartesian(\dots)
cset_combn(x, m)

\method{cut}{cset}(x, level = 1, type = c("alpha", "nu"), strict = FALSE, \dots)
\method{mean}{cset}(x, \dots, na.rm = FALSE)
% \method{median}{cset}(x, na.rm = FALSE)
## S3 method for class 'cset'
## median(x, na.rm = FALSE, ...)     [R >= 3.4.0]
## median(x, na.rm)                  [R < 3.4.0]
\method{length}{cset}(x)
\method{lengths}{cset}(x, use.names = TRUE)
}
\arguments{
  \item{x}{For \code{as.cset()} and \code{is.cset()}:
    an \R object. A (c)set object otherwise.}
  \item{y}{A (c)set object.}
  \item{gset}{A generalized set (or some other \R object coercible to it).}
  \item{matchfun}{A function for matching similar elements, comparable 
    to \code{\link{match}}, taking two arguments: \code{x} (vector of
    elements to be matched) and \code{table} (vector of elements to be
    matched against). The return value is an integer vector of the
    matching positions (or \code{NA} if there is no match).
    Note that the default behavior is to test for \emph{identity}.}
  \item{FUN}{A predicate testing for equality of two objects.}
  \item{orderfun}{A function taking a list and returning an integer vector,
    specifying the order in which an iterator processes the set
    elements. Alternatively, the index vector can be specified directly.}
  \item{value}{A new match function (order function).}
  \item{type}{For \code{gset_cardinality()}:
    cardinality type (either \code{"absolute"} or
    \code{"relative"}). For \code{gset_mean()}: mean type
    (\code{"arithmetic"}, \code{"geometric"}, or \code{"harmonic"}).
    For \code{"cut"}: either \code{"alpha"} or \code{"nu"}.}
  \item{strict}{Logical indicating whether the cut level must be 
    exceeded strictly (\dQuote{greater than}) or not (\dQuote{greater
      than or equal}).}
  \item{height}{Double from the unit interval for scaling memberships.}
  \item{e}{An object of class \code{element}.}
  \item{filter}{Optional vector of elements to be filtered.}
  \item{m}{Number of elements to choose.}
  \item{method}{Currently, only \code{"Jaccard"} is implemented.}
  \item{level}{The minimum membership level.}
  \item{use.names}{logical; should the names of \code{x} be used in the result?}
  \item{na.rm}{logical indicating whether \code{NA} values should be
    removed.}
  \item{\dots}{For \code{cset_\var{foo}()}: (c)set objects. For
    the mean and sort methods: additional parameters internally passed to
    \code{\link{mean}} and \code{\link{order}}, respectively. For
    \code{gset_transform_memberships}: further arguments passed to
    \code{FUN}. For \code{cut}: currently not used.}
}
\details{
  Customizable sets extend generalized sets in two ways: First, users
  can control the way elements are matched, i.e., define equivalence
  classes of elements. Second, an order function (or permutation index)
  can be specified for each set for changing the order in which
  iterators such as \code{as.list} process the elements. The latter in
  particular influences the labeling and print methods for
  customizable sets.

  The match function needs to be vectorized in a similar way than
  \code{\link{match}}. \code{matchfun} can be used to create such a
  function from a \dQuote{simple} predicate testing for equality
  (such as, e.g., \code{\link{identical}}). Make sure, however, to
  create the same function only once.

  Note that operations on customizable sets require the same match
  function for all sets involved. The order function can differ, but
  will then be stripped from the result.

  \code{sets_options} can be used to
  conveniently switch the default match and/or
  order function if a number of \code{cset} objects need to be created.
}
\seealso{
  \code{\link{set}} for (\dQuote{ordinary}) sets, 
  \code{\link{gset}} for generalized sets, 
  \code{\link{cset_outer}}, and
  \code{\link{tuple}} for tuples (\dQuote{vectors}).
}
\references{
  D. Meyer and K. Hornik (2009),
  Generalized and customizable sets in R,
  \emph{Journal of Statistical Software} \bold{31}(2), 1--27.
  \doi{10.18637/jss.v031.i02}.
}
\examples{
## default behavior of sets: matching of elements is very strict
## Note that on most systems, 3.3 - 2.2 != 1.1
x <- set("1", 1L, 1, 3.3 - 2.2, 1.1)
print(x)

y <- set(1, 1.1, 2L, "2")
print(y)
1L \%e\% y

set_union(x, y)
set_intersection(x, y)
set_complement(x, y)

## Now use the more sloppy match()-function. 
## Note that 1 == "1" == 1L ...
X <- cset(x, matchfun = match)
print(X)
Y <- cset(y, matchfun = match)
print(Y)
1L \%e\% Y

cset_union(X, Y)
cset_intersection(X, Y)
cset_complement(X, Y)

## Same using all.equal().
## This is a non-vectorized predicate, so use matchfun
## to generate a vectorized version:
FUN <- matchfun(function(x, y) isTRUE(all.equal(x, y)))
X <- cset(x, matchfun = FUN)
print(X)
Y <- cset(y, matchfun = FUN)
print(Y)
1L \%e\% Y

cset_union(X, Y)
cset_intersection(X, Y)
cset_complement(X, Y)

### change default functions via set_option
sets_options("matchfun", match)
cset(x)
cset(y)

cset(1:3) <= cset(c(1,2,3))

### restore package defaults
sets_options("matchfun", NULL)

### customized order function
FUN <- function(x) order(as.character(x), decreasing = TRUE)
Z <- cset(letters[1:5], orderfun = FUN)
print(Z)
as.character(Z)

## converter for ordered factors keeps order
o <- ordered(c("a", "b", "a"), levels = c("b", "a"))
as.set(o)
as.cset(o)

## converter for other data types keep order if the elements are unique:
as.cset(c("A", "quick", "brown", "fox"))
as.cset(c("A", "quick", "brown", "fox", "quick"))
}
\keyword{math}

\name{sets-internal}
\title{Internal}
\alias{.cartesian_product}
\alias{.exact_match}
\alias{.get_memberships}
\alias{.make_gset_from_support_and_memberships}
\alias{.make_set_of_tuples_from_list_of_lists}
\alias{.ntuple}
\alias{.set_subset}
\alias{fuzzy_logic_predicates}
\description{Internal functions not intended for public use.}
\keyword{ internal }

\name{closure}
\alias{closure}
\alias{closure.set}
\alias{binary_closure}
\alias{reduction}
\alias{reduction.set}
\alias{binary_reduction}
\title{Closure and reduction}
\description{Closure and reduction of (g)sets.}
\usage{
\method{closure}{set}(x, operation = c("union", "intersection"), ...)
binary_closure(x, operation = c("union", "intersection"))
\method{reduction}{set}(x, operation = c("union", "intersection"), ...)
binary_reduction(x, operation = c("union", "intersection"))
}
\arguments{
  \item{x}{For \code{binary_closure} and \code{binary_reduction}: a
    binary matrix.  A set of (g)sets otherwise.}
  \item{operation}{The set operation under which the closure or
    reduction shall be computed.}
  \item{\dots}{Currently not used.}
}
\details{
  The closure of a set \eqn{S} under some operation \eqn{OP}
  contains all elements
  of \eqn{S}, and the results of \eqn{OP} applied to all element pairs
  of  \eqn{S}.

  The reduction of a set \eqn{S} under some operation \eqn{OP}
  is the minimal subset
  of \eqn{S} having the same closure than \eqn{S} under \eqn{OP}.

  Note that the closure and reduction methods for sets
  are currently only implemented for sets of \emph{(g)sets} (families)
  and will give an error for other cases.

  \code{binary_closure} and \code{binary_reduction}
  interface efficient C code for computing
  closures and reductions of binary patterns.
  They are used by the
  high-level methods if \code{x} contains only objects of class \code{sets}.
}
\value{
  An object of same type than \code{x}.
}

\seealso{
  \code{\link{set}}, \code{\link{gset}}.
}
\author{
  The C code for binary closures is provided by Christian Buchta.
}
\examples{
## ordinary set
s <- set(set(1),set(2),set(3))
(cl <- closure(s))
(re <- reduction(cl))
stopifnot(s == re)

(cl <- closure(s, "intersection"))
(re <- reduction(cl, "intersection"))
stopifnot(s == re)

## multi set
s <- set(gset(1,1),gset(2,2),gset(3,3))
(cl <- closure(s))
(re <- reduction(cl))
stopifnot(s == re)

## fuzzy set
s <- set(gset(1,1/3),gset(2,2/3),gset(3,3/3))
(cl <- closure(s))
(re <- reduction(cl))
stopifnot(s == re)

## fuzzy multiset
s <- set(gset(1,list(set(1,0.8))), gset(2, list(gset(1,3))), gset(3,0.3))
(cl <- closure(s))
(re <- reduction(cl))
stopifnot(s == re)
}
\keyword{math}

\name{fuzzyfuns}
\alias{charfun_generator}
\alias{is.charfun_generator}
\alias{fuzzy_tuple}
\alias{fuzzy_normal}
\alias{fuzzy_two_normals}
\alias{fuzzy_bell}
\alias{fuzzy_cone}
\alias{fuzzy_sigmoid}
\alias{fuzzy_trapezoid}
\alias{fuzzy_triangular}
\alias{fuzzy_pi3}
\alias{fuzzy_pi4}
\alias{fuzzy_normal_gset}
\alias{fuzzy_two_normals_gset}
\alias{fuzzy_bell_gset}
\alias{fuzzy_sigmoid_gset}
\alias{fuzzy_trapezoid_gset}
\alias{fuzzy_triangular_gset}
\alias{fuzzy_cone_gset}
\alias{fuzzy_pi3_gset}
\alias{fuzzy_pi4_gset}
\title{Fuzzy membership functions}
\description{Fuzzy membership and set creator functions.}
\usage{
charfun_generator(FUN, height = 1)
fuzzy_tuple(FUN = fuzzy_normal, n = 5, \dots,
                 universe = NULL, names = NULL)
is.charfun_generator(x)

fuzzy_normal(mean = NULL, sd = 1, log = FALSE, height = 1, chop = 0)
fuzzy_two_normals(mean = NULL, sd = c(1,1), log = c(FALSE, FALSE),
                  height = 1, chop = 0)
fuzzy_bell(center = NULL, cross = NULL, slope = 4, height = 1, chop = 0)
fuzzy_sigmoid(cross = NULL, slope = 0.5, height = 1, chop = 0)
fuzzy_trapezoid(corners = NULL, height = c(1,1),
                return_base_corners = TRUE)
fuzzy_triangular(corners = NULL, height = 1,
                 return_base_corners = TRUE)
fuzzy_cone(center = NULL, radius = 2, height = 1,
           return_base_corners = TRUE)
fuzzy_pi3(mid = NULL, min = NULL, max = NULL, height = 1,
          return_base_corners = TRUE)
fuzzy_pi4(knots, height = 1, return_base_corners = TRUE)

fuzzy_normal_gset(mean = NULL, sd = 1, log = FALSE, height = 1,
                  chop = 0, universe = NULL)
fuzzy_two_normals_gset(mean = NULL, sd = c(1,1), log = c(FALSE, FALSE),
                      height = 1, chop = 0, universe = NULL)
fuzzy_bell_gset(center = NULL, cross = NULL, slope = 4, height = 1,
                chop = 0, universe = NULL)
fuzzy_sigmoid_gset(cross = NULL, slope = 0.5, height = 1,
                   chop = 0, universe = NULL)
fuzzy_trapezoid_gset(corners = NULL, height = c(1,1), universe = NULL,
                     return_base_corners = TRUE)
fuzzy_triangular_gset(corners = NULL, height = 1, universe = NULL,
                      return_base_corners = TRUE)
fuzzy_cone_gset(center = NULL, radius = 2, height = 1, universe = NULL,
                return_base_corners = TRUE)
fuzzy_pi3_gset(mid = NULL, min = NULL, max = NULL, height = 1,
              universe = NULL, return_base_corners = TRUE)
fuzzy_pi4_gset(knots, height = 1,
               universe = NULL, return_base_corners = TRUE)
}
\arguments{
  \item{x}{An \R object.}
  \item{n}{Positive integer indicating the number of sets to be created,
    or integer vector of location parameters used to create the sets.}
  \item{FUN}{Function to be used for creating a membership function.
    Needs to be vectorized, i.e., is expected to take a vector of set
    elements and to return a vector of numeric values.}
  \item{height}{Numeric value in the unit interval specifying the height
    of the set resulting from applying the membership function to the
    universe, i.e., the maximum value to which the values will be scaled to.}
  \item{chop}{Threshold value below which the membership function is
    truncated, i.e., has a value of 0.}
  \item{center, mean}{Numeric mean value(s) used for
    the resulting membership function.}
  \item{sd}{Numeric scale factor(s) (standard deviation(s)) used for the
    resulting membership function.}
  \item{radius}{Double added/subtracted from \code{center} to get the
    base line corners of the cone.}
  \item{log}{Logical (vector), indicating whether normal or log-normal
    distributions should be used.}
  \item{cross}{Double indicating the cross-over point for the sigmoidal
    distribution.}
  \item{slope}{Double indicating the slope at the cross-over point.}
  \item{corners}{Double values (vector of length four)
    indicating the abscissas of the four corners of the resulting
    trapezoid.}
  \item{min, mid, max}{Doubles indicating the abscissas of the three
    spline knots \eqn{(\code{min}, 0)}, \eqn{(\code{mid},
      \code{height})} and \eqn{(\code{max}, 0)}
    the curve passes through.}
  \item{knots}{Vector of doubles of length four indicating the abscissas
    of the spline knots the curve passes through.}
  \item{return_base_corners}{Logical indicating whether membership grades
    for the base line corner elements should be returned as small values
    (\code{.Machine$double.eps}) instead of 0.
    Otherwise, if a set is created from the memberships,
    the corner elements would be removed from the set.}
  \item{universe}{Universal set used for computing the memberships
    grades.}
  \item{names}{optional character vector of component labels for the return value.}
  \item{\dots}{Further arguments passed to \code{FUN}.}
  }
\details{
  These functions can be used to create sets with certain membership
  patterns. 

  The core functions are function \emph{generators}, taking parameters
  and returning a corresponding fuzzy function (i.e., with values in the
  unit interval).  All of them are normalized, i.e., scaled to have a
  maximum value of \code{height} (default: 1):

  \describe{
    \item{\code{fuzzy_normal}}{is simply based on \code{\link[stats]{dnorm}}.}
    
    \item{\code{fuzzy_two_normals}}{returns a function composed of the left and
      right parts of two normal distributions (each normalized),
      with possibly different means and standard
      deviations.}
    
    \item{\code{fuzzy_bell}}{returns a function defined as:
      \eqn{\frac{1}{\left(1 + |\frac{x - c}{w}| \right) ^ {2s}}}{1 / (1
      + |(x - c) / w|) ^ 2s}  with center \eqn{c},
      crossover points \eqn{c \pm w}{c +/- w}, and slope at the crossover
      points of \eqn{\frac{s}{2w}}{s / 2w}.}
    
    \item{\code{fuzzy_sigmoid}}{yields a function whose values are
      computed as \eqn{\frac{1}{1 + e ^ {s (c - x)}}}{1 / (1 + exp{-s (x -
	  c)})}  with slope \eqn{s} at crossover point \eqn{c}.}
    
    \item{\code{fuzzy_trapezoid}}{creates a function with
      trapezoidal shape, defined by four \code{corners} elements and two
      \code{height} values for the second and third corner (the heights of
      the first and fourth corner being fixed at 0).}
    
    \item{\code{fuzzy_triangular}}{similar to the above with only three
      corners.}
    
    \item{\code{fuzzy_cone}}{is a special case of \code{fuzzy_triangular},
      defining an isosceles triangle with corners (element, membership degree):
      \eqn{(\code{center} - \code{radius}, 0)},
      \eqn{(\code{center}, \code{height})}, and
      \eqn{(\code{center} + \code{radius}, 0)}.}
    
    \item{\code{fuzzi_pi3}}{constructs a spline based on two quadratic
      functions, passing through the knot points \eqn{(\code{min}, 0)},
      \eqn{(\code{mid}, \code{height})} and \eqn{(\code{max}, 0)}.}
    
    \item{\code{fuzzi_pi4}}{constructs a spline based on an S-shaped and
      a Z-shaped curve forming a \eqn{\pi}-shaped one, passing through
      the four knots \eqn{(\code{knots[1]}, 0)}, \eqn{(\code{knots[2]},
	\code{height})}, \eqn{(\code{knots[3]}, \code{height})}, and \eqn{(\code{knots[4]}, 0)}.
    }
  }

  \code{charfun_generator} takes a vectorized function as argument,
  returning a function normalized by \code{height}.

  The \code{fuzzy_\var{foo}_gset} functions directly generate
  generalized sets from \code{fuzzy_\var{foo}}, using the values defined by
  \code{universe}, \code{sets_options("universe")}, or \code{seq(0, 20, by
    = 0.1)} (in that order, whichever is not \code{NULL}).

  \code{fuzzy_tuple} generates a sequence of \code{n}
  sets based on any of the generating functions (except
  \code{fuzzy_trapezoid} and \code{fuzzy_triangular}).  The chosen
  generating function \code{FUN} is called with \code{n} different
  values chosen along the \code{universe} passed to the
  first argument, thus varying the position or the resulting graph.
}
\value{
  For \code{charfun_generator}, a \emph{generating function}
  taking an argument list of parameters, and returning a membership
  function, mapping elements to membership values (from of the unit
  interval).
  
  For \code{fuzzy_tuple}, a tuple of \code{n} fuzzy sets.
  
  For \code{is.charfun_generator}, a logical.

  For \code{fuzzy_\var{foo}_gset}, a fuzzy set.

  For the other functions, a membership function.
}

\seealso{
  \code{\link{set}}, \code{\link{gset}}, and \code{\link{tuple}} for the
  set types, and \code{\link{plot.gset}} for the available plot functions.
}
\examples{
## creating a fuzzy normal function
N <- fuzzy_normal(mean = 0, sd = 1)
N(-3:3)

## create a fuzzy set with it
gset(charfun = N, universe = -3:3)

## same using wrapper
fuzzy_normal_gset(universe = -3:3)

## creating a user-defined fuzzy function
fuzzy_poisson <- charfun_generator(dpois)
gset(charfun = fuzzy_poisson(10), universe = seq(0, 20, 2))

## creating a series of fuzzy normal sets
fuzzy_tuple(fuzzy_normal, 5)

## creating a series of fuzzy cones with specific locations
fuzzy_tuple(fuzzy_cone, c(2,3,7))
}
\keyword{math}

\name{gset}
\alias{gset}
\alias{as.gset}
\alias{is.gset}
\alias{gset_is_empty}
\alias{gset_is_subset}
\alias{gset_is_proper_subset}
\alias{gset_is_equal}
\alias{gset_contains_element}
\alias{gset_union}
\alias{gset_sum}
\alias{gset_product}
\alias{gset_mean}
\alias{gset_difference}
\alias{gset_intersection}
\alias{gset_symdiff}
\alias{gset_complement}
\alias{gset_power}
\alias{gset_cartesian}
\alias{gset_combn}
\alias{gset_core}
\alias{gset_peak}
\alias{gset_height}
\alias{gset_transform_memberships}
\alias{gset_concentrate}
\alias{gset_dilate}
\alias{gset_normalize}
\alias{gset_defuzzify}
\alias{\%D\%}
\alias{\%e\%}
\alias{e}
\alias{is_element}
\alias{gset_bound}
\alias{gset_cardinality}
\alias{gset_charfun}
\alias{gset_memberships}
\alias{gset_support}
\alias{gset_universe}
\alias{gset_has_missings}
\alias{gset_is_crisp}
\alias{gset_is_fuzzy_set}
\alias{gset_is_multiset}
\alias{gset_is_set_or_fuzzy_set}
\alias{gset_is_set_or_multiset}
\alias{gset_is_fuzzy_multiset}
\alias{gset_is_set}
\alias{length.gset}
\alias{lengths.gset}
\alias{mean.gset}
\alias{median.gset}
\alias{cut.gset}

\title{Generalized sets}
\description{Creation and manipulation of generalized sets.}
\usage{
gset(support, memberships, charfun, elements, universe, bound,
     assume_numeric_memberships)
as.gset(x)
is.gset(x)
gset_support(x)
gset_core(x, na.rm = FALSE)
gset_peak(x, na.rm = FALSE)
gset_height(x, na.rm = FALSE)
gset_universe(x)
gset_bound(x)

gset_memberships(x, filter = NULL)
gset_transform_memberships(x, FUN, \dots)
gset_concentrate(x)
gset_dilate(x)
gset_normalize(x, height = 1)
gset_defuzzify(x,
               method = c("meanofmax", "smallestofmax",
                          "largestofmax", "centroid"))

gset_is_empty(x, na.rm = FALSE)
gset_is_subset(x, y, na.rm = FALSE)
gset_is_proper_subset(x, y, na.rm = FALSE)
gset_is_equal(x, y, na.rm = FALSE)
gset_contains_element(x, e)

gset_is_set(x, na.rm = FALSE)
gset_is_multiset(x, na.rm = FALSE)
gset_is_fuzzy_set(x, na.rm = FALSE)
gset_is_set_or_multiset(x, na.rm = FALSE)
gset_is_set_or_fuzzy_set(x, na.rm = FALSE)
gset_is_fuzzy_multiset(x)
gset_is_crisp(x, na.rm = FALSE)
gset_has_missings(x)

gset_cardinality(x, type = c("absolute", "relative"), na.rm = FALSE)
gset_union(\dots)
gset_sum(\dots)
gset_difference(\dots)
gset_product(\dots)
gset_mean(x, y, type = c("arithmetic", "geometric", "harmonic"))
gset_intersection(\dots)
gset_symdiff(\dots)
gset_complement(x, y)
gset_power(x)
gset_cartesian(\dots)
gset_combn(x, m)

e(x, memberships = 1L)
is_element(e)

\method{cut}{gset}(x, level = 1, type = c("alpha", "nu"), strict = FALSE, \dots)
\method{mean}{gset}(x, \dots, na.rm = FALSE)
% \method{median}{gset}(x, na.rm = FALSE)
## S3 method for class 'gset'
## median(x, na.rm = FALSE, ...)     [R >= 3.4.0]
## median(x, na.rm)                  [R < 3.4.0]
\method{length}{gset}(x)
\method{lengths}{gset}(x, use.names = TRUE)
}
\arguments{
  \item{x}{For \code{e()}, \code{as.gset()} and \code{is.gset()}:
    an \R object. A (g)set object otherwise. \code{gset_memberships()}
    also accepts tuple objects.}
  \item{y}{A (g)set object.}
  \item{e}{An object of class \code{element}.}
  \item{filter}{Optional vector of elements to be filtered.}
  \item{m}{Number of elements to choose.}
  \item{support}{A set of elements giving the support of the gset
    (elements with non-zero memberships). Must be a subset of the
    universe, if specified.}
  \item{memberships}{For an (\dQuote{ordinary}) set: 1L (or simply missing).
    For a fuzzy set: a value between 0 and 1.  For a multiset: a
    positive integer.  For a fuzzy multiset: a list of
    multisets with elements from the unit interval (or a list of vectors
    interpreted as such).
    Otherwise, the argument will be transformed using \code{as.gset}.} 
  \item{elements}{A set (or list) of \code{e} objects which are
    object/memberships-pairs.}
  \item{charfun}{A function taking an object and returning the
    membership.}
  \item{bound}{Integer used to compute the absolute complement for
    (fuzzy) multisets. If \code{NULL},
    defaults to the value of \code{sets_options("bound")}.
    If the latter is also \code{NULL}, the maximum multiplicity
    will be used in computations.}
  \item{assume_numeric_memberships}{When applying \code{carfun()} to
        the universe, should numeric memberships (i.e. fuzzy sets or
	multisets) be assumed (default)? If \code{FALSE}, fuzzy multisets
	will be created.}
  \item{FUN}{A function, to be applied to a membership vector.}
  \item{type}{For \code{gset_cardinality()}:
    cardinality type (either \code{"absolute"} or
    \code{"relative"}). For \code{gset_mean()}: mean type
    (\code{"arithmetic"}, \code{"geometric"}, or \code{"harmonic"}).
    For \code{"cut"}: either \code{"alpha"} or \code{"nu"}.}
  \item{strict}{Logical indicating whether the cut level must be 
    exceeded strictly (\dQuote{greater than}) or not (\dQuote{greater
      than or equal}).}
  \item{height}{Double from the unit interval for scaling memberships.}
  \item{universe}{An optional set of elements. If \code{NULL},
    defaults to the value of \code{sets_options("universe")}.
    If the latter is also \code{NULL}, the support
    will be used in computations.}
  \item{method}{\code{"centroid"} computes the arithmetic
    mean of the set elements, using the membership values as
    weights. \code{"smallestofmax"} / \code{"meanofmax"} /
    \code{"largestofmax"} returns the minimum/mean/maximum of all
    set elements with maximal membership degree.}
  \item{level}{The minimum membership level.}
  \item{na.rm}{logical indicating whether \code{NA} values should be
    removed.}
  \item{use.names}{logical; should the names of \code{x} be used in the result?}
  \item{\dots}{For \code{gset_\var{foo}()}: (g)set objects.  For
    the mean and sort methods: additional parameters internally passed to
    \code{\link{mean}} and \code{\link{order}}, respectively. For
    \code{gset_transform_memberships}: further arguments passed to
    \code{FUN}. For \code{cut}: currently not used.}
}
\details{
  These functions represent basic infrastructure for handling
  \emph{generalized} sets of general (\R) objects.

  A generalized set (or gset) is set of pairs \eqn{(e, f)}, where
  \eqn{e} is some set element and \eqn{f} is the characteristic (or
  membership) function. For (\dQuote{ordinary}) sets
  \eqn{f} maps to \eqn{\{0, 1\}},
  for fuzzy sets into the unit interval, for multisets into the natural
  numbers, and for fuzzy multisets \eqn{f} maps to the set of multisets
  over the unit interval.
  
  The \code{gset_is_\var{foo}()} predicates
  are vectorized. In addition
  to the methods defined, one can use the following operators:
  \code{|} for the union, \code{&} for the
  intersection, \code{+} for the sum, \code{-} for
  the difference,  \code{\%D\%} for the symmetric difference,
  \code{*} and \code{^\var{n}} for the
  (\eqn{n}-fold) cartesian product, \code{2^} for the power set,
  \code{\%e\%} for the element-of predicate,
  \code{<} and \code{<=} for
  the (proper) subset predicate, \code{>} and \code{>=} for
  the (proper) superset predicate, and \code{==} and \code{!=} for
  (in)equality.
  The \code{Summary} methods do also work if
  defined for the set elements. 
  The \code{\link{mean}} and \code{\link[stats]{median}}
  methods try to convert the object to a numeric vector before calling
  the default methods. \code{set_combn} returns the gset of all
  subsets of specified length. 

  \code{gset_support}, \code{gset_core}, and \code{gset_peak}
  return the set of elements with memberships greater than zero, equal
  to one, and equal to the maximum membership, respectively.
  \code{gset_memberships} returns the membership
  vector(s) of a given (tuple of) gset(s), optionally
  restricted to the elements specified by \code{filter}.
  \code{gset_height} returns only
  the largest membership degree.
  \code{gset_cardinality} computes either the absolute or the
  relative cardinality, i.e. the memberships sum, or the absolute
  cardinality divided by the number of elements, respectively.
  The \code{length} method for gsets gives the (absolute) cardinality.
  The \code{lengths} method coerces the set to a list
  before applying the \code{length} method on its elements.
  \code{gset_transform_memberships} applies function \code{FOO} to 
  the membership vector of the supplied gset and returns the transformed
  gset. The transformed memberships are guaranteed to be in the unit
  interval.
  \code{gset_concentrate} and \code{gset_dilate} are convenience
  functions, using the square and the square root,
  respectively. \code{gset_normalize} divides the memberships by their
  maximum and scales with \code{height}.
  \code{gset_product} (\code{gset_mean}) of some gsets
  compute the gset with the corresponding memberships multiplied (averaged).
  
  The \code{cut} method provides both \eqn{\alpha}- and \eqn{\nu}-cuts.
  \eqn{\alpha}-cuts \dQuote{filter} all elements with memberships
  greater than (or equal to) \code{level}---the result, thus, is a crisp
  (multi)set. \eqn{\nu}-cuts select those elements with a
  \emph{multiplicity} exceeding \code{level}
  (only sensible for (fuzzy) multisets). 
  
  Because set elements are unordered, it is not allowed to use
  \emph{positional} indexing. However, it is possible to
  do indexing using element labels or
  simply the elements themselves (useful, e.g., for subassignment).
  In addition, it is possible to iterate over
  \emph{all} elements using \code{for} and \code{lapply}/\code{sapply}.

  \code{gset_contains_element} is vectorized in \code{e}, that is, if \code{e}
  is an atomic vector or list, the is-element operation is performed
  element-wise, and a logical vector returned. Note that, however,
  objects of class \code{tuple} are taken as atomic objects to
  correctly handle sets of tuples.
  
}
\seealso{
  \code{\link{set}} for \dQuote{ordinary} sets, 
  \code{\link{gset_outer}}, and
  \code{\link{tuple}} for tuples (\dQuote{vectors}).
}
\references{
  D. Meyer and K. Hornik (2009),
  Generalized and customizable sets in R,
  \emph{Journal of Statistical Software} \bold{31}(2), 1--27.
  \doi{10.18637/jss.v031.i02}.
}
\examples{
## multisets
(A <- gset(letters[1:5], memberships = c(3, 2, 1, 1, 1)))
(B <- gset(c("a", "c", "e", "f"), memberships = c(2, 2, 1, 2)))
rep(B, 2)
gset_memberships(tuple(A, B), c("a","c"))

gset_union(A, B)
gset_intersection(A, B)
gset_complement(A, B)

gset_is_multiset(A)
gset_sum(A, B)
gset_difference(A, B)

## fuzzy sets
(A <- gset(letters[1:5], memberships = c(1, 0.3, 0.8, 0.6, 0.2)))
(B <- gset(c("a", "c", "e", "f"), memberships = c(0.7, 1, 0.4, 0.9)))
cut(B, 0.5)
A * B
A <- gset(3L, memberships = 0.5, universe = 1:5)
!A

## fuzzy multisets
(A <- gset(c("a", "b", "d"),
         memberships = list(c(0.3, 1, 0.5), c(0.9, 0.1),
                            gset(c(0.4, 0.7), c(1, 2)))))
(B <- gset(c("a", "c", "d", "e"),
         memberships = list(c(0.6, 0.7), c(1, 0.3), c(0.4, 0.5), 0.9)))
gset_union(A, B)
gset_intersection(A, B)
gset_complement(A, B)

## other operations
mean(gset(1:3, c(0.1,0.5,0.9)))
median(gset(1:3, c(0.1,0.5,0.9)))

## vectorization
list(gset(1, 0.5), gset(2, 2L), gset()) <= gset(1, 2L)
}
\keyword{math}

\name{set}
\alias{set}
\alias{as.set}
\alias{make_set_with_order}
\alias{is.set}
\alias{set_is_empty}
\alias{set_is_subset}
\alias{set_is_proper_subset}
\alias{set_is_equal}
\alias{set_contains_element}
\alias{set_union}
\alias{set_intersection}
\alias{set_symdiff}
\alias{set_complement}
\alias{set_cardinality}
\alias{length.set}
\alias{lengths.set}
\alias{set_power}
\alias{set_cartesian}
\alias{set_combn}
\title{Sets}
\description{Creation and manipulation of sets.}
\usage{
set(\dots)
as.set(x)
make_set_with_order(x)
is.set(x)

set_is_empty(x)
set_is_subset(x, y)
set_is_proper_subset(x, y)
set_is_equal(x, y)
set_contains_element(x, e)

set_union(\dots)
set_intersection(\dots)
set_symdiff(\dots)
set_complement(x, y)
set_cardinality(x)
\method{length}{set}(x)
\method{lengths}{set}(x, use.names = TRUE)
set_power(x)
set_cartesian(\dots)
set_combn(x, m)
}
\arguments{
  \item{x}{For \code{as.set()} and \code{is.set()}:
    an \R object. A set object otherwise.}
  \item{y}{A set object.}
  \item{e}{An \R object.}
  \item{m}{Number of elements to choose.}
  \item{use.names}{logical; should the names of \code{x} be used in the result?}
  \item{\dots}{For \code{set()}: \R objects, and set objects otherwise.}
}
\details{
  These functions represent basic infrastructure for handling sets
  of general (\R) objects. The \code{set_is_\var{foo}()} predicates
  are vectorized. In addition
  to the methods defined, one can use the following operators:
  \code{|} for the union,
  \code{-} for the difference (or complement), \code{&} for the
  intersection, \code{\%D\%} for the symmetric difference,
  \code{*} and \code{^\var{n}} for the
  (\eqn{n}-fold) cartesian product, \code{2^} for the power set,
  \code{\%e\%} for the element-of predicate,
  \code{<} and \code{<=} for
  the (proper) subset predicate, \code{>} and \code{>=} for
  the (proper) superset predicate, and \code{==} and \code{!=} for
  (in)equality. The \code{length} method for sets gives the
  cardinality. The \code{lengths} method coerces the set to a list
  before applying the \code{length} method on its elements.
  \code{set_combn} returns the set of all
  subsets of specified length. The \code{Summary} methods do also work if
  defined for the set elements. The \code{\link{mean}} and
  \code{\link[stats]{median}}
  methods try to convert the object to a numeric vector before calling
  the default methods.

  Because set elements are unordered, it is not allowed to use
  \emph{positional} indexing. However, it is possible to
  do indexing using element labels or
  simply the elements themselves (useful, e.g., for subassignment).
  In addition, it is possible to iterate over
  \emph{all} elements using \code{for} and \code{lapply}/\code{sapply}.

  Note that converting objects to sets may change the internal order
  of the elements, so that iterating over the original data
  might give different results than iterating over the corresponding
  set. The permutation can be obtained using the generic function
  \code{make_set_with_order}, returning both the set and the ordering.
  \code{as.set} simply calls
  \code{make_set_with_order} internally and strips the order
  information, so user-defined
   methods for coercion have to be provided for the latter and not for
  \code{as.set}.

  Note that \code{set_union}, \code{set_intersection}, and
  \code{set_symdiff} accept any number of arguments. The \eqn{n}-ary
  symmetric difference of sets contains
  just elements which are in an odd number of the sets.

  \code{set_contains_element} is vectorized in \code{e}, that is, if \code{e}
  is an atomic vector or list, the is-element operation is performed
  element-wise, and a logical vector returned. Note that, however,
  objects of class \code{tuple} are taken as atomic objects to
  correctly handle sets of tuples.
  
}
\value{
  For the predicate functions, a vector of logicals.
  For \code{make_set_with_order},
  a list with two components \code{"set"} and \code{"order"}. For
  \code{set_cardinality} and the length method, an integer value.
  For the \code{lengths} method, an integer vector. For all
  others, a set.
}
\seealso{
  \code{\link{set_outer}},
  \code{\link{gset}} for generalized sets,
  and \code{\link{tuple}} for tuples (\dQuote{vectors}).
}
\references{
  D. Meyer and K. Hornik (2009),
  Generalized and customizable sets in R,
  \emph{Journal of Statistical Software} \bold{31}(2), 1--27.
  \doi{10.18637/jss.v031.i02}.
}
\examples{
## constructor
s <- set(1L, 2L, 3L)
s

## named elements
snamed <- set(one = 1, 2, three = 3)
snamed

## indexing by label
snamed[["one"]]

## subassignment
snamed[c(2,3)] <- c("a","b")
snamed

## a more complex set
set(c, "test", list(1, 2, 3))

## converter
s2 <- as.set(2:5)
s2

## converter with order
make_set_with_order(5:1)

## set of sets
set(set(), set(1))

## cartesian product
s * s2
s * s
s ^ 2 # same as above
s ^ 3

## power set
2 ^ s

## tuples
s3 <- set(tuple(1,2,3), tuple(2,3,4))
s3

## Predicates:

## element
1:2 \%e\% s
tuple(1,2,3) \%e\% s3

## subset
s <= s2
s2 >= s # same

## proper subset
s < s

## complement, union, intersection, symmetric difference:
s - set(1L)
s + set("a") # or use: s | set("a")
s & s
s \%D\% s2
set(1,2,3) - set(1,2)
set_intersection(set(1,2,3), set(2,3,4), set(3,4,5))
set_union(set(1,2,3), set(2,3,4), set(3,4,5))
set_symdiff(set(1,2,3), set(2,3,4), set(3,4,5))

## subsets:
set_combn(as.set(1:3),2)

## iterators:
sapply(s, sqrt)
for (i in s) print(i)

## Summary methods
sum(s)
range(s)

## mean / median
mean(s)
median(s)

## cardinality
s <- set(1, list(1, 2))
length(s)
lengths(s)

## vectorization
list(set(1), set(2), set()) == set(1)
}

\keyword{math}

\name{options}
\alias{sets_options}
\title{Options for the `sets' package}
\description{Function for getting and setting options for the
  \pkg{sets} package.}
\usage{
sets_options(option, value)
}
\arguments{
  \item{option}{character string indicating the option to get
    or set (see details). If missing, all options are returned as a list.}
  \item{value}{Value to be set. If omitted, the current value is returned.}
}
\details{
  Currently, the following options are available:

  \describe{
    \item{\code{"quote"}:}{logical specifying whether labels for
      character elements are quoted or not (default: \code{TRUE}).}

    \item{\code{"hash"}:}{logical specifying whether set elements
      are hashed or not (default: \code{TRUE}).}

    \item{\code{"matchfun"}:}{the default matching function for
      \code{\link{cset}} (default: \code{NULL}).}

    \item{\code{"orderfun"}:}{the default ordering function for
      \code{\link{cset}} (default: \code{NULL}).}

    \item{\code{"universe"}:}{the default universe for
      generalized sets (default: \code{NULL}).}
  }
}
\seealso{
  \code{\link{cset}}
}
\examples{
sets_options()
sets_options("quote", TRUE)
print(set("a"))
sets_options("quote", FALSE)
print(set("a"))
}
\keyword{math}
